/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include "winsock.h"
#include <stdio.h>
#include "hash.h"
#include "main.h"
#include "resource.h"
#include "windows.h"
#include <time.h>


int allowNSLookups = 1;
int workingNSLookups = 0;

typedef struct {
	char address[512];
	char hostent[MAXGETHOSTSTRUCT];
	int cancelled;
	HANDLE handle;
	int completed;
} RESOLVENAME_STRUCT;

BOOL CALLBACK ResolveNameDlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	RESOLVENAME_STRUCT* rs;
	
	/*if (msg == WM_INITDIALOG)
	{
		rs = (RESOLVENAME_STRUCT*) lparam;
		SetProp(hwnd,"RS",(HANDLE)lparam);

		if (rs)
			rs->handle = WSAAsyncGetHostByName(hwnd,2050,rs->address,rs->hostent,MAXGETHOSTSTRUCT);
		return 0;
	}*/
	
	if (msg == WM_COMMAND)
	{
		if (wparam == IDCANCEL)
		{
			rs = RemoveProp(hwnd,"RS");
			if (rs)
			{
				rs->cancelled = 1;
				WSACancelAsyncRequest(rs->handle);
				rs->completed = 1;
			}
			return 0;
		}
	}
	if (msg == 2050)
	{
		rs = RemoveProp(hwnd,"RS");
		if (HIWORD(lparam) != 0)
			rs->completed = -1;
		else rs->completed = 1;
		return 0;
	}
	return 0;
}

int ResolveNameDlg(RESOLVENAME_STRUCT* rs)
{
	int start;
	int rem, lastrem;
	MSG msg;

/*#define NUM_HWND 64

	static HWND arrHwnd[NUM_HWND];
	static int inuseHwnd[NUM_HWND];
	int i, thisHwnd;
	static int numHwnds = 0; */
	HWND hwnd = NULL;
	char buf[32];


	rs->completed = 0;

	/*
	// find a free handle
	thisHwnd = -1;
	for (i=0;i<NUM_HWND;i++)
	{
		if (inuseHwnd[i] == 0)
			break;
	}

	if (i < NUM_HWND)
	{
		thisHwnd = i;
		if (!arrHwnd[thisHwnd])
			hwnd = CreateDialogParam(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_RESOLVENAME),PROG_HWND,ResolveNameDlgProc,(long)rs);
		else hwnd = arrHwnd[thisHwnd];
		inuseHwnd[thisHwnd] = 1;
	}
	else 
	*/
		hwnd = CreateDialogParam(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_RESOLVENAME),PROG_HWND,ResolveNameDlgProc,(long)rs);

	SetProp(hwnd,"RS",rs);
	rs->handle = WSAAsyncGetHostByName(hwnd,2050,rs->address,rs->hostent,MAXGETHOSTSTRUCT);

	//ShowWindow(hwnd,SW_SHOW);

	start = time(NULL);
	lastrem = -1;
	
	while (1)
	{
		if (rs->completed)
		{
			//DestroyWindow(hwnd);
			/*if (thisHwnd != -1)
			{
				ShowWindow(hwnd,SW_HIDE);
				inuseHwnd[thisHwnd] = 0;
			}	
				else */
			DestroyWindow(hwnd);

			return rs->completed;
		}

		if (PeekMessage(&msg, (HWND)hwnd, 0, 0, PM_NOREMOVE))
		{
			if (GetMessage(&msg, (HWND) hwnd, 0, 0) != 0 
				&& (!GetActiveWindow() || !IsDialogMessage(GetActiveWindow(),&msg)))
			{
				TranslateMessage(&msg);         
				DispatchMessage(&msg);     
			}
		} 

		rem = 6 - (time(NULL) - start);
		if (rem != lastrem && rem != 0)
		{
			if (rem == 5)
				ShowWindow(hwnd,SW_SHOWNORMAL);

			sprintf(buf,"Cancel in (%d)",rem);
			SetDlgItemText(hwnd,IDCANCEL,buf);
			lastrem = rem;
		}
		if (rem == 0)
			SendMessage(hwnd,WM_COMMAND,IDCANCEL,0);
		Sleep(1);
	}
}

unsigned long ResolveName(char* address)
{
        HOSTENT* host;
        char tmp[128];
		int i;
		int result;
		RESOLVENAME_STRUCT rs;
		unsigned long tmpaddr;

		strcpy(tmp,address);
		for (i=0;i<(int)strlen(tmp);i++)
			if (tmp[i] == ':')
				tmp[i] = 0;

		if (tmp[0] >= '0' && tmp[0] <= '9') // its an ip
	        return inet_addr(tmp);

		if ((int) HashExistsKey(HASH_GLOBAL,"DNS_CACHE",tmp) != -1)
		{
			sscanf(HashSelectText(HASH_GLOBAL,"DNS_CACHE",tmp),"%lu",&tmpaddr);
			return tmpaddr;
		}

		
        if (!allowNSLookups)
			return 0;

        if (1 || !workingNSLookups)
		{
			strcpy(rs.address,tmp);
			rs.cancelled = 0;
			result = ResolveNameDlg(&rs);
			
			if (result == -1)
				return 0;
			
			if (rs.cancelled)
			{
				allowNSLookups = 0;
//				MSG_DialogBox(NULL,NULL,"DNS lookups will be disabled for the remainder of this session");

				return 0;
			}

			host = (HOSTENT*)rs.hostent;
			// No longer bring up the cancel box
			workingNSLookups = 1;
		}
		else
			host = gethostbyname(tmp);

        if (host) 
		{
			sprintf(tmp,"%lu",*(unsigned long*) host->h_addr_list[0]);
			HashInsertText(HASH_GLOBAL,"DNS_CACHE",address,tmp);
            return *(unsigned long*) host->h_addr_list[0];
		}               
        else {
			HashInsertText(HASH_GLOBAL,"DNS_CACHE",address,"0");	
			return 0;
		}
}

// Short functions for splitting up addr:port
unsigned long GetServerAddr(char* addr) {
	unsigned i;
	char server_addr[256];	

	strcpy(server_addr, addr);
	for (i=0;i<strlen(server_addr);i++)
		if (server_addr[i] == ':')
			server_addr[i] = 0;
	return ResolveName(server_addr);

}

short int GetServerPort(char* addr, short otherwise) {
	int port = otherwise;
	int i;

	for (i=strlen(addr) - 1;i>=0;i--)
		if (addr[i] == ':') {
			sscanf(addr+i+1,"%d",&port);
			break;
		}

	return port;
}

int CheckSocket(SOCKET s)
{
	fd_set set;
	int result;
	struct timeval timeout;
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
    
	FD_ZERO(&set);
	FD_SET(s, &set);

	result = select(0,&set,NULL,NULL, &timeout);

	if (result == SOCKET_ERROR) 
		return SOCKET_ERROR; /* Not much else we can do */

	if (FD_ISSET(s, &set)) /* Incoming data/socket closed ? */
		return 0;

	return -1;
}

int CloseSocket(SOCKET s)
{
	return closesocket(s);
}

int initWinSock(void)
{
	WSADATA wsadata;
	WORD version;

	version = MAKEWORD(1,1);

	return WSAStartup(version, &wsadata);
}

int shutdownWinSock(void)
{
	return WSACleanup();
}

