/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <windows.h>
#include "servers.h"
#include "list.h"
#include "groupshared.h"

server_t* (*MAIN_AddServer)(list_t* list, server_t* svr);
void (*MAIN_EmptyList)(list_t* list);


void AllPostUpdate(list_t* list)
{
	server_t* svr;

	MAIN_EmptyList(list);

	if (HSIG("Settings","DualPane") == 0 || HSIG("Settings","ListAllServers") == 0)
		return;

	svr = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);
	while (svr)
	{
		// Don't re-add servers that another list has dropped
		if (HashCount(svr,"lists"))
			MAIN_AddServer(list,svr);
		svr = HashNext(HASH_GLOBAL,"AllServers",HASH_PTR);
	}
}

// Create the 'internet', 'lan' and 'all' types
int __stdcall Init(group_shared_t* info) {
	group_t *grp, *def;
	
	HashInit(info->hashes,info->numHashes);
	
	def = (group_t*) info->def;
	grp = (group_t*)info->groups[0];

	// Create the standard 'internet' type
	strcpy(grp->name, "NetServers");
	strcpy(grp->desc, "Internet Servers");
	grp->add_server = def->add_server;
	grp->delete_server = def->delete_server;
	grp->cut_server = def->cut_server;
	grp->copy_server = def->copy_server;
	grp->paste_server = def->paste_server;
	grp->pastable_server = def->pastable_server;
	grp->delete_list = def->delete_list;
	grp->properties_list = def->properties_list;
	grp->add_favourites = def->add_favourites;
	grp->location = grp;

	// Create the 'lan' type
	grp = (group_t*)info->groups[1];
	strcpy(grp->name, "LANServers");
	strcpy(grp->desc, "LAN Servers");
	grp->copy_server = def->copy_server;
	grp->properties_list = def->properties_list;
	grp->refresh_list = HashSelectPtr(HASH_GLOBAL,"Funcs","RefreshLANServers");
	grp->add_favourites = def->add_favourites;
	grp->location = grp;

	// Create the standard 'favourites' type
	grp = (group_t*)info->groups[2];
	strcpy(grp->name, "Favourites");
	strcpy(grp->desc, "Favourites");
	grp->add_server = def->add_server;
	grp->delete_server = def->delete_server;
	grp->cut_server = def->cut_server;
	grp->copy_server = def->copy_server;
	grp->paste_server = def->paste_server;
	grp->pastable_server = def->pastable_server;
	grp->delete_list = NULL;
	grp->properties_list = def->properties_list;
	grp->add_favourites = NULL;
	grp->location = grp;

	
	// Create the 'all' type
	grp = (group_t*)info->groups[3];
	strcpy(grp->name, "ListAllServers");
	strcpy(grp->desc, "List All Servers");
	grp->copy_server = def->copy_server;
	grp->location = grp;
	grp->properties_list = def->properties_list;
	grp->add_favourites = def->add_favourites;
	grp->major_update = AllPostUpdate;

	MAIN_AddServer = HashSelectPtr(HASH_GLOBAL,"Funcs","AddServer");
	MAIN_EmptyList = HashSelectPtr(HASH_GLOBAL,"Funcs","EmptyList");

	
	return 4;
}