/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include <winbase.h>
#include "hash.h"

typedef struct {
	void (*func)(void* data);
	void* ptr;
	int length;
	int inittime;
	int repeat;
} EVENT_DATA;

void EVENTS_Cancel(void* data)
{
	HashDeleteValue(HASH_GLOBAL,"Events",data,HASH_PTR);
	return;
}
void EVENTS_Free(void) {
	EVENT_DATA* data;

	while ((data = HashReset(HASH_GLOBAL,"Events",HASH_PTR)) != NULL)
	{
		free (data);
		HashDeleteValue(HASH_GLOBAL,"Events",data,HASH_PTR);
	}
	return;
}

void* EVENTS_Register(void (*func)(void*),void* ptr, float length, float repeat)
{
	EVENT_DATA* data;

	data = calloc(sizeof(EVENT_DATA),1);

	data->func = func;
	data->ptr = ptr;
	data->length = (int) (length * 1000);
	data->repeat = (int) (repeat * 1000);
	data->inittime = GetTickCount();

	HashInsert(HASH_GLOBAL,"Events",HASH_ANY,data,HASH_PTR);
	return data;
}

void EVENTS_Check(void) {
	EVENT_DATA* data;

	for (data = HashReset(HASH_GLOBAL,"Events",HASH_PTR);data;)
	{
		if (data->inittime + data->length < (int) GetTickCount())
		{
			data->func(data->ptr);
			if (data->repeat) {
				data->inittime = GetTickCount();
				data->length = data->repeat;
			}
			else {
				HashDeleteValue(HASH_GLOBAL,"Events",data,HASH_PTR);
				free(data);
				data = HashReset(HASH_GLOBAL,"Events",HASH_PTR);
				continue;
			}
		}
		data = HashNext(HASH_GLOBAL,"Events",HASH_PTR);
	}
	return;
}


