/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include "types.h"
#include "list.h"
#include "http.h"
#include "events.h"
#include "comm.h"
#include "main.h"

// *** Function definitions ***

// Server operations
server_t*		(*MAIN_NewServer)			(char* name, char* ipaddr, char* qstat_type);
server_t*		(*MAIN_AddServer)			(list_t* list, server_t* svr);
void			(*MAIN_DeleteServer)		(server_t* server, list_t* list);
types_t*		(*MAIN_GetServerType)		(int type);
types_t*		(*MAIN_GetServerType_qstat)	(char* type);

// List operations
void			(*MAIN_ServersListUpdate)	(int forced);
void			(*MAIN_EmptyList)			(list_t* list);
list_t*			(*MAIN_ListNamedWithParent)	(char* name, list_t* parent);
list_t*			(*MAIN_CreateList)			(char* desc, char* group, list_t* parent);
group_t*		(*MAIN_GroupOfList)			(list_t* list);
void			(*MAIN_ListPingAllServers)	(list_t*);

// Hash operations
HASH*			(*MAIN_HashAlloc)			(void* owner, char*name, int size);

// HTTP/Events 
void			(*MAIN_HTTP_Download)		(HTTP_DATA* data);
void			(*MAIN_EVENTS_Register)		(void (*func)(void*),void* ptr, float length, int repeat);

// Memory operations
void*			(*MAIN_calloc)				(size_t,size_t);
void*			(*MAIN_malloc)				(size_t);
void*			(*MAIN_realloc)				(void*,size_t);

// IP/port manipulation
unsigned long	(*MAIN_GetServerAddr)		(char* addr);
short int		(*MAIN_GetServerPort)		(char* addr, short otherwise);


typedef struct {
	char* name;
	void* outfunc, *infunc;
} func_t;

func_t funcs[] = {
	{ "malloc", malloc, &MAIN_malloc },
	{ "calloc", calloc, &MAIN_calloc },
	{ "realloc", realloc, &MAIN_realloc },
	{ "GetServerPort", GetServerPort, &MAIN_GetServerPort },
	{ "GetServerAddr", GetServerAddr, &MAIN_GetServerAddr },
	{ "ServersListUpdate", ServersListUpdate, &MAIN_ServersListUpdate },
	{ "AddServer", AddServer, &MAIN_AddServer },
	{ "NewServer", NewServer, &MAIN_NewServer },
	{ "DeleteServer", DeleteServer, &MAIN_DeleteServer },
	{ "EmptyList", EmptyList, &MAIN_EmptyList },
	{ "HashAlloc", HashAlloc, &MAIN_HashAlloc },
	{ "GetServerType", GetServerType, &MAIN_GetServerType },
	{ "GetServerType_qstat", GetServerType_qstat, &MAIN_GetServerType_qstat },
	{ "ListNamedWithParent", ListNamedWithParent, &MAIN_ListNamedWithParent },
	{ "CreateList", CreateList, &MAIN_CreateList },
	{ "GroupOfList", GroupOfList, &MAIN_GroupOfList },
	{ "HTTP_Download", HTTP_Download, &MAIN_HTTP_Download },
	{ "EVENTS_Register", EVENTS_Register, &MAIN_EVENTS_Register },
	{ "ListPingAllServers", ListPingAllServers, &MAIN_ListPingAllServers },
	{ "", NULL, NULL }
};

void FUNC_Import(void) 
{
	int i;

	for (i=0;funcs[i].infunc;i++)
		memcpy (funcs[i].infunc, HashSelectPtr(HASH_GLOBAL,"Global",funcs[i].name), sizeof(void*));
}

void FUNC_Export(void) 
{
	int i;

	for (i=0;funcs[i].outfunc;i++)
		HashInsertPtr(HASH_GLOBAL,"Global",funcs[i].name, funcs[i].outfunc);
}
