/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include "resource.h"

#include "types.h"
#include "main.h"
#include "list.h"
#include "http.h"
#include "qstat.h" // type conversion

#include "groupshared.h"

// Function definitions
server_t* (*MAIN_NewServer)(char* name, char* ipaddr, char* qstat_type);
server_t* (*MAIN_AddServer)(list_t* list, server_t* svr);
void (*MAIN_ServersListUpdate)(int forced);
void (*MAIN_EmptyList)(list_t* list);
HASH* (*MAIN_HashAlloc)(void* owner, char*name, int size);

types_t* (*MAIN_GetServerType)(int type);
list_t* (*MAIN_ListNamedWithParent)(char* name, list_t* parent);
list_t* (*MAIN_CreateList)(char* desc, char* group, list_t* parent);
group_t* (*MAIN_GroupOfList)(list_t* list);
void (*MAIN_HTTP_Download)(HTTP_DATA* data);
void (*MAIN_EVENTS_Register)(void (*func)(void*),void* ptr, float length, int repeat);
void* (*MAIN_calloc)(size_t,size_t);
void (*MAIN_ListPingAllServers)(list_t*);


void GSPYMASTER_Receive(HTTP_DATA* data) {
	
	char name[128];
	char ipaddr[128];
	server_t* svr;

	char* str = data->data;
	char svr_type[32], svr_ip[128], svr_name[128];
	int type_num, tl;
	list_t* list;
	TV_ITEM item;

	list = data->ptr;
	
	if (CURRENT_SERVER && CURRENT_LIST == list)
	{
		strcpy(name,CURRENT_SERVER->name);
		strcpy(ipaddr,CURRENT_SERVER->ipaddr);
	}

	item.hItem = list->hTreeItem;
	item.mask = TVIF_STATE;
	TreeView_GetItem(PROG_TREEVIEW,&item);

	MAIN_EmptyList(list);
	MAIN_ServersListUpdate(UPDATE_DISABLE);

	while (str<data->data + data->length) 
	{
		strcpy(svr_name,"New Server");
		strcpy(svr_ip,"");
		strcpy(svr_type,"");

		if (*str == '\\')
			str++;

		if (memcmp(str,"ip\\",3) == 0)
		{
			str+=3;
			memcpy(svr_ip,str,32);
			for (tl=0;tl<32;tl++)
				if (svr_ip[tl] == '\\')
				{
					svr_ip[tl] = 0;
					break;
				}
			str+=strlen(svr_ip);
		
			//if (strstr(svr_ip,"202.") || strstr(svr_ip,"203.")) {
			if (svr_ip[0] >= '0' && svr_ip[0] <= '9') {
				type_num = (int) HashSelect(list,"Extra","Type",HASH_INT);
				strcpy(svr_name, svr_ip);

				svr = MAIN_AddServer(list, MAIN_NewServer(svr_name,svr_ip,type_num));
				if (svr == NULL) {

					if (item.state & TVIS_EXPANDED)
						TreeView_Expand(PROG_TREEVIEW,list->hTreeItem,TVE_EXPAND);

					MAIN_ServersListUpdate(UPDATE_ENABLE);
					MAIN_ServersListUpdate(0);
					return;
				}

				svr->unnamed = 1;
			
				if (name[0] && !strcmp(svr_ip,ipaddr)) {
						/*TreeView_SelectItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),
							svr->hTreeItem);*/
						SET_CURRENT_SERVER(svr);
				}
			}

		}
		else for (tl=0;tl<2;tl++) {
			while (str< data->data+ data->length) {
				if (str[0] == '\\')
				{
					if (!tl) str++;
					break;
				}
				else str++;
			}
		}
	}

	if (item.state & TVIS_EXPANDED)
		TreeView_Expand(PROG_TREEVIEW,list->hTreeItem,TVE_EXPAND);

	MAIN_ServersListUpdate(UPDATE_ENABLE);
	MAIN_ServersListUpdate(0);

	MAIN_ListPingAllServers(list);
	return;
}


void GSPYMASTER_UpdateListRun(list_t* list, int windowed)
{
	HTTP_DATA* data;
	
	data = MAIN_calloc(sizeof(HTTP_DATA),1);
	
	data->windowed = windowed;
	data->url = NULL;
	if (HashExists(list,"extra"))
		data->url = HashSelect(list,"Extra","URL",HASH_STRING);
	if (!data->url)
		data->url = HashSelect(&list->group, "Extra","URL",HASH_STRING);

	data->custom = malloc(128);
	sprintf(data->custom,"\\gamename\\query\\final\\\\list\\\\gamename\\%s",
		MAIN_GetServerType(HashSelectInt(list,"Extra","Type"))->gamespy_type );
	
	data->func = GSPYMASTER_Receive;
	data->ptr = list;
	MAIN_HTTP_Download(data);
}

void GSPYMASTER_UpdateList(list_t* list) {
	GSPYMASTER_UpdateListRun(list,1);
}

void GSPYMASTER_UpdateLists(void* unused)
{
	list_t* list;

	if (!HashSelect(HASH_GLOBAL,"Settings","RefreshPupServers",HASH_INT))
		return;
	
	for (list = HashReset(HASH_GLOBAL,"Lists",HASH_PTR);list;list = HashNext(HASH_GLOBAL,"Lists",HASH_PTR))
		if (strcmp(MAIN_GroupOfList(list)->name,"GSpyMaster") == 0)
			GSPYMASTER_UpdateListRun(list,0);
	return;
}

// Create the 'GSpyList' group type
int __stdcall Init(group_shared_t* info) {

	group_t* grp = (group_t*)info->groups[0];
	group_t* def = (group_t*)info->def;

	HashInit(info->hashes,info->numHashes);

	MAIN_NewServer = HashSelectPtr(HASH_GLOBAL,"Funcs","NewServer");
	MAIN_AddServer = HashSelectPtr(HASH_GLOBAL,"Funcs","AddServer");
	MAIN_EmptyList = HashSelectPtr(HASH_GLOBAL,"Funcs","EmptyList");
	MAIN_HashAlloc = HashSelectPtr(HASH_GLOBAL,"Funcs","HashAlloc");
	MAIN_ServersListUpdate = HashSelectPtr(HASH_GLOBAL,"Funcs","ServersListUpdate");

	MAIN_GetServerType = HashSelectPtr(HASH_GLOBAL,"Funcs","GetServerType");
	MAIN_ListNamedWithParent = HashSelectPtr(HASH_GLOBAL,"Funcs","ListNamedWithParent");
	MAIN_CreateList = HashSelectPtr(HASH_GLOBAL,"Funcs","CreateList");
	MAIN_GroupOfList = HashSelectPtr(HASH_GLOBAL,"Funcs","GroupOfList");
	MAIN_HTTP_Download = HashSelectPtr(HASH_GLOBAL,"Funcs","HTTP_Download");
	MAIN_EVENTS_Register = HashSelectPtr(HASH_GLOBAL,"Funcs","EVENTS_Register");
	MAIN_calloc = HashSelectPtr(HASH_GLOBAL,"Funcs","calloc");
	MAIN_ListPingAllServers = HashSelectPtr(HASH_GLOBAL,"Funcs","ListPingAllServers");

	strcpy(grp->name, "GSpyMaster");
	strcpy(grp->desc, "GameSpy Master");
	grp->refresh_list = GSPYMASTER_UpdateList;
	grp->properties_list = def->properties_list;
	grp->delete_list = def->delete_list;
	MAIN_HashAlloc(grp,"extra",4);
	HashInsert(grp,"extra", "URL", "www.serverquery.com:28900",HASH_STRING);
	HashInsert(grp,"extra", "Type", "6",HASH_STRING);
	grp->copy_server = def->copy_server;
	grp->add_favourites = def->add_favourites;

	// Set up automated downloading of lists
	MAIN_EVENTS_Register(GSPYMASTER_UpdateLists,
		NULL, 1,60); 
		
	return 1;
}
