/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "resource.h"
#include <time.h>

#include "main.h"
#include "types.h"

//#include "hash.h"
#include "msg.h"
#include <ctype.h>

#define DEF_FIELD_SIZE	32
#define DEF_FIELD_EXTRA	16

/*char uc[256];

char* uppercase(char* str)
{
	int i;

	for (i=0;i<strlen(str);i++)
		uc[i] = toupper(str[i]);

	uc[i] = 0;
	return uc;

}*/

/* 
Write's query.log with info about what's happening internally.  
For debugging purposes only
*/
int consoleAllocated = 0;
int DEBUG_ShowMessages = 0;

void WRITE_DEBUG(char* text) {
	//FILE* fp;
	//time_t ltime;
	int written;
	char buf[4096];

#if 0
	if (strcmp(text,"date"))
	{
		fp = fopen("query.log","at");
		fprintf(fp,"%s\n",text);
	}
	else  {
		fp = fopen("query.log","wt");
		time(&ltime);
		fprintf(fp,"%s\n",ctime(&ltime));
	}
	fclose(fp);
#endif

	/*if (strlen(text)>6 && !memcmp(text,"memcmp",6))
		return;*/
	//return;

#if 0
#ifndef _DEBUG
	if (DEBUG_ShowMessages == 0)
		return; 
#endif
#else
	return;
#endif

	if (!consoleAllocated &&
			AllocConsole())
		consoleAllocated = 1;

	_snprintf(buf,4096,"%s\n",text);
	WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE),buf,strlen(buf),&written,NULL);

	//return;

	
}
int SQstricmp(char* str1, char* str2)
{
	return stricmp(str1,str2);
}

void SQstrtoupper(char* str)
{
	int i;
	for (i=0;i<(int)strlen(str);i++)
		str[i] = toupper(str[i]);
}

char* GetValue(server_t* svr, char* key) 
{
	char* str;
	if (!svr)
		return "";
	str = HashSelect(svr,"items", key,HASH_STRING);
	if (!str)
		str = "";
	return str;
	
}

#define HASH_MAXKEYS	16384

HASH** hashes;
int* numHashes;

int HashIsKeyNumeric(void* owner, char* name, void* key, char* operation)
{
	char buf[2048];

	if ((int)key< HASH_MAXKEYS)
		return 1;
	else if ((int)key >= HASH_MAXKEYS && (int)key <= HASH_MAXKEYS + 100)
	{
		sprintf(buf,"ERROR: Trying to access invalid key via %s.  Owner %x, name %s, key %d",
			operation,owner,name,(int)key);
		WRITE_DEBUG(buf);
		//MSG_DialogBox(NULL,NULL,buf,IDD_ERROR);
	}
	return 0;
}

/* Cache the indexes of recently used hashes */
#define HASH_CACHE_SIZE		16

int cache[HASH_CACHE_SIZE];
int cache_size = 0;

void HashInit(HASH* h[], int* nh)
{
	hashes = h;
	numHashes = nh;

}

/* Check whether the specified hash exists within the hashes array */
HASH* HashExists(void* owner, char* name) {
	int i;
	
	//uppercase(name);

	for (i=0;i<*numHashes;i++)
		if (owner == hashes[i]->owner && 
			!SQstricmp(hashes[i]->name, name))
				return hashes[i];
	return NULL;
}

int HASH_NO_EXIST = 0;

int HashCompare(const void* p1, const void* p2)
{
	HASH* h1 = *(HASH**) p1;
	HASH* h2 = *(HASH**) p2;

	if (!h1 && !h2)
		return 0;
	else if (!h1)
		return -1;
	else if (!h2)
		return 1;
	if (h1->owner < h2->owner)
		return -1;
	else if (h1->owner > h2->owner)
		return 1;
	else return strcmp(h1->name,h2->name);
}

void HashSort(void)
{
	qsort(hashes,*numHashes,sizeof(HASH*),HashCompare);
}


/* Grab the hash* with specified owner and name */
HASH* HashGetFromAll(void* owner, char* name)
{	
	int i, index = -1;
	HASH* hash = NULL;
	void* ptr;
	HASH temp;
	HASH* temp2;
#ifdef _DEBUG
	FILE* fp;
#endif

	temp.owner = owner;
	strcpy(temp.name,name);
	temp2 = &temp;

	ptr = bsearch(&temp2,hashes,*numHashes,sizeof(HASH*),HashCompare);
	if (ptr)
	{
		hash = *(HASH**)ptr;
		return hash;
	}
		


	// if we didn't find the hash, look through the entire collection
	for (i=0;i<*numHashes;i++)
		if (hashes[i]->owner == owner &&
			!SQstricmp(hashes[i]->name,name)) 
		{
			hash = hashes[i];
			/*fp = fopen("misses.log","at");
			fprintf(fp,"(%x) %s\n",hash->owner,hash->name);
			fclose(fp);*/
			index = i;
			break;
		}

	if (index != -1) {
		for (i=HASH_CACHE_SIZE-1;i>0;i--)
			cache[i] = cache[i-1];
		cache[0] = index;
		if (cache_size != HASH_CACHE_SIZE)
			cache_size++;
		return hash;
	}

#if 0
#ifdef _DEBUG
	fp = fopen("alloc.log","at");
	fprintf(fp,"(%x) %s\n",owner,name);
	fclose(fp);
#endif
#endif

	HASH_NO_EXIST = 1;
	return HashAlloc(owner,name,8);
}



/* Grab the hash* with specified owner and name */
HASH* HashGet(void* owner, char* name) {
	int i, index = -1;
	HASH* hash = NULL;
#ifdef _DEBUG
	FILE* fp;
#endif
	
	// look through the cache
	for (i=0;i<cache_size;i++)
		if (cache[i] < *numHashes &&
			hashes[cache[i]]->owner == owner &&
			!SQstricmp(hashes[cache[i]]->name,name)) 
		{
				hash = hashes[cache[i]];
				return hash;
		}

	return HashGetFromAll(owner,name);
}


static void* (*SQ_calloc)(size_t,size_t) = NULL;
static void* (*SQ_malloc)(size_t) = NULL;
static void* (*SQ_realloc)(void*,size_t) = NULL;

void* SQcalloc(int size, int width) {
	if (!SQ_calloc) {
		if (HashExists(HASH_GLOBAL,"funcs") &&
				HashExistsKey(HASH_GLOBAL,"funcs","calloc") != (HASH_PARAM)-1)
			SQ_calloc = HashSelectPtr(HASH_GLOBAL,"funcs","calloc");
		else SQ_calloc = calloc;
	}
	return SQ_calloc(size,width);
}
void* SQmalloc(int size) {
	if (!SQ_malloc) {
		if (HashExists(HASH_GLOBAL,"funcs") &&
				HashExistsKey(HASH_GLOBAL,"funcs","malloc") != (HASH_PARAM)-1)
			SQ_malloc = HashSelectPtr(HASH_GLOBAL,"funcs","malloc");
		else SQ_malloc = malloc;
	}
	return SQ_malloc(size);
}
void* SQrealloc(void* ptr, int size) {
	void* tmp;

	if (!SQ_realloc) {
		if (HashExists(HASH_GLOBAL,"funcs") &&
				HashExistsKey(HASH_GLOBAL,"funcs","realloc") != (HASH_PARAM)-1)
			SQ_realloc = HashSelectPtr(HASH_GLOBAL,"funcs","realloc");
		else SQ_realloc = realloc;
	}
	tmp = SQ_realloc(ptr,size);
	if (!tmp)
		tmp = SQmalloc(size);
		
	return tmp;
}
HASH* HashIncreaseSize(HASH* hash)
{
	int size, j;
#ifdef _DEBUG	 
	FILE* fp;

	fp = fopen("increase.log","at");
	fprintf(fp,"(%x) %s increased to %d\n",hash->owner,hash->name,hash->max*2);
	fclose(fp);
#endif _DEBUG	

	size = hash->max * 2;
	hash->entries = SQrealloc(hash->entries,size * sizeof(item_t*));
	for (j=hash->max;j<size;j++)
	{
		hash->entries[j] = SQcalloc(1,sizeof(item_t));
		hash->entries[j]->keylen = 
			hash->entries[j]->valuelen = DEF_FIELD_SIZE;
		hash->entries[j]->key = SQmalloc(DEF_FIELD_SIZE);
		hash->entries[j]->value = SQmalloc(DEF_FIELD_SIZE);
	}
	hash->max = size;
	return hash;
}



/* Allocate a new hash and add it to the hashes array */
HASH* HashAlloc(void* owner, char* name, int size) {
	int i, j; //, w;
	HASH* hash;

	/* If hash already exists, just return the currently
		existing one */
	if (HASH_NO_EXIST == 0 && HashExists(owner,name))
		return HashGet(owner,name);

	HASH_NO_EXIST = 1;

	j = *numHashes;

	if (j == MAX_HASHES) {
		//MSG_DialogBox(PROG_HINSTANCE,PROG_HWND,"Unable to allocate new hash.",IDD_ERROR);
		PostQuitMessage(0);
		return NULL;
	}

	// FIXME - check calloc doesn't fail	
	hashes[j] = SQcalloc(1,sizeof(HASH));
	
	/* Set hash's owner and name properties */
	hashes[j]->owner = owner;
	strcpy(hashes[j]->name, name);
	
	/* Set the hash's max key/value pairs, and allocate space for them */
	hashes[j]->max = size;
	if (( hashes[j]->entries = SQmalloc(size * sizeof(item_t*)) ) == NULL) {
		PostQuitMessage(0);
		return NULL;
	}
	
	//w = ((sizeof(item_t) + 15) /16) * 16;
	/*hashes[j]->entries[0] = calloc(size,sizeof(item_t)); 
	for (i=1;i<size;i++)
		hashes[j]->entries[i] = hashes[j]->entries[0] + i; */
	
	for (i=0;i<size;i++)
	{
		hashes[j]->entries[i] = SQmalloc(sizeof(item_t));
		hashes[j]->entries[i]->keylen = hashes[j]->entries[i]->valuelen = DEF_FIELD_SIZE;
		hashes[j]->entries[i]->key = SQmalloc(DEF_FIELD_SIZE);
		hashes[j]->entries[i]->value = SQmalloc(DEF_FIELD_SIZE);
	}

	(*numHashes)++;

	hash = hashes[j];
	HashSort();

	return hash;
}

/* Free the memory allocated to a specified hash, and removes it
   from the array of hash's */
int HashFree(void* owner, char* name) {
	int i, j;
	HASH* hash = HashGet(owner,name);

	/* Find the hash */
	for (i=0;i<*numHashes;i++)
		if (hashes[i] == hash)
			break;

	/* Take it out of the list of hashes */
	//for (j=i;j<(*numHashes)-1;j++)
	//		hashes[j] = hashes[j+1];
	memmove(hashes+i,hashes+i+1,sizeof(HASH*) * (*numHashes - i - 1));
	(*numHashes)--;

	/* Adjust the cached indexes */
	//cache_size = 0;
	for (j=0;j<cache_size;j++) {
		if (cache[j] > i)
			cache[j]--;
		else if (cache[j] == i) {
			memcpy(cache+j,cache+j+1,(cache_size - j - 1) * sizeof(int));
			cache_size--;
			j = -1;
		}
	}
	
	/* Free each allocated key/value pair */
	for (i=0;i<hash->max;i++)
	{
		free (hash->entries[i]->key);
		free (hash->entries[i]->value);
		free (hash->entries[i]);
	}
	//free (hash->entries[0]);

	/* Free the pointers to the pairs */
	free (hash->entries);
	/* Free the hash itself */
	free (hash);

	return TRUE;
}

/* Free's all associated hashes */
void HashFreeAssoc(void* owner) {
	int i;
	
	for (i=0;i<*numHashes;)
		if (hashes[i]->owner == owner)
			HashFree(hashes[i]->owner,hashes[i]->name);
		else i++;
}

/* Free's all allocated hashes */
void HashFreeAll(void) {

#ifdef _DEBUG
	FILE* fp;
	
	fp = fopen("unclaimed.log","wt");
	while (*numHashes > 0)
	{
		fprintf(fp,"Unclaimed hash: (%x) %s\n",hashes[0]->owner,hashes[0]->name);
		HashFree(hashes[0]->owner,hashes[0]->name);
	}
	fclose(fp);
#else
	/* While hashes exist, remove the first one.  The list collapses on
	itself, so hashes[0] is always valid whilst numHashes  != 0 */

	while (*numHashes > 0)
		HashFree(hashes[0]->owner,hashes[0]->name);
#endif

	cache_size = 0;
}

/* Copies an existing hash to a new hash */
HASH* HashCopy(void* owner, char* name, void* orig_owner, char* orig_name) {
	HASH* hash, *orig;
	int i;

	orig = HashGet(orig_owner, orig_name);
	hash = HashAlloc(owner, name, orig->max);
	
	for (i=0;i<orig->num;i++) {
		strcpy(hash->entries[i]->key,orig->entries[i]->key);
		strcpy(hash->entries[i]->value,orig->entries[i]->value);
	}
	hash->num = orig->num;
	
	return hash;
}

void HashDump(void)
{
	FILE* fp;
	int i, j;

	return;

	fp = fopen("hash.log","wt");

	fprintf(fp,"%d hashes\n\n",*numHashes);

	for (i=0;i<*numHashes;i++)
	{
		fprintf(fp,"\n%s (%d entries)\n--------\n",hashes[i]->name,hashes[i]->num);
		for (j=0;j<hashes[i]->num;j++)
			fprintf(fp,"%s\n",hashes[i]->entries[j]->key);
	}
	fclose(fp);
}
	
/* Inserts a value of type 'type' into the given hash, under key 'key' */
void* HashInsert(void* owner, char* name, void* key, void* value, char* type)
{
	int i; //,j, size;
	HASH *hash = HashGet(owner,name);
	char temp[256];
	
	if (!HashIsKeyNumeric(owner,name,key,"insert"))
	{
		for (i=0;i<hash->num;i++)
			if (!SQstricmp(hash->entries[i]->key, key))
				break;

		if (i >= hash->max) 
			HashIncreaseSize(hash);
			//MSG_DialogBox(PROG_HINSTANCE,PROG_HWND,"No space remaining",IDD_ERROR);
			//return NULL;
		//}

		if (hash->entries[i]->keylen <= (int) strlen(key))
		{
			hash->entries[i]->keylen = strlen(key) + DEF_FIELD_EXTRA;
			hash->entries[i]->key = SQrealloc(hash->entries[i]->key,
				hash->entries[i]->keylen);
		}
		strcpy(hash->entries[i]->key, key);

	} else {
		i = (int)key;

		if (i < 0)
			i = hash->num;

		while (i >= hash->max)
		{
			sprintf (temp,"No space left in hash '%s' of object '%x' (allocating key '%d')",name,owner,key);
			WRITE_DEBUG(temp);

//			MSG_DialogBox(PROG_HINSTANCE,PROG_HWND,temp,IDD_ERROR);
			

			//return "";
			HashIncreaseSize(hash);
		}

		// NOTE DEF_FIELD_SIZE Must be able to fit this
		sprintf(hash->entries[i]->key, "array_item%d",i);
	}

	// NOTE DEF_FIELD_SIZE Must be able to hold other types
	if (strcmp(type,"%s") ==  0 &&
		hash->entries[i]->valuelen <= (int) strlen(value))
	{
		hash->entries[i]->valuelen = strlen(value) + DEF_FIELD_EXTRA;
		hash->entries[i]->value = SQrealloc(hash->entries[i]->value,
			hash->entries[i]->valuelen);
	}

	sprintf(hash->entries[i]->value, type, value);

	if (i >= hash->num)
			hash->num = i + 1;

	HashDump();
	return value;
}

extern types_t* server_types[MAX_SERVER_TYPES];

/* Selects the specified key, returning a value of type 'type' */
void* HashSelectReason(void* owner, char* name, void* key, char* type, char* reason) {
	int i;
	HASH* hash = HashGet(owner,name);
	void* result;
	//char buf[128];

/*	if (server_types[0] && (server_types[0]->type < 0 || server_types[0]->type > 100))
	{
		WRITE_DEBUG("ERROR!");
	}*/

	if (!HashIsKeyNumeric(owner,name,key,reason)) {
		for (i=0;i<hash->num;i++)
			if (!SQstricmp(hash->entries[i]->key,key)) 
				break;
			
	}
		else i = (int)key;

	if (i >= hash->num) {
		if (!strcmp(type,"%s"))
			return "";
		return NULL;
	}

	if (!strcmp(type, "%s"))
		return hash->entries[i]->value;

	result = NULL;
	sscanf(hash->entries[i]->value,type,&result);

	/*if (key > HASH_MAXKEYS)
		sprintf(buf,"'%p'->'%s'->'%s' = '%p'",
			owner,name,key,result);
	else sprintf(buf,"'%p'->'%s'->'%d' = '%p'",
			owner,name,(int)key,result);*/
	//WRITE_DEBUG(buf);
	return result;
}

void* HashSelect(void* owner, char* name, void* key, char* type) {
	return HashSelectReason(owner,name,key,type,"select");
}


/* Deletes the specified key from the hash */
int HashDelete(void* owner, char* name, void* key) {
	int i, j;
	item_t* temp;
	HASH* hash = HashGet(owner,name);
	
	if (!HashIsKeyNumeric(owner,name,key,"delete")) {
		for (i=0;i<hash->num;i++)
			if (!SQstricmp(hash->entries[i]->key,key))
				break;
	}
		else i = (int) key;
	
	if (i >= hash->num)
		return 0;

	/* just move the selected key/value pair to the end of the array */
	temp = hash->entries[i];
	for (j=i;j<hash->num-1;j++)
		hash->entries[j] = hash->entries[j+1];
	hash->entries[j] = temp;

	hash->num--;

	return 1;
}

/* Deletes the specified value from the hash */
int HashDeleteValue(void* owner, char* name, void* value, char* type) {
	char* temp;
	char str[128];
	int i = 0;
	sprintf(str,type,value);

	for (temp = HashReset(owner,name,HASH_STRING);temp;temp = HashNext(owner,name,HASH_STRING)) {
		if (!strcmp(temp,str)) {
			HashDelete(owner,name,(HASH_PARAM)i);
			return 1;
		}
		i++;
	}
	return 0;
}

/* Clears all key/value pairs from the hash */
int HashDeleteAll(void* owner, char* name)  {
	HASH* hash = HashGet(owner,name);
	hash->num = 0;
	return 1;
}

/* Returns the current # of key/value pairs in the hash */
int HashCount(void* owner, char* name)
{
	HASH* hash = HashGet(owner, name);
	return hash->num;
}

/* Moves a (numeric) key/value pair to another location in the hash */
int HashMove(void* owner, char* name, int from, int to)
{
	HASH* hash = HashGet(owner, name);
	int i;
	item_t* temp;

	if (from >= hash->num || to >= hash->num)
		return 0;

	temp = hash->entries[from];
	for (i=from;i<hash->num - 1;i++)
		hash->entries[i] = hash->entries[i+1];

	for (i=hash->num-1;i>to;i--)
		hash->entries[i] = hash->entries[i-1];

	hash->entries[to] = temp;
	return 1;
}

/* Returns the next value from the hash */
void* HashNext(void* owner, char* name, char* type)
{
	void* p;
	HASH* hash = HashGet(owner,name);
	
	p = HashSelectReason(owner,name,(HASH_PARAM)(hash->array),HASH_PTR,"next+select");
	if (!p)
		return NULL;

	return HashSelectReason(owner,name,(HASH_PARAM)(hash->array ++), type,"next+select");
}

/* Returns the first value from the hash */
void* HashReset(void* owner, char* name, char* type)
{
	HASH* hash = HashGet(owner, name);
	
	hash->array = 0;
	return HashNext(owner, name, type);
}

/* Assuming each value within the hash is a pointer, this function
   searches the memory pointed at for a specified string; returns
   the first matching pointer with contents as specified. */
void* HashFind(void* owner, char* name, char* str, int offset) {
	char* ptr;

	for (ptr = HashReset(owner,name,HASH_PTR);ptr;ptr = HashNext(owner,name,HASH_PTR))
	{
		if (SQstricmp(ptr + offset, str) == 0)
			return ptr;
	}
	return NULL;
}

void* HashExistsKey(void* owner, char* name, void* key) {
	int i;
	HASH* hash = HashGet(owner,name);

	if (HashIsKeyNumeric(owner,name,key,"exists")) {
		if ((int)key < hash->num)
			return hash->entries[(int)key]->key;
		return (HASH_PARAM) -1;
	}
	
	for (i=0;i<hash->num;i++)
		if (!SQstricmp(hash->entries[i]->key,key))
			return (HASH_PARAM)i;
	return (HASH_PARAM)-1;
}

void* HashExistsValue(void* owner, char* name, void* value, char* type) {
	int i;
	HASH* hash = HashGet(owner,name);
	char temp[128];

	sprintf(temp, type, value);
	
	for (i=0;i<hash->num;i++)
		if (!SQstricmp(hash->entries[i]->value,temp))
			return (HASH_PARAM)i;
	
	return (HASH_PARAM)-1;
}

int HashInsertHash(void* owner, char* name, void* other_owner, char* other_name)
{
	HASH* hash = HashGet(other_owner,other_name);
	int i;
	char buf[256];

	for (i=0;i<hash->num;i++) 
	{
		sprintf(buf,"(%d) Inserting into '%x'->'%s' value (%x) from '%x'->'%s'",
				i,owner,name,hash->entries[i]->value,other_owner,other_name);
		WRITE_DEBUG(buf);
		HashInsert(owner,name,HASH_ANY,hash->entries[i]->value,HASH_STRING);
	}
	WRITE_DEBUG("HashInsertHash completed");
	return 1;
}

int HashExistsSubset(void* owner, char* name, void* other_owner, char* other_name)
{
	char* temp, *full;
	
	temp = HashReset(other_owner, other_name, HASH_STRING);
	while (temp)
	{
		full = HashReset(other_owner, other_name, HASH_STRING);
		while (full)
		{
			if (!SQstricmp(temp,full))
				break;
			full = HashNext(owner,name,HASH_STRING);
		}
		if (!full)
			return 0;
			
		temp = HashNext(other_owner,other_name, HASH_STRING);
	}
	return 1;
}

