/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <winsock.h>
#include <time.h>
#include <stdio.h>
#include "resource.h"

#include "http.h"
#include "resolve.h"
#include "hash.h"
#include "main.h"
#include "comm.h"
#include "msg.h"

void HTTP_CloseConnection(HTTP_DATA* data)
{
	closesocket(data->s);
	if (data->hwnd) {
		HashFree(data->hwnd,"Global");
		DestroyWindow(data->hwnd);
	}
	
	HashDeleteValue(HASH_GLOBAL,"HTTP_DATA",data, HASH_PTR);
}


int WINAPI HTTP_DlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	HTTP_DATA* data;
	if (msg == WM_COMMAND && wparam == IDB_ABORT) {
		data = HashSelectPtr(hwnd,"Global","HTTP_DATA");
		HTTP_CloseConnection(data);
		free (data);
	}
	return 0;
}

char HTTP_hostname[1024], HTTP_path[1024];

void HTTP_URLSplit(char* str)
{
	int i;
	// Look for end of 'http://'

	for (i=0;i<(int) strlen(str)-1;i++)
		if (str[i] == '/' && str[i+1] == '/')
			break;

	if (i == (int) strlen(str) - 1)
		i = 0;
	else i+= 2;

	strcpy(HTTP_hostname, str + i);
	
	for (i=0;i<(int) strlen(HTTP_hostname);i++)
	{
		if (HTTP_hostname[i] == '/')
		{
			strcpy(HTTP_path,HTTP_hostname+i);
			HTTP_hostname[i] = 0;
			break;
		}
	}
	return;
}

extern int allowNSLookups;

void HTTP_Download(HTTP_DATA* data)
{
	SOCKADDR_IN addr;
	int len, i;
	char str_addr[512];
	char buf[1024];
	
	data->s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
	if (data->s < 0) {
		HTTP_CloseConnection(data);
		free(data);
		return;
	}

	if (data->windowed) {
		data->hwnd = CreateDialog(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_HTTP),
			NULL,HTTP_DlgProc);
		HashInsertPtr(data->hwnd,"Global","HTTP_DATA",data);
		ShowWindow(data->hwnd,SW_SHOW);
	} else 
		data->hwnd = NULL;

	len = 1;
	ioctlsocket(data->s,FIONBIO,&len); // don't block on connect attempt

	addr.sin_family = AF_INET;
	// FIXME

	for (i=0;i<(int) strlen(data->url)-1;i++)
		if (data->url[i] == '/' && data->url[i+1] == '/')
			break;

	if (i == (int) strlen(data->url) - 1)
		i = 0;
	else i+= 2;

	strcpy(str_addr, data->url + i);
	for (i=0;i<(int) strlen(str_addr);i++)
		if (str_addr[i] == '/')
		{
			str_addr[i] = 0;
			break;
		}

	if (GetServerPort(str_addr,80) == 80 && strcmp(HSTG("Settings","HttpProxy"),"") != 0)
		strcpy(str_addr,HSTG("Settings","HttpProxy") );
	
	WRITE_DEBUG("HTTP: resolve");
	if ((addr.sin_addr.s_addr = ResolveName(str_addr)) == 0) {
		HTTP_CloseConnection(data);
		
		if (data->windowed)
		{
			sprintf(buf,"Unable to resolve '%s'; DNS failed.",str_addr);
			if (!allowNSLookups)
				strcat(buf," (DNS lookups are currently disabled)");
			MSG_DialogBox(NULL,NULL,buf,IDD_ERROR);
		}		
		free(data);
		return;
	}
	
	addr.sin_port = htons(GetServerPort(str_addr,80));

	WRITE_DEBUG("HTTP: connect");
	connect(data->s,(SOCKADDR*)&addr,sizeof(SOCKADDR_IN));
	if (WSAGetLastError() != WSAEWOULDBLOCK) {
		HTTP_CloseConnection(data);
		free(data);
		return;
	}

	//WRITE_DEBUG("HTTP: finish");
	data->inittime = time(NULL);
	HashInsert(HASH_GLOBAL,"HTTP_DATA",HASH_ANY,data,HASH_PTR);	
	return;
}
	

void HTTP_CheckConnect(HTTP_DATA* data)
{
	FD_SET set;
	TIMEVAL tm;

	char fullurl[512];
	int  result;
	
	FD_ZERO(&set);
	FD_SET(data->s,&set);
	tm.tv_sec = 0;
	tm.tv_usec = 0;

	if (select(0,NULL,&set,NULL,&tm) <= 0)
	{
		return;
	}
	
	/* Connection established, so do some work */
	data->connected = 1;

	HTTP_URLSplit(data->url);
	//sprintf(fullurl,"GET %s HTTP/1.1\r\nHost: %s\r\n\r\n",HTTP_path,HTTP_hostname);
	sprintf(fullurl,"GET %s HTTP/1.1\r\nHost: %s\r\nCache-control: no-cache\r\nConnection: close\r\n\r\n",data->url,HTTP_hostname);

	if (data->custom)
		strcpy(fullurl, data->custom);
	
	do {
		result = send(data->s,fullurl,strlen(fullurl),0);
	} while (result == -1 && WSAGetLastError() == WSAEWOULDBLOCK);

	data->length = 0;	
	return;
}

int HTTP_WaitResponse(HTTP_DATA* data) {
	FD_SET set;
	TIMEVAL tm;

	char temp[8192];
	char* str;
	int t, result;

	FD_ZERO(&set);
	FD_SET(data->s,&set);
	tm.tv_sec = 0;
	tm.tv_usec = 0;

	result = select(0,&set,NULL,NULL,&tm);

	if (result == 0)
		return 1; 
	else if (result < 0) {
		HTTP_CloseConnection(data);
		if (data->header)
			free(data->header);
		free(data);
		return 0;
	}
		
	t = recv(data->s,temp,4096,0);
	if (t > 0) {
		if (data->data) {
			str = malloc(data->length + t + 16);
			memcpy(str,data->data,data->length);
			memcpy(str+data->length,temp,t);
			free(data->data);
			data->header = data->data = str;
			data->length += t;
		}
		else {
			data->header = data->data = malloc(t+16);
			memcpy(data->data,temp,t);
			data->length = t;
		}
		if (data->hwnd)
		{
			sprintf(temp,"Server Query is downloading a server list.\n\n%d",data->length);
			SetDlgItemText(data->hwnd,IDT_WAIT,temp);
		}
		//send(data->s,"\n",1,0);
		/*if (data->length < 4096) */
			return 1;
	}
	
	/* completed */
	HTTP_CloseConnection(data);

	if (data->length == 0) {
		if (data->data)
			free(data->data);
		free(data);
		return 0;
	}

	data->data[data->length] = 0;
	data->header = data->data;
	
	if (!data->custom) {
		// push past the http headers
		str = data->data;
		WRITE_DEBUG("memcmp() http.c(181)");
		while (str < data->data + data->length && memcmp(str,"\r\n\r\n",4))
			str++;
		str+=4;
		data->data = str;
		data->length -= (data->data - data->header);
	}
	WRITE_DEBUG("func setup http.c(185)");
	if (data->func)
		data->func(data);
	
	WRITE_DEBUG("free http.c(181)");
	free(data->header);
	free(data);
	return 0;
}

void HTTP_Free(void) {
	HTTP_DATA* data;

	while ((data = HashReset(HASH_GLOBAL,"HTTP_DATA",HASH_PTR)) != NULL)
	{
		closesocket(data->s);
		free(data);
		HashDeleteValue(HASH_GLOBAL,"HTTP_DATA",data,HASH_PTR);
	}
}

void HTTP_Update(void)
{
	HTTP_DATA* data;

	for (data = HashReset(HASH_GLOBAL,"HTTP_DATA",HASH_PTR);data;)
	{
		if (!data->connected)
			HTTP_CheckConnect(data);
		else if (!HTTP_WaitResponse(data))
		{ // dropped an item, so restart list
			data = HashReset(HASH_GLOBAL,"HTTP_DATA",HASH_PTR);
			continue;
		}
		data = HashNext(HASH_GLOBAL,"HTTP_DATA",HASH_PTR);
	}	
}