/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "resource.h"

#include "types.h"
#include "list.h"
#include "main.h"
#include "gencfg.h"
#include "msg.h"
#include "ui.h"
#include "limits.h"
#include "events.h"

int CALLBACK ListCompare(LPARAM item1, LPARAM item2, LPARAM p)
{
	list_t* list, *list1, *list2, *parent;
	server_t* svr, *svr1, *svr2;

	int i;
	HASH* hash;
	ITEM_TYPE* it1, *it2;

	HWND hwnd = HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR);
	parent = (list_t*)p;

	it1 = (ITEM_TYPE*)item1;
	it2 = (ITEM_TYPE*)item2;

	if (it1->type == IT_LIST && it2->type != IT_LIST)
		return -1;
	if (it2->type == IT_LIST && it1->type != IT_LIST)
		return 1;

	if (it1->type == IT_LIST) {
		list1 = it1->ptr;
		list2 = it2->ptr;

	
		hash = HashGet(HASH_GLOBAL,"Lists");
		for (i=0; i<HashCount(HASH_GLOBAL,"Lists");i++)
		{
			list = HashSelectPtr(HASH_GLOBAL,"Lists",(HASH_PARAM)i);
			if (list == list1)
				return -1;
			if (list == list2)
				return 1;
		}
	}
	else {
		svr1 = it1->ptr;
		svr2 = it2->ptr;

		hash = HashGet(parent,"servers");

		for (i=0;i<HashCount(parent,"servers");i++)
		{
			svr = HashSelectPtr(parent,"servers",(HASH_PARAM)i);

			if (svr == svr1)
				return -1;
			if (svr == svr2)
				return 1;
		}
	}
	return 0;
}

void ListSort(list_t* list) {
	TV_SORTCB sort;
	
	if (! list )
		sort.hParent = HashSelect(HASH_GLOBAL,"Global","ServersTreeItem",HASH_PTR); 
	else sort.hParent = list->hTreeItem;
	sort.lpfnCompare = ListCompare;
	sort.lParam = (LPARAM) list;
	TreeView_SortChildrenCB(PROG_TREEVIEW,&sort,0);
	return;
}
group_t* GroupNamed(char* name) {
	return HashFind(HASH_GLOBAL,"Groups",name,(int)((group_t*)NULL)->name);
}

group_t* GroupOfList(list_t* list) {
	return &(list->group);
}

void DeleteList(list_t* list) 
{
	void* ptr;
	char buf[128];

	// Cancel any running ping of list
	EVENTS_Cancel(HashSelectPtr(list,"pinglist","event"));
	HashDelete(list,"pinglist","event");

	//WRITE_DEBUG("Removing servers from list");
	/* Remove the servers from the list */
	DeleteServerList(list);
	
	sprintf(buf,"Removing trevview item (%p)",list);
	//WRITE_DEBUG(buf);
	sprintf(buf,"desc (desc, %s - item, %p - treeview %p)",list->desc,list->hTreeItem,
		HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR));
	WRITE_DEBUG(buf);

	/* Delete the list item */
	if (list->hTreeItem != TVI_ROOT && list->hTreeItem) {
	//	WRITE_DEBUG("Deleting item");
		 TreeView_DeleteItem((HWND) HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),
			(HTREEITEM) list->hTreeItem);
		 //WRITE_DEBUG("DONE");
	}


	WRITE_DEBUG("Removing list from lists hash");
	/* Remove this list from the hash of lists */
	HashDeleteValue(HASH_GLOBAL,"Lists",list,HASH_PTR);


	//WRITE_DEBUG("Free'ing item_type");
	/* Free the associated item_type */
	if ((ptr = HashSelectPtr(list,"ptrs","item_type")) != NULL)
		free(ptr);

	//WRITE_DEBUG("Deleting other associated hashes");
	/* Delete associated hashes */
	HashFreeAssoc(list);

	//WRITE_DEBUG("Free'ing list itself");
	free(list);

	return;
}

list_t* ListNamedWithParent(char* name, list_t* parent) {
	list_t* list;

	for (list = HashReset(HASH_GLOBAL,"Lists",HASH_PTR);list != NULL;list = HashNext(HASH_GLOBAL,"Lists",HASH_PTR))
		if (!strcmp(list->desc,name) && HashSelectPtr(list,"ptrs","parent") == parent)
			return list;
	return NULL;
}

// ***
// This performs two things:
// a) Create a item in the server tree-view
//    to list relevant servers under

// b) Set up a struct in the 'Lists' hash
//    containing relevant details about this 
//    list
// ***
list_t* CreateList(char* desc, char* group, list_t* parent)
{
	list_t* list;
	ITEM_TYPE* it;
	TV_INSERTSTRUCT tv;
	HWND hwnd = HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR);
	HTREEITEM hTreeItem;
	TV_ITEM item;

	if (stricmp(group,"LANServers") == 0 && !HSIG("Settings","ShowLANList"))
		return NULL;

	if (!parent && (stricmp(group,"LANServers") == 0 || stricmp(group,"ListAllServers") == 0) )
	{
		for (list = HashReset(NULL,"Lists",HASH_PTR);list;
			list = HashNext(NULL,"Lists",HASH_PTR) )
				if (stricmp(GroupOfList(list)->name,group) == 0) {
					DeleteList(list);
					break;
				}
	}
	if (desc && ListNamedWithParent(desc,parent))
	{
		HashDeleteAll(ListNamedWithParent(desc,parent),"Servers");
		return ListNamedWithParent(desc,parent);
	}

	
	list = (list_t*) calloc(sizeof(list_t),1);
	it = (ITEM_TYPE*) calloc(sizeof(ITEM_TYPE),1);

	it->type = IT_LIST;
	it->ptr = list;

	HashInsertPtr(list,"ptrs","item_type",it);
	
	if (desc)
	{
		if (!parent)
			tv.hParent = HashSelect(HASH_GLOBAL,"Global","ServersTreeItem",HASH_PTR); //NULL;
		else {
			tv.hParent = parent->hTreeItem;

			// Mark as having children
			item.hItem = list->hTreeItem;
			item.mask = TVIF_CHILDREN;
			item.cChildren = 1;
			TreeView_SetItem(PROG_TREEVIEW,&item);
		}

		tv.hInsertAfter = TVI_LAST;
		if (stricmp(group,"LANServers") == 0)
			tv.hInsertAfter = TVI_FIRST;

		tv.item.pszText = desc;
		tv.item.iImage = tv.item.iSelectedImage = 
			HashSelectInt(HASH_GLOBAL,"Indexes","Closed");
		tv.item.mask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
		tv.item.state = tv.item.stateMask = TVIS_EXPANDED;
		tv.item.lParam = (LPARAM)it;	
	
		if (hwnd)
		{
			hTreeItem = TreeView_InsertItem(hwnd,&tv);
			TreeView_Expand(hwnd,hTreeItem,TVE_EXPAND);
		}
	
	
	//FIXME
	//TreeView_EnsureVisible(hwnd,&tv.item);
		strcpy(list->desc,desc);
		list->hTreeItem = hTreeItem;
	}

	if (GroupNamed(group))
		memcpy(&list->group,GroupNamed(group),sizeof(group_t));
	else memcpy(&list->group,GroupNamed("NetServers"),sizeof(group_t));

	if (HashExists(GroupOfList(list)->location,"Extra"))
		HashCopy(list,"Extra", GroupOfList(list)->location, "Extra");
	
	
	HashAlloc(list,"Servers",128);

	if (parent)
		HashInsertPtr(list,"ptrs","parent",parent);

	HashInsert(HASH_GLOBAL,"Lists",HASH_ANY,list,HASH_PTR);

	return list;
}






int WINAPI CREATELIST_DlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	group_t* grp;
	char temp[128];
	int i;
	
	switch (msg) {
	case WM_INITDIALOG:
		for (grp = HashReset(HASH_GLOBAL,"Groups",HASH_PTR);grp;grp = HashNext(HASH_GLOBAL,"Groups",HASH_PTR))
			if (grp->add_server || HashExistsKey(grp,"extra","URL") != (HASH_PARAM)-1)
				SendMessage(GetDlgItem(hwnd,IDC_GROUP),CB_ADDSTRING,0,(LPARAM)grp->desc);
		SendMessage(GetDlgItem(hwnd,IDC_GROUP),CB_SETCURSEL,0,0);
		SendMessage(GetDlgItem(hwnd,IDC_GROUP),CB_SELECTSTRING,0,(LPARAM)GroupNamed("NetServers")->desc);

		for (i=0;i<MAX_SERVER_TYPES;i++)
			if (!GetServerType(i)->invalid)
				SendMessage(GetDlgItem(hwnd,IDC_SERVERTYPE),CB_ADDSTRING,0,(LPARAM)GetServerType(i)->name);
		break;
	case WM_COMMAND:
		if (LOWORD(wparam) == IDC_GROUP && HIWORD(wparam) == CBN_SELCHANGE)
		{
			GetDlgItemText(hwnd,IDC_GROUP,temp,128);
			grp = HashFind(HASH_GLOBAL,"Groups",temp,(int)((group_t*)NULL)->desc);
			if ((int)HashExistsKey(grp,"extra","URL") != -1) {
				
				SendMessage(GetDlgItem(hwnd,IDT_URL),CB_RESETCONTENT,0,0);
				if (!strcmp(grp->name,"QStatList"))
				{
					SendMessage(GetDlgItem(hwnd,IDT_URL),CB_ADDSTRING,0,(LPARAM)"http://serverquery.qgl.org/powerup/");
					SendMessage(GetDlgItem(hwnd,IDT_URL),CB_ADDSTRING,0,(LPARAM)"http://serverquery.qgl.org/wireplay/");
					SendMessage(GetDlgItem(hwnd,IDT_URL),CB_ADDSTRING,0,(LPARAM)"http://serverquery.qgl.org/optus/");
					SendMessage(GetDlgItem(hwnd,IDT_URL),CB_ADDSTRING,0,(LPARAM)"http://serverquery.qgl.org/eisa/");
				}

				SetDlgItemText(hwnd,IDT_URL,HashSelect(grp,"extra","URL",HASH_STRING));
				EnableWindow(GetDlgItem(hwnd,IDT_URL),TRUE);

				/*else if (!strcmp(grp->name,"GSpyList"))
					SendMessage(GetDlgItem(hwnd,IDT_URL),CB_ADDSTRING,0,(LPARAM)"http://serverquery.qgl.org/wireplay/");*/
		
			} else {
				SetDlgItemText(hwnd,IDT_URL,"");
				EnableWindow(GetDlgItem(hwnd,IDT_URL),FALSE);
			}

			if ((int)HashExistsKey(grp,"extra","Grouped") != -1) {
				SendMessage(GetDlgItem(hwnd, IDC_GROUPED), BM_SETCHECK, HashSelectInt(grp,"extra","Grouped") ? BST_CHECKED : 0, 0);
				EnableWindow(GetDlgItem(hwnd,IDC_GROUPED),TRUE);
			} else {
				SendMessage(GetDlgItem(hwnd, IDC_GROUPED), BM_SETCHECK, 0, 0);
				EnableWindow(GetDlgItem(hwnd,IDC_GROUPED),FALSE);
			}
			
			if ((int)HashExistsKey(grp,"extra","Type") != -1) {
				SendMessage(GetDlgItem(hwnd,IDC_SERVERTYPE),CB_SELECTSTRING,
					0,(LPARAM)GetServerType((int)HashSelect(grp,"extra","Type",HASH_INT))->name);
				//SendMessage(GetDlgItem(hwnd,IDC_SERVERTYPE),WM_ENABLE,1,0);
				EnableWindow(GetDlgItem(hwnd,IDC_SERVERTYPE),TRUE);
			} else {
				SendMessage(GetDlgItem(hwnd,IDC_SERVERTYPE),CB_SETCURSEL,-1,0);
				
				//SetDlgItemText(hwnd,IDC_SERVERTYPE,"");
				EnableWindow(GetDlgItem(hwnd,IDC_SERVERTYPE),FALSE);
				
			}
			break;
		}
		if (wparam == IDOK) {
			GetDlgItemText(hwnd,IDT_NAME,temp,128);
			if (! temp[0] || ListNamedWithParent(temp,CURRENT_LIST))
			{
				MSG_DialogBox(NULL,NULL,"Unable to create new list.\n\nPlease enter a name to uniquely identify this server list. The name can not be blank or the same as that of another list.",IDD_ERROR);
				return 0;
			}
			HashInsert(HASH_GLOBAL,"Status","ListName",temp,HASH_STRING);
			GetDlgItemText(hwnd,IDC_GROUP,temp,128);
			grp = HashFind(HASH_GLOBAL,"Groups",temp,(int) ((group_t*)NULL)->desc);
			if (grp)
				HashInsert(HASH_GLOBAL,"Status","GroupName",grp->name,HASH_STRING);
			GetDlgItemText(hwnd,IDT_URL,temp,128);
			HashInsert(HASH_GLOBAL,"Status","URL",temp,HASH_STRING);

			if (SendMessage(GetDlgItem(hwnd, IDC_GROUPED), BM_GETCHECK, 0, 0) == BST_CHECKED)
				HashInsertInt(HASH_GLOBAL,"Status","Grouped",1);
			else HashInsertInt(HASH_GLOBAL,"status","Grouped",0);

			//
			// Types no longer supported
			//

			/* GetDlgItemText(hwnd,IDC_SERVERTYPE,temp,128);
			for (i=0;i<MAX_SERVER_TYPES;i++)
				if (strcmp(GetServerType(i)->name,temp) == 0)
					HashInsert(HASH_GLOBAL,"Status","Type",(HASH_PARAM)GetServerType(i)->type,HASH_INT);
					*/
			EndDialog(hwnd,1);
		}
		else if (wparam == IDCANCEL) {
			EndDialog(hwnd,0);
		}
		return 0;
	case WM_CLOSE:
		EndDialog(hwnd,0);
	}
	return 0;
}

int CREATELIST_DialogBox(HINSTANCE hinstance, HWND hwnd)
{
	list_t* list;
	group_t* group;
	int root;

	HashDelete(HASH_GLOBAL,"Status","ListName");
	HashDelete(HASH_GLOBAL,"Status","GroupName");
	
	if (DialogBox(NULL,MAKEINTRESOURCE(IDD_CREATELIST),hwnd,CREATELIST_DlgProc)) {
		if ((int) HashExistsKey(HASH_GLOBAL,"Status","ListName") == -1)
			return 0;
		if (GroupNamed(HashSelect(HASH_GLOBAL,"Status","GroupName",HASH_STRING)) == NULL)
			return 0;
		root = HSI(CURRENT_LIST,"extra","root");

		list = CreateList(HashSelect(HASH_GLOBAL,"Status","ListName",HASH_STRING),
			HashSelect(HASH_GLOBAL,"Status","GroupName",HASH_STRING),
			root ? NULL : HashSelect(HASH_GLOBAL,"Status","CurrentList",HASH_PTR));
		
		group = GroupNamed(list->group.name);

		if (HashExists(group,"extra"))
		{
			//if ((int)HashExistsKey(group,"extra", "Type") != -1) 
				//HashInsert(list,"extra","Type",HashSelect(HASH_GLOBAL,"Status","Type",HASH_STRING),HASH_STRING);

			if ((int)HashExistsKey(group,"extra", "URL") != -1) {
				HashInsert(list,"extra","URL",HashSelect(HASH_GLOBAL,"Status","URL",HASH_STRING),HASH_STRING);
				if (GroupOfList(list)->refresh_list)
					GroupOfList(list)->refresh_list(list);
			}

			if ((int)HashExistsKey(group,"extra", "Grouped") != -1) 
				HashInsert(list,"extra","Grouped",HashSelect(HASH_GLOBAL,"Status","Grouped",HASH_INT),HASH_INT);
		}
		return 1;
	}
	return 0;
}

void RENAMELIST_Modify(list_t* list, char* desc)
{
	TV_INSERTSTRUCT tv;
	ITEM_TYPE* it = HashSelectPtr(list,"ptrs","item_type");

	HWND hwnd = HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR);

	tv.hParent = HashSelect(HASH_GLOBAL,"Global","ServersTreeItem",HASH_PTR);
	//tv.hInsertAfter = TVI_SORT;
	tv.hInsertAfter = NULL;
	tv.item.pszText = desc;
	tv.item.lParam = (LPARAM) HashSelectPtr(list,"ptrs","item_type");
	tv.item.mask = TVIF_TEXT | TVIF_PARAM;
	tv.item.hItem = list->hTreeItem;
	//list->hTreeItem = TreeView_InsertItem(hwnd,&tv);

	TreeView_SetItem(hwnd,&(tv.item));
/*	TreeView_SortChildren(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),
		HashSelect(HASH_GLOBAL,"Global","ServersTreeItem",HASH_PTR),0);*/

	strcpy(list->desc, desc);
}



int WINAPI RENAMELIST_DlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	char listname[128];
	list_t* list;

	switch (msg) {
	case WM_INITDIALOG:
		SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)"List Properties");
		list = (list_t*)lparam;
		SetProp(hwnd,"List",list);
		SetDlgItemText(hwnd,IDT_VALUE,list->desc);
		if ((int)HashExistsKey(list,"extra","URL") != -1)
			SetDlgItemText(hwnd,IDT_URL,HashSelect(list,"extra","URL",HASH_STRING));

		if ((int)HashExistsKey(list,"extra","Grouped") != -1) {
			EnableWindow(GetDlgItem(hwnd,IDC_GROUPED),TRUE);
			SendMessage(GetDlgItem(hwnd,IDC_GROUPED),BM_SETCHECK,
				HashSelectInt(list,"extra","Grouped"), 0);
		}

		//
		// Types no longer supported
		//
		/*
		if ((int)HashExistsKey(list,"extra","Type") != -1) {
			for (i=0;i<MAX_SERVER_TYPES;i++)
				if (GetServerType(i)->type)
					SendMessage(GetDlgItem(hwnd,IDC_SERVERTYPE),CB_ADDSTRING,0,(LPARAM)GetServerType(i)->name);
			SendMessage(GetDlgItem(hwnd,IDC_SERVERTYPE),CB_SELECTSTRING,0,(LPARAM)GetServerType((int) HashSelect(list,"extra","Type",HASH_INT))->name);
		}*/

		break;
	case WM_COMMAND:
		if (wparam == IDOK) {
			list = GetProp(hwnd,"List");
			GetDlgItemText(hwnd,IDT_VALUE,listname,128);
			RENAMELIST_Modify(list,listname);
			if ((int)HashExistsKey(list,"extra","URL") != -1)
			{
				GetDlgItemText(hwnd,IDT_URL,listname,128);
				HashInsert(list,"extra","URL",listname,HASH_STRING);
			}
			if ((int)HashExistsKey(list,"extra","Grouped") != -1)
				HashInsertInt(list,"extra","Grouped",
					SendMessage(GetDlgItem(hwnd,IDC_GROUPED),BM_GETCHECK,0,0));

			//
			//Types no longer supported
			//
			/*
			if ((int)HashExistsKey(list,"extra", "Type") != -1)
			{
				GetDlgItemText(hwnd,IDC_SERVERTYPE,listname,128);
				for (i=0;i<MAX_SERVER_TYPES;i++)
					if (strcmp(GetServerType(i)->name,listname) == 0)
						HashInsert(list,"extra","Type",(HASH_PARAM)GetServerType(i)->type,HASH_INT);
			}
			*/
			RemoveProp(hwnd,"List");
			EndDialog(hwnd,1);
		}
		else if (wparam == IDCANCEL) {
			RemoveProp(hwnd,"List");
			EndDialog(hwnd,0);
		}
		return 0;
	case WM_CLOSE:
		RemoveProp(hwnd,"List");
		EndDialog(hwnd,0);
	}
	return 0;
}

int WINAPI RENAMELIST_DialogBox(HINSTANCE hinstance, HWND hwnd, list_t* list, int window)
{
	return DialogBoxParam(NULL,MAKEINTRESOURCE(window),hwnd,RENAMELIST_DlgProc,(LPARAM)list);
}

void RenameList(list_t* list) {
	int window;

	if ((int) HashExistsKey(list,"extra","Type") != -1)
		window = IDD_GSPYLIST;
	else if ((int) HashExistsKey(list,"extra","URL") != -1)
		window = IDD_QSTATLIST;
	else window = IDD_EDIT;

	RENAMELIST_DialogBox(PROG_HINSTANCE,PROG_HWND,list,window);
}

void StateImageList(char* name)
{
	HIMAGELIST img;
	HBITMAP bmp;
//	int i, index,rs;

	img = ImageList_Create(16,16, ILC_COLOR24|ILC_MASK,3,4); 
	

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_NONE));
	HashInsertInt(NULL,"StateIndexes","Blank",ImageList_AddMasked(img,bmp,RGB(255,255,255)) );
	DeleteObject(bmp);

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_BUDDY));
	HashInsertInt(NULL,"StateIndexes","Buddy",ImageList_AddMasked(img,bmp,RGB(255,255,255)) );
	DeleteObject(bmp);

	HashInsertPtr(NULL,"Global",name,img);
 
	return;
}
void TreeviewImageList(char* name) {
	HIMAGELIST img;
	HBITMAP bmp;
	int i, index,rs;

	img = ImageList_Create(16,16, ILC_COLOR24|ILC_MASK,3,16); 
	

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_BLANK));
	HashInsertInt(NULL,"Indexes","Blank",ImageList_AddMasked(img,bmp,RGB(255,255,255)) );
	DeleteObject(bmp);

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_OPEN));
	HashInsertInt(NULL,"Indexes","Open",ImageList_AddMasked(img,bmp,RGB(255,255,255)) );
	DeleteObject(bmp);

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_CLOSED));
	HashInsertInt(NULL,"Indexes","Closed", ImageList_AddMasked(img,bmp,RGB(255,255,255)) );
	DeleteObject(bmp);

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_MESSAGE));
	HashInsertInt(NULL,"Indexes","Message", ImageList_AddMasked(img,bmp,RGB(255,0,0)) );
	DeleteObject(bmp);
	
	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_NONE));
	HashInsertInt(NULL,"Indexes","None",ImageList_AddMasked(img,bmp,RGB(255,255,255)) );
	DeleteObject(bmp);

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_FAVOURITES));
	HashInsertInt(NULL,"Indexes","Favourites", ImageList_AddMasked(img,bmp,RGB(255,0,0)) );
	DeleteObject(bmp);

	bmp = LoadBitmap(PROG_HINSTANCE,MAKEINTRESOURCE(IDB_LAN));
	HashInsertInt(NULL,"Indexes","LAN", ImageList_AddMasked(img,bmp,RGB(255,0,0)) );
	DeleteObject(bmp);

	
	for (i=0;i<num_server_types;i++)
		if (server_types[i]->iImage) {
			index = ImageList_AddMasked(img,(void*)server_types[i]->iImage,RGB(255,0,255) );
			rs = DeleteObject((void*)server_types[i]->iImage);
			server_types[i]->iImage = index;
		}

	//ImageList_SetBkColor(img,RGB(255,255,255) );
	HashInsertPtr(NULL,"Global",name,img);
 
	return;
}

void ListDestroyTreeviewOfList(list_t* list, int remove) {
	list_t* other;
	server_t* svr;
	
	int i, idx;
	HASH* hash;

	HWND hwnd = HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR);

	hash = HashGet(HASH_GLOBAL,"Lists");

	for (i=0; i<HashCount(HASH_GLOBAL,"Lists");i++)
	{
		other = HashSelectPtr(HASH_GLOBAL,"Lists",(HASH_PARAM)i);

		if (HashSelectPtr(list,"ptrs","parent") == list)
			ListDestroyTreeviewOfList(other,1);
	}

	if (!list) 
		return;

	if (remove) {
		TreeView_DeleteItem(hwnd,list->hTreeItem);
		list->hTreeItem = NULL;
	}
		
	for (svr = HashReset(list,"servers",HASH_PTR);
			svr != NULL; svr = HashNext(list,"servers",HASH_PTR)) 
	{
		idx = (int) HashExistsValue(svr,"lists",list,HASH_PTR);
		if (idx != -1)
		{
			TreeView_DeleteItem(hwnd,HashSelect(svr,"hTreeItems",(HASH_PARAM)idx,HASH_PTR) );
			HashDelete(svr,"hTreeItems",(HASH_PARAM)idx);
		}
	}
			
	return;
}

void ListRebuildTreeviewOfList(list_t* list, int add) {
	list_t* other, *parent;
	ITEM_TYPE* it;
	server_t* svr;
	TV_INSERTSTRUCT tv;
	int i;
	HASH* hash;
	int index;

	HWND hwnd = HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR);

	hash = HashGet(HASH_GLOBAL,"Lists");

	for (i=0; i<HashCount(HASH_GLOBAL,"Lists");i++)
	{
		other = HashSelectPtr(HASH_GLOBAL,"Lists",(HASH_PARAM)i);

		if (HashSelectPtr(list,"ptrs","parent") == list)
			ListRebuildTreeviewOfList(other,1);
	}

	if (!list || HSI(list,"Extra","HiddenList")) 
		return;

	if (! (it = HashSelectPtr(list,"ptrs","item_type")) ) {
		it = (ITEM_TYPE*) calloc(sizeof(ITEM_TYPE),1);
		it->type = IT_LIST;
		it->ptr = list;
		HashInsertPtr(list,"ptrs","item_type",it);
	}
	
	if (! (parent = HashSelectPtr(list,"ptrs","parent") ) )
		tv.hParent = HashSelect(HASH_GLOBAL,"Global","ServersTreeItem",HASH_PTR); //NULL;
	else tv.hParent = parent->hTreeItem;

	tv.hInsertAfter = TVI_LAST;
	tv.item.pszText = list->desc;
	 
	index = HashSelectInt(NULL,"Indexes","Closed");
	if (strcmp(GroupOfList(list)->name,"Favourites") == 0)
		index = HashSelectInt(NULL,"Indexes","Favourites");
	if (strcmp(GroupOfList(list)->name,"LANServers") == 0)
		index = HashSelectInt(NULL,"Indexes","LAN");
	tv.item.iImage = tv.item.iSelectedImage = index;
	tv.item.mask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	tv.item.lParam = (LPARAM)it;	
	
	if (add)
		list->hTreeItem = TreeView_InsertItem(hwnd,&tv);

	for (svr = HashReset(list,"servers",HASH_PTR);
		svr != NULL; svr = HashNext(list,"servers",HASH_PTR))
			UI_AddServer(svr,list);

	ListSort(HashSelectPtr(list,"ptrs","parent"));

	return;
}
		
void ListRebuildTreeview(void) {
	static ITEM_TYPE it = {IT_ROOT,NULL};	
	TV_INSERTSTRUCT tv;
	HWND hwnd = HashSelectPtr(HASH_GLOBAL,"Global","ServersTree");
	
	TreeView_DeleteAllItems(hwnd);

	TreeView_SetImageList(hwnd, HashSelectPtr(NULL,"Global","ImageList"), TVSIL_NORMAL); 
	
	tv.hParent = NULL;
	tv.hInsertAfter = TVI_LAST;
	tv.item.pszText = "Servers";
	tv.item.iImage = HashSelectInt(NULL,"Indexes","Open");
	tv.item.iSelectedImage = HashSelectInt(NULL,"Indexes","Open");

	tv.item.mask = TVIF_TEXT | TVIF_STATE | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	
	it.type = IT_LIST;
	it.ptr = GetAllServersList();
	HII(it.ptr,"extra","root",1);

	tv.item.lParam = (LPARAM)&it;
	tv.item.stateMask = TVIS_EXPANDED;
	tv.item.state = TVIS_EXPANDED;

	HashInsert(HASH_GLOBAL, "Global","ServersTreeItem", TreeView_InsertItem(hwnd,&tv), HASH_PTR);

	ListRebuildTreeviewOfList(NULL,1);
}

void ListAddServer(list_t* list)
{
	PROPERTIES_DialogBox(NULL,PROG_HWND,list,1);
}

void EmptyList(list_t* list) 
{
	/*char name[128];
	char group[128];
	list_t* parent;
	strcpy(name,list->desc);
	strcpy(group,GroupOfList(list)->name);
	parent = HashSelectPtr(list,"ptrs","parent");

	DeleteList(list);
	CreateList(name,group,parent);*/
	DeleteServerList(list);
}


void ListMove(list_t* list, int count, int dir) 
{
	list_t* other, *parent;
	int i;
	int index;
	HASH* hash;
	//HTREEITEM hTreeItem;
	int c = 0;
	//list_t* p;

	parent = HashSelectPtr(list,"ptrs","parent");
	

	/* Find where we are in the list */
	index = (int) HashExistsValue(HASH_GLOBAL,"Lists",list,HASH_PTR);

	hash = HashGet(HASH_GLOBAL,"Lists");
	for (i=index+dir; i>= 0 && i<HashCount(HASH_GLOBAL,"Lists");i+=dir)
	{
		other = HashSelectPtr(HASH_GLOBAL,"Lists",(HASH_PARAM)i);

		if (HashSelectPtr(other,"ptrs","parent") == parent &&
				++c == count) {
			//ListDestroyTreeviewOfList(list,0);
			//ListDestroyTreeviewOfList(other,0);

			HashMove(HASH_GLOBAL,"Lists",index,i);
			ListSort(HashSelectPtr(other,"ptrs","parent") );

			/*hTreeItem = list->hTreeItem;
			list->hTreeItem = other->hTreeItem;
			other->hTreeItem = hTreeItem;
			RENAMELIST_Modify(list,list->desc);
			RENAMELIST_Modify(other,other->desc);

			ListRebuildTreeviewOfList(list,0);
			ListRebuildTreeviewOfList(other,0);*/

			//TreeView_SelectItem(PROG_TREEVIEW,list->hTreeItem);
			return;
		}
	}
}

void ListMoveUp(list_t* list) {
	ListMove(list,1,-1);
}
void ListMoveDown(list_t* list) {
	ListMove(list,1,+1);
}
