/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include <io.h>
#include <time.h>
#include <process.h>

#include "resource.h"
#include <windows.h>
#include <winuser.h>
#include <direct.h>
#include <ddeml.h>

#include "stringtable.h"
#include "comm.h"
#include "types.h"
#include "hashdef.h"
#include "progcfg.h"
#include "buddy.h"
#include "msg.h"
#include "main.h"
#include "commctrl.h"
#include "scheme.h"
#include "utils.h"

#include "shared.h"
#include "gencfg.h"
#include "default.h"
#include "list.h"
#include "disk.h"
#include "password.h"
#include "about.h"
#include "http.h"
#include "events.h"
#include "motd.h"
#include "onlinelists.h"
#include "systray.h"
#include "ui.h"
#include "thread_recv.h"
#include "comm.h"
#include "limits.h"


SOCKET InitSocket(void);

#include "group.h"
#include "funcexport.h"
/* 
'Detailed Server Info' dialog box procedure.
Does nothing but close the window on user's request.
*/
int WINAPI SERVERINFO_DlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	if (msg == WM_CLOSE) {
		HashDelete(HASH_GLOBAL,"Global","ServerInfo");
		DestroyWindow(hwnd);
	}
	return 0;
}

/* Display the ServerInfo dialog box when requested by the user. */
void SERVERINFO_DialogBox(void) {
	RECT r;

	HashInsert(HASH_GLOBAL,"Global","ServerInfo",CreateDialog(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_SERVERINFO),PROG_HWND,SERVERINFO_DlgProc),HASH_PTR);
	GetWindowRect(PROG_HWND,&r);
	SetWindowPos(HashSelect(HASH_GLOBAL,"Global","ServerInfo",HASH_PTR),0,r.right-200,r.bottom-150,0,0,SWP_NOSIZE|SWP_NOREPOSITION);
}

list_t* GetCurrentList(void)
{
	return HashSelectPtr(HASH_GLOBAL,"Status","CurrentList");
}

list_t* GetLANList(void)
{
	list_t* list;

	list = HashFind(HASH_GLOBAL,"Lists","LANServers",(int)GroupOfList(NULL)->name);
				
	if (!list && HSIG("Settings","ShowLANList"))
		list = CreateList("LAN","LANServers",NULL);

	return list;
}

void* GetAllServersList(void)
{
	list_t* list;

	list = HashFind(HASH_GLOBAL,"Lists","ListAllServers",(int)GroupOfList(NULL)->name);
				
	if (!list)
	{
		list = CreateList(NULL,"ListAllServers",NULL);
		HII(list,"Extra", "HiddenList", 1);
		HII(list,"Extra","UnsavableList", 1);
	}

	return (void*)list;
}



TV_ITEM current_server_info;



extern int G_numHashes;

int shutdownTime;

void timeTaken()
{
	char buf[128];
	sprintf(buf,"time taken: %d ms",GetTickCount() - shutdownTime);
//	WRITE_DEBUG(buf);

	shutdownTime = GetTickCount();
}

void ShutDown(void)
{
	int i;
	list_t* list;
	char buf[128];

	//ShowWindow(PROG_HWND,SW_HIDE);
	
	sprintf(buf,"Shutting down, %d hashes",G_NumHashes);
	WRITE_DEBUG(buf);
	HIIG("Status","Shutdown",1);

	shutdownTime = GetTickCount();

	WRITE_DEBUG("Empty clipboard");
	// Free any server on the clipboard
	ServerClipboardEmpty();
	timeTaken();

	WRITE_DEBUG("Close systray");
	// Shut down systray icon (if any)
	SystrayClose();
	timeTaken();
	
	WRITE_DEBUG("Save settings");
	// Save current settings and servers to disk
	ServersClose();
	timeTaken();

	// Hide window
	//ShowWindow(PROG_HWND,SW_HIDE);

	WRITE_DEBUG("Close chat");
	// Shut down chat
#ifdef _SQCHAT
	if ((ptr = HashSelectPtr(HASH_GLOBAL,"Status","ChatActive")) != NULL)
		CHAT_Cleanup(ptr);
	timeTaken();
#endif


	WRITE_DEBUG("Free events");
	// free allocated events
	EVENTS_Free();
	timeTaken();

	WRITE_DEBUG("Stop HTTP");
	// stop HTTP connections
	HTTP_Free();
	timeTaken();

	WRITE_DEBUG("Close socket");
	// Close our socket
	closesocket(HashSelectInt(HASH_GLOBAL,"Global","Socket"));
	timeTaken();

	// prevent time-consuming hash re-indexing from occuring
	ServersListUpdateEnable(0);
	
	WRITE_DEBUG("Free lists and servers");
	// free all lists and associated servers
	while ((list=HashReset(HASH_GLOBAL,"Lists",HASH_PTR)) != NULL)
	{
		sprintf(buf,"Deleting '%s'",list->desc);
		WRITE_DEBUG(buf);

		// Don't free the list, as the servers belong to someone else
		//if (strcmp(GroupOfList(list)->name,"ListAllServers") != 0)
			DeleteList(list);	
		//else
		//	HashDeleteValue(HASH_GLOBAL,"Lists",list,HASH_PTR);

		timeTaken();
	}

	i = HashCount(HASH_GLOBAL,"AllServers");

	// No lists left, so mark this as so until final shutdown is complete
	HashInsertPtr(HASH_GLOBAL,"Status","CurrentList",NULL);
	
	WRITE_DEBUG("Free groups");
	// free the allocated server list group types
	FreeGroupTypes();
	timeTaken();

	// Close the main window
	DestroyWindow(PROG_HWND);
	timeTaken();

	WRITE_DEBUG("Free imagelist");
	// free the treeview imagelist
	ImageList_Destroy(HashSelectPtr(HASH_GLOBAL,"Global","imagelist"));
	ImageList_Destroy(HashSelectPtr(HASH_GLOBAL,"Global","statelist"));
	timeTaken();

	// Do windows winsock shut down
	shutdownWinSock();
	timeTaken();

	WRITE_DEBUG("Free DLLs");
	// Chuck out the loaded DLL modules
	for (i=0;i<num_server_types;i++)
	{
		if (server_types[i]->port 
			&& server_types[i]->module) 
			// since some modules load more than one server type
			FreeLibrary((HMODULE)server_types[i]->module);
		free(server_types[i]->parameters);
		free(server_types[i]);
	}
	timeTaken();

	WRITE_DEBUG("Free hashes");
	// Get rid of remaining hashes
	HashFreeAll();
	timeTaken();

	return;
}


void ShowPing(server_t* svr)
{
	char str[128];
	if (!HashSelect(HASH_GLOBAL,"Settings","ShowPing",HASH_INT))
		return;
	
	sprintf(str, "%d",svr->ping * 5/6);
	if (svr->ping <= 0 || svr->ping > 10000)
		sprintf(str,"");

	if (HashSelectPtr(HASH_GLOBAL,"Global","ServerInfo")) {
		SetDlgItemText(HashSelectPtr(HASH_GLOBAL,"Global","ServerInfo"),IDC_PING,"Ping:");
		SetDlgItemText(HashSelectPtr(HASH_GLOBAL,"Global","ServerInfo"),IDT_PING,str);
	}
	else if (TEXT_DESCS) {
		SetDlgItemText(PROG_HWND,IDC_PING,"Ping:");
		SetDlgItemText(PROG_HWND,IDT_PING,str);
	}

	HashInsertText(HASH_GLOBAL,"Status","Ping",str);
	return;
}

void Refresh(int update)
{
	server_t* svr;
	types_t* t;

	svr = CURRENT_SERVER;
	if (!svr) {
		// This makes no sense anymore
		//temp = GetDlgItem(PROG_HWND, IDL_PLAYERS);
		//SendMessage(temp, LB_RESETCONTENT, 0, 0);
		return;
	}
	t = svr->gametype;
	if (!t)
		MSG_DialogBox(NULL,NULL,"Unknown server type!", IDD_ERROR);
	else {
		UI_UpdateServer(svr);
		
		t->print_info(svr, PROG_HWND, update);
		if (HashSelectPtr(HASH_GLOBAL,"Global","ServerInfo"))
			t->print_info(svr, HashSelect(HASH_GLOBAL,"Global","ServerInfo",HASH_PTR), update);
		ShowPing(svr);
	}

	return;
}

void UI_SetStatusText(char* text)
{
	SetWindowText(GetDlgItem(PROG_HWND,IDT_STATUS),text);
}

void UI_ProgressBar(int set)
{
	if (set != HSIG("Status","ShowProgressBar"))
	{
		HIIG("Status","ShowProgressBar",set);
		SendMessage(PROG_HWND,WM_SIZE,0,0);
	}
}

void ServerUpdate(server_t* svr)
{
	int limit, players = 0, total = 0;
	list_t* list = CURRENT_LIST;
	int i;
	char buf[128];

	limit = HashCount(list,"servers");

	for (svr=HashReset(list,"servers",HASH_PTR);svr;svr=HashNext(list,"servers",HASH_PTR))
	{
		players += HashSelectInt(svr,"stats","numplayers");
		total += HashSelectInt(svr,"stats","maxplayers");
	}
	sprintf(buf,"%d servers, %d/%d players.",limit,players,total);

	i = HashSelectInt(list,"pinglist","lastsvr");
	if (i && i < limit)
	{
		sprintf(buf+strlen(buf)," (Refreshing list, %d servers remaining.)",limit-i);
		SendMessage(GetDlgItem(PROG_HWND,IDC_PROGRESS),PBM_SETRANGE,
			0, MAKELPARAM(0,limit) );
		SendMessage(GetDlgItem(PROG_HWND,IDC_PROGRESS),PBM_SETPOS,i,0);
		UI_ProgressBar(1);

	}
    else
	{
		HashInsertInt(list,"pinglist","lastsvr",0);
		SendMessage(GetDlgItem(PROG_HWND,IDC_PROGRESS),PBM_SETPOS,0,0);
		UI_ProgressBar(0);
	}
		

	UI_SetStatusText(buf);

}

// USED TO CALCULATE SERVER PING
unsigned int last_req = 0;

void SetServerPing(server_t* svr, int ping)
{
	if (ping < 0)
		ping = 0;

	svr->ping = ping;
	svr->ping_count++;
	if (svr->ping_count == 0)
	{
		svr->ping_count = 1;
		svr->avg_ping = 0;
	}
	svr->avg_ping += (ping - svr->avg_ping) / (int)svr->ping_count;
}

#define REQ_CACHE_SIZE	32
static server_t* RequestCache[REQ_CACHE_SIZE];

void AddRequestCache(server_t* svr)
{
	memmove(RequestCache+1,RequestCache,(REQ_CACHE_SIZE-1)*sizeof(server_t*));
	RequestCache[0] = svr;
}
void DropRequestCache(server_t* svr)
{
	int i;
	for (i=0;i<REQ_CACHE_SIZE;i++)
		if (RequestCache[i] == svr)
			RequestCache[i] = NULL;
}
			


int CheckIncomingData(void)
{
	SOCKADDR_IN dest_addr;
	char buf[4096], ipaddr[128];
	int len, which, buf_len, type, tack = 'B';
	server_t* svr, *temp, *current = CURRENT_SERVER;
//	unsigned int process_time;
	int used = 0;
	list_t* lanlist, *list;
	RECV_BUFFER** root, *next;
	static int ignore_future;
	int received;
	int cache_index = 0;


	if (0 && ignore_future)
		return 0;

	which = -1;

	WRITE_DEBUG("Checking for incoming data");
	//if (!CheckSocket((int) HashSelect(HASH_GLOBAL,"Global","Socket",HASH_INT))) {
	//	return 0;

	RecvThread(NULL);

	root = RecvThreadGetRoot();
	if (*root)
	{
		if ((*root)->received - GetTickCount() < 100)
			return 0; 

		WRITE_DEBUG("Data found, receiving");
		//received = GetTickCount();
		len = sizeof(SOCKADDR_IN);
		//buf_len = recvfrom((int) HashSelect(HASH_GLOBAL,"Global","Socket",HASH_INT), buf, 2048, 0, (SOCKADDR*)&dest_addr, &len);
		dest_addr = (*root)->dest_addr;
		memcpy(buf,(*root)->data,4096);
		buf_len = (*root)->len;
		received = (*root)->received;

		memset(buf+buf_len,0,4096-buf_len);

		

//		RecvThreadWaitSafe();
		next = (*root)->next;
		free (*root);
		*root = next;
//		RecvThreadSetSafe(1);


		WRITE_DEBUG("Checking lists for possible owner");
		for (list = HashReset(HASH_GLOBAL,"Lists",HASH_PTR);list;
				list = HashNext(HASH_GLOBAL,"Lists",HASH_PTR) )
		{
			RecvThread(NULL);

			if (GroupOfList(list)->is_packet_for &&
				GroupOfList(list)->is_packet_for(list,(SOCKADDR*)&dest_addr) )
			{
				GroupOfList(list)->packet(list,buf,buf_len);
				used = 1;
			}
		}

		if (used)
			return 1;

		WRITE_DEBUG("No list owner found");

		WRITE_DEBUG("Checking server cache");

		

		used = 0;
		cache_index = 0; svr = NULL;
		//cache_index = -1; svr = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);


		while (cache_index != -1 || svr)
		{
			RecvThread(NULL);

			if (cache_index != -1)
				svr = RequestCache[cache_index];

			if (svr && svr->gametype->isPacketFrom(svr,(SOCKADDR*)&dest_addr))
			{	// server already on list, mark it as still existing
				WRITE_DEBUG("Server found");
				svr->timeout = 0;

				if (buf_len < 0)
				{
					WRITE_DEBUG("Timed out");
					//error condition - timeout server immediately
					svr->timeout = 30;
					Refresh(0);
					return 0;
				}
				if (svr->gametype->receive(svr, buf, buf_len)) 
				{
					WRITE_DEBUG("Received data successfully");
					
					/*if (svr->lan)
					{
						WRITE_DEBUG("Setting ping");
						SetServerPing(svr, received - HSIG("LastRequest",(HASH_PARAM)svr->type) );
					}
					else */
					if (svr->last_req)
					{
						WRITE_DEBUG("Setting ping");
						SetServerPing(svr, received - svr->last_req);
					}
					svr->last_req = 0;
			
					if ((!strcmp(svr->name, "New Server") || HSIG("settings","UpdateNames") || svr->lan || svr->unnamed || !strcmp(svr->name,svr->ipaddr)) && svr->gametype->hostname(svr)[0])
					{
						WRITE_DEBUG("Setting new hostname");
						free (svr->name);
						svr->name = (char*) malloc(128);
						strcpy(svr->name, svr->gametype->hostname(svr));
						svr->name[63] = 0;
						svr->unnamed = 0;

						// update tree item
						UI_UpdateServer(svr);
						
						//ServersListUpdate(1); // force update due to name change
					}
					

					if (svr == current)
					{
						WRITE_DEBUG("Refreshing");
						Refresh(0);
					}
					
					// Dual pane shows more info, so update it
					if (HSIG("Settings","DualPane"))
					{
						WRITE_DEBUG("Updating second pane");
						UI_UpdateServer(svr);

						if (HSIG("Settings","AlwaysSort"))
							for (list=HashReset(svr,"lists",HASH_PTR);list;list=HashNext(svr,"lists",HASH_PTR))
								UI_SortList(list);
												
						ServerUpdate(svr);
						//UI_SortList(svr->list);
					}
				}
				used = 1;
				 break;   
				// - allow all servers a chance to grab the packet... this
				// allows players to enter two servers of same ip and port.
				// 
				 // Redisabled this due to cpu load
			}

			if (cache_index == -1)
				svr = HashNext(HASH_GLOBAL,"AllServers",HASH_PTR);
			else
			{
				if (++cache_index == REQ_CACHE_SIZE)
				{
					WRITE_DEBUG("Checking all servers");		
					cache_index = -1;
					svr = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);
				}
			}
		}
		
		
		if (!used && HSIG("Settings","ShowLANList")) 
		{ // server not on list, add a new LAN server
			WRITE_DEBUG("Server not found, adding LAN server");
							
			for (type = 0; type<MAX_SERVER_TYPES;type++)
			{
				RecvThread(NULL);

				if (GetServerType(type)->port == ntohs(dest_addr.sin_port) ||
						(GetServerType(type)->recv_port &&
						GetServerType(type)->recv_port == ntohs(dest_addr.sin_port)))
					break;
			}
			if (type != MAX_SERVER_TYPES)
			{
				ignore_future = 1;

				WRITE_DEBUG("LAN Type found, adding");
				HashInsert(HASH_GLOBAL,"Status","IgnoreLAN",(HASH_PARAM)1,HASH_INT);

				
				//list = HashFind(HASH_GLOBAL,"Lists","LANServers",(int)GroupOfList(NULL)->name);
				//if (!list)
				//	list = CreateList("LAN","LANServers",NULL);
				list = GetLANList();

				if (HashSelect(HASH_GLOBAL,"Settings","LANSortByGame",HASH_INT)) {
					lanlist = list;
					list = ListNamedWithParent(GetServerType(type)->name,list);
					if (!list)
						list = CreateList(GetServerType(type)->name,"LANServers",lanlist);
				}

				sprintf(ipaddr, "%s:%d",inet_ntoa(dest_addr.sin_addr), GetServerType(type)->port );

				svr = AddServer(list,NewServer("", ipaddr, GetServerType(type)->qstat_type) );
				svr->lan = 1;
				//SetServerPing(svr, received - HSIG("LastRequest",(HASH_PARAM)type) );
				SetServerPing(svr, received - last_req);

				
				svr->gametype->receive(svr, buf, buf_len);
				free(svr->name); svr->name = malloc(64);
				memcpy(svr->name, svr->gametype->hostname(svr),64);
				svr->name[63] = 0;
					
				if (!strcmp(svr->name, ""))
					strcpy(svr->name, "Noname");
				sprintf(buf,"%s",svr->name);
				
				temp = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);
				while (temp)
				{
					RecvThread(NULL);

					if (temp != svr && !strcmp(temp->name, buf))
					{
						sprintf(buf,"%s %c",svr->name,tack);
						tack++;
						temp = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);
					}
					else temp = HashNext(HASH_GLOBAL,"AllServers",HASH_PTR);
				}
				strcpy(svr->name, buf);
				UI_UpdateServer(svr);

				if (HSIG("Settings","AlwaysSort"))
					for (list=HashReset(svr,"lists",HASH_PTR);list;list=HashNext(svr,"lists",HASH_PTR))
						UI_SortList(list);

				ServerUpdate(svr);

				//UI_SortList(svr->list);
//				ServersListUpdate(2); // name change, force an update
				//HashInsert(HASH_GLOBAL,"Status","ListUpdate",(HASH_PARAM)(GetTickCount() + 50),HASH_INT);
			}
		}
		return 1;
	}
	return 0;
}


void RequestNewData(int timeout_servers)
{
	int refresh = 0;
	char buf[2048];

	server_t* svr, *current;
	

	current = CURRENT_SERVER;

	if (current)
	{	sprintf(buf,"Current server: %s (%s)",current->name,current->ipaddr);
		//WRITE_DEBUG(buf);
	} //else WRITE_DEBUG("No current server");

	
	if (!HashSelect(HASH_GLOBAL,"Settings","PingAllNetServers",HASH_INT))
	{	// ping current server only, otherwise will get done further down
		//WRITE_DEBUG("only pinging current server");
		if (current ) // && !current->lan) 
		{
		//	WRITE_DEBUG("Current server is not lan based");
			RequestServerStatus(current);
			AddRequestCache(current);
		//	WRITE_DEBUG("Saving last request time");
			//current->last_req = GetTickCount();
		}
	//	WRITE_DEBUG("check for waiting server");
		if ((svr = HashSelectPtr(HASH_GLOBAL,"Status","WaitingServer")) != NULL)
		{
		//	WRITE_DEBUG("check waiting server");
			RequestServerStatus(svr);
			AddRequestCache(svr);
			//svr->last_req = GetTickCount();
		}
	}
	else 
	{	// ping all servers on net
	//	WRITE_DEBUG("ping all servers");
		svr = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);		   
		while (svr)
		{
			sprintf(buf,"Checking server %s (%s)",svr->name,svr->ipaddr);
		//	WRITE_DEBUG(buf);
		//	WRITE_DEBUG("check if its a lan serveR");
			if (HashExistsValue(HASH_GLOBAL,"LANServers",svr,HASH_PTR) == (HASH_PARAM)-1) {
		//		WRITE_DEBUG("it isn't, ping it");
				RequestServerStatus(svr);
				AddRequestCache(svr);
				//svr->last_req = GetTickCount();
				svr = HashNext(HASH_GLOBAL,"AllServers",HASH_PTR);
			}
		}
	}
	//WRITE_DEBUG("ping lan servers?");
	/*
	if (!HashSelect(HASH_GLOBAL,"Settings","NoLAN",HASH_INT) &&
		timeout_servers) 	// ping LAN servers if permitted
						    // and request was not caused by a server change
	{
		//WRITE_DEBUG("yes");
		WRITE_DEBUG("initiate lan server status");
		RequestLANServersStatus();
		WRITE_DEBUG("done lan server status");
	}
	*/

	WRITE_DEBUG("save last request time for lan servers");
	last_req = GetAccTickCount();

	WRITE_DEBUG("timeout servers?");
	if (!timeout_servers) {
		WRITE_DEBUG("no, just return now");
		HashInsertInt(HASH_GLOBAL,"Status","NextRefresh",GetTickCount() + CLOCKS_PER_SEC * HashSelectInt(HASH_GLOBAL,"Settings","Refresh"));	
		return;
	}

	WRITE_DEBUG("yes, step through all servers");
	svr = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);
	while (svr != NULL)
	{
		sprintf(buf,"checking server %s (%s)",svr->name,svr->ipaddr);
		WRITE_DEBUG(buf);

		// FIXME: bit of a hack.  Should be handled better.
		svr->WAITING_FIRST_PACKET = 1;

		WRITE_DEBUG("check if server is on lan, current, or whether all servers are being pinged");

		if (svr->lan || HashSelectInt(HASH_GLOBAL,"Settings","PingAllNetServers") || 
			svr == current)
		{
			WRITE_DEBUG("yes, so increment the timeout value");
			svr->timeout += HashSelectInt(HASH_GLOBAL,"Settings","Refresh");

			WRITE_DEBUG("has this server timed out?");
			if ((!svr->lan && svr->timeout >= HashSelectInt(HASH_GLOBAL,"Settings","Timeout")) ||
				(svr->lan && svr->timeout >= HashSelectInt(HASH_GLOBAL,"Settings","LANTimeout")))
			{
				WRITE_DEBUG("yes");
				if (svr->lan) {
					WRITE_DEBUG("Lan server timed out, remove it");
					DeleteServer(svr,GetLANList());
					refresh = 1;
					svr = HashReset(HASH_GLOBAL,"AllServers",HASH_PTR);
				}
				else if (svr == current)
				{
					WRITE_DEBUG("current server timed out, so display a blank page");
					svr->gametype->print_info(svr, PROG_HWND, 0);
				}
			}
		}
		svr = HashNext(HASH_GLOBAL,"AllServers",HASH_PTR);
	}
	HashInsertInt(HASH_GLOBAL,"Status","NextRefresh",GetTickCount() + CLOCKS_PER_SEC * HashSelectInt(HASH_GLOBAL,"Settings","Refresh"));	
	WRITE_DEBUG("done, returning to main()");
}


// From cvars.c
void CVARS_DialogBox(server_t* svr);

int WINAPI SERVERFULL_DlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	char* textmsg = "The server you have requested to connect to is currently "
		"reporting all player spots as full.  If you attempt to connect, "
		"you will most likely get a 'Server Is Full' error.\n\n"
		"You may select 'Wait for free spot' to connect once a connection "
		"should be possible, or Cancel the connection process altogether.";

	if (msg == WM_INITDIALOG)
		SetDlgItemText(hwnd,IDT_MSG,textmsg);

	if (msg == WM_COMMAND) {
		if (wparam == IDB_CONNECT)
			EndDialog(hwnd,1);
		if (wparam == IDB_WAIT)
			EndDialog(hwnd,2);
		if (wparam == IDB_AUDIO)
			EndDialog(hwnd,3);
		if (wparam == IDB_CANCEL)
			EndDialog(hwnd,0);
	}
	if (msg == WM_CLOSE)
		EndDialog(hwnd,0);

	return 0;
}

int ConnectFullServerWarning(server_t* server)
{
	int i;

	i = DialogBox(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_SERVERFULL),
		PROG_HWND,SERVERFULL_DlgProc);
	
	if (i == 1)
		return 1;
	else if (i == 2)
	{
		HashInsertPtr(HASH_GLOBAL,"Status","WaitingServer",CURRENT_SERVER);
		return 0;
	}
	else if (i == 3)
	{
		HashInsertPtr(HASH_GLOBAL,"Status","WaitingServer",CURRENT_SERVER);
		HashInsertInt(HASH_GLOBAL,"Status","WaitingServerAudio",1);
		return 0;
	}
	else return 0;
}

void str_replace(char* str, char* match, char* replace)
{
	char* idx;
	char tmp[1024];

	while ((idx = strstr(str,match)) != NULL)
	{
		strcpy(tmp,idx+strlen(match));
		strcpy(idx,replace);
		strcpy(idx+strlen(replace),tmp);
	}
}

int ConnectServer(server_t* svr)
{
	char buf[1024], dir[256];
	char tmp[1024];
	char tmpl[128];
	char ip[32]; char port[32];
	char* idx;

	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	int i;
	time_t tt;
	struct tm* tm;

	//FILE* fp;
	
	types_t *t;
	int (*isfull)(server_t*);

	if (svr == NULL)
	{
		return 1;
	}

	t = svr->gametype;

	if ((isfull = HashSelectPtr(t,"funcs","isfull") ) != NULL)
	{
		if (isfull(svr) && !ConnectFullServerWarning(svr))
			return 0;
	}
	else if (atoi(GetValue(svr, "numplayers")) && atoi(GetValue(svr,"numplayers")) == atoi(GetValue(svr, "maxplayers"))
			&& !ConnectFullServerWarning(svr))
		return 0;

	HashDelete(HASH_GLOBAL,"Status","WaitingServer");
		
	
	// FIXME
	// DISABLED 23/4/2001
	if (0 ) { // && (svr->type == 7 || svr->type == 10)) {
		// special unreal start up
		sprintf(buf,"start unreal://%s/",svr->ipaddr);
		if (WinExec(buf,0) <= 31)
			return 1;
	}
	else {
	
	if (t->path[0] == 0)
	{
		MSG_DialogBox(NULL,NULL, "You have not specified an executable to use when connecting to this server.",IDD_PROGERROR);
		return 0;
	}
	strcpy(dir, t->path);
	for (i = strlen(dir); i>=0; i--)
		if (dir[i] == '\\') {
			dir[i] = 0;
			break;
		}
	
	// New connect method
	strcpy(buf, t->connect_template);
	if (buf[0] == 0)
		strcpy(buf,"$EXE +connect $IP $PARAMS");

	sprintf(tmp,"\"%s\"",t->path);
	str_replace(buf,"$EXE", tmp);

	str_replace(buf,"$IP", svr->ipaddr);
	str_replace(buf,"$PARAMS",t->parameters);

	// break up ip and port for weird games
	strcpy(ip,svr->ipaddr);
	strcpy(port,"");
	idx = strchr(ip,':');
	if (idx)
	{
		*idx = 0;
		strcpy(port,idx+1);
	}

	str_replace(buf,"$HOSTIP",ip);
	str_replace(buf,"$HOSTPORT",port);

	if (GetValue(svr,t->mod_format) == NULL || strcmp(GetValue(svr,t->mod_format),"") == 0)
		strcpy(tmpl,"$MOD");
	else
		strcpy(tmpl, t->mod_format);
	
	str_replace(tmpl,"$MOD", GetValue(svr,t->mod_var) );
	str_replace(buf,"$MOD",tmpl);

	if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
	{
		strcpy(tmpl, t->password_format);
		str_replace(tmpl,"$PASSWORD", HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING) );
	}
	else
		strcpy(tmpl,"");

	str_replace(buf,"$PASSWORD",tmpl);

	if (strcmp(HashSelectText(HASH_GLOBAL,"Status","Param"),"") != 0)
		sprintf(buf+strlen(buf)," %s",HashSelectText(HASH_GLOBAL,"Status","Param"));

	time(&tt);
	tm = localtime(&tt);
	sprintf(tmp,"%04d%02d%02d-%02d%02d%02d",tm->tm_year+1900,tm->tm_mon+1,
				tm->tm_mday,tm->tm_hour,tm->tm_min,tm->tm_sec);

	str_replace(buf,"$DT",tmp);


#if 0	
	strcpy(buf,t->path);

	if (strlen(t->parameters))
		sprintf(buf+strlen(buf)," %s",t->parameters);
	
	if (t->cmdline)
		sprintf(buf+strlen(buf), " %s", t->cmdline(svr));
	else {
		if (GetValue(svr, "game") == NULL)
			strcpy(game, "\"\"");
		else if (!strcmp(GetValue(svr, "game"),""))
			strcpy(game, "\"\"");
		else strcpy(game, GetValue(svr, "game"));

		// Quake2 styles
		sprintf(buf+strlen(buf), " +set game %s " /*"+set clientport %d " */ "+connect %s",/*t->path,*/ game, /*27900 + rand() % 100,*/ svr->ipaddr);

		if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
			sprintf(buf+strlen(buf)," +set password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

		/*if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
			sprintf(buf+strlen(buf)," +set password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));*/
	}

	if (strcmp(HashSelectText(HASH_GLOBAL,"Status","Param"),"") != 0)
		sprintf(buf+strlen(buf)," %s",HashSelectText(HASH_GLOBAL,"Status","Param"));

	for (i=0;i<(int)strlen(buf);i++)
	{
		if (memcmp(buf+i,"$DT",3) == 0)
		{
			strcpy(tmp,buf);
			time(&tt);
			tm = localtime(&tt);
			sprintf(tmp+i,"%04d%02d%02d-%02d%02d%02d",tm->tm_year+1900,tm->tm_mon+1,
				tm->tm_mday,tm->tm_hour,tm->tm_min,tm->tm_sec);
			strcat(tmp,buf+i+3);
			strcpy(buf,tmp);
		}
	}
#endif

	//MessageBox(NULL,buf,"Launch",MB_OK);

	HashDelete(HASH_GLOBAL,"Status","Password");
	HashDelete(HASH_GLOBAL,"Status","Param");
	GetStartupInfo(&si);

	if (!CreateProcess(NULL,buf,NULL,NULL,0,NORMAL_PRIORITY_CLASS,NULL,dir,&si,&pi)) {
		sprintf(buf,"Unable to start %s.  Ensure you have specified the correct path.\n\n(Missing or corrupt %s)",t->name, t->path);
		HashInsertPtr(HASH_GLOBAL,"Status","ProgErrorRejoin",svr);
		MSG_DialogBox(NULL,NULL, buf, IDD_PROGERROR);
		return 0;
	}
	
	}

	if ((int) HashSelect(HASH_GLOBAL,"Settings","ExitOnConnect",HASH_INT) == 1) 
		PostQuitMessage(0); 

	return 1;
}

void WaitingServerAlert(void* param)
{
	int i;

	PlaySound("SystemExclamation",NULL,SND_ASYNC|SND_ALIAS|SND_NOWAIT);

	i = HSIG("Status","WaitingServerAudioCount");
	i--;

	HIIG("Status","WaitingServerAudioCount",i);

	if (i == 0)
		EVENTS_Cancel( HSPG("Status","WaitingServerAudioEV") );
}



void CheckWaitingServer(void)
{
	server_t* svr;
	types_t *t;
	int (*isfull)(server_t*);
	int notfull = 0;

	svr = HashSelectPtr(HASH_GLOBAL,"Status","WaitingServer");
	if (!svr)
		return;

	t = svr->gametype;
	
	if ((isfull = HashSelectPtr(t,"funcs","isfull") ) != NULL)
	{
		if (!isfull(svr))
			notfull = 1;
			
	}
	else if (atoi(GetValue(svr, "numplayers")) && atoi(GetValue(svr, "maxplayers")) &&  atoi(GetValue(svr,"numplayers")) != atoi(GetValue(svr, "maxplayers")) )
		notfull = 1;

	if (notfull)
	{
		if (HSIG("Status","WaitingServerAudio"))
		{
			HashDelete(HASH_GLOBAL,"Status","WaitingServer");
			HashDelete(HASH_GLOBAL,"Status","WaitingServerAudio");
			HashInsertInt(HASH_GLOBAL,"Status","WaitingServerAudioCount",3);
			HashInsertPtr(HASH_GLOBAL,"status","WaitingServerAudioEV",
				EVENTS_Register(WaitingServerAlert,NULL,(float)0.1,(float)1));
			return;
		}
		else
			ConnectServer(svr);
	}
	
	return;

}

#define SCM_DELETE		3001
#define SCM_ADDSERVER	3002
#define SCM_PROPERTIES	3003
#define SCM_CVARS		3005
#define SCM_CONNECT		3006
#define SCM_COPYDESC	3007
#define SCM_REFRESHLIST	3008
#define SCM_PASSWORD	3009
#define SCM_CREATELIST	3010
#define SCM_CUT			3011
#define SCM_COPY		3012
#define SCM_PASTE		3013
#define SCM_DELLIST		3014
#define SCM_RENAMELIST	3015
#define SCM_PARAM		3016
#define SCM_LISTMVUP	3017
#define SCM_LISTMVDOWN	3018
#define SCM_WAITFREE	3019
#define SCM_EXPORTQSTAT	3020
#define SCM_COPYIP		3021
#define SCM_ADDFAVOURITES	3022

void MENU_CreateServerMenu(HMENU menu, server_t* svr,list_t* list) {
	int (*isfull)(server_t*);

	if (svr == NULL)
	{
		AppendMenu(menu,MF_STRING|MF_GRAYED,0,"None selected");
		return;
	}
	AppendMenu(menu,MF_STRING,SCM_CONNECT,"Connect");
	SetMenuDefaultItem(menu,SCM_CONNECT,FALSE);

	if (! svr->gametype->disallowPath) {
		AppendMenu(menu,MF_STRING,SCM_PASSWORD,"Connect with password...");
		AppendMenu(menu,MF_STRING,SCM_PARAM,"Connect with parameters...");
	}

	if ((isfull = HashSelectPtr(svr->gametype,"funcs","isfull")) !=NULL)
	{
		if (isfull(svr))
			AppendMenu(menu,MF_STRING,SCM_WAITFREE,"Wait for free spot");
	}

	AppendMenu(menu,MF_SEPARATOR,0,"");
	AppendMenu(menu,MF_STRING,SCM_CVARS,"View CVars...");
	AppendMenu(menu,MF_STRING,SCM_COPYDESC,"Copy Desc");
	AppendMenu(menu,MF_STRING,SCM_COPYIP,"Copy IP");
	AppendMenu(menu,MF_SEPARATOR,0,"");
	if (GroupOfList(list)->add_favourites)
		AppendMenu(menu,MF_STRING,SCM_ADDFAVOURITES,"Add to favourites");
	if (GroupOfList(list)->cut_server)
		AppendMenu(menu,MF_STRING,SCM_CUT,"Cut");
	if (GroupOfList(list)->copy_server)
		AppendMenu(menu,MF_STRING,SCM_COPY,"Copy");
	if (GroupOfList(list)->delete_server)  {
		AppendMenu(menu,MF_STRING,SCM_DELETE,"Delete");
		AppendMenu(menu,MF_SEPARATOR,0,"");
	}
	AppendMenu(menu,MF_STRING,SCM_PROPERTIES,"Properties");
}

void MENU_CreateListMenu(HMENU menu, list_t* list) {

	int blank = 1;
	
	if (!list || HSI(list,"extra","root")) {
		AppendMenu(menu,MF_STRING,SCM_CREATELIST,"Create new list...");
		AppendMenu(menu,MF_STRING,SCM_EXPORTQSTAT,"Export QSTAT list...");
		return;
	}
	if (GroupOfList(list)->add_server || GroupOfList(list)->refresh_list)  {
		AppendMenu(menu,MF_STRING,SCM_CREATELIST,"Create sublist...");
		blank = 0;
	}
	if (GroupOfList(list)->delete_list) {
		AppendMenu(menu,MF_STRING,SCM_DELLIST,"Delete list");
		AppendMenu(menu,MF_STRING,SCM_EXPORTQSTAT,"Export QSTAT list...");
		blank = 0;
	}
	if (GroupOfList(list)->add_server) {
		AppendMenu(menu,MF_STRING,SCM_ADDSERVER,"Add server...");
		blank = 0;
	}
	if (GroupOfList(list)->refresh_list) {
		AppendMenu(menu,MF_STRING,SCM_REFRESHLIST,"Refresh");
		blank = 0;
	}
	if (GroupOfList(list)->paste_server &&
		GroupOfList(list)->pastable_server(list)) {
		AppendMenu(menu,MF_STRING,SCM_PASTE,"Paste server");
		blank = 0;
	}
	if (GroupOfList(list)->properties_list)  {
		if (!blank) 
			AppendMenu(menu,MF_SEPARATOR,0,"");
		AppendMenu(menu,MF_STRING,SCM_LISTMVUP,"Move Up");
		AppendMenu(menu,MF_STRING,SCM_LISTMVDOWN,"Move Down");
		AppendMenu(menu,MF_SEPARATOR,0,"");
		AppendMenu(menu,MF_STRING,SCM_RENAMELIST,"Properties");
		blank = 0;
	}

	
	if (blank)
		AppendMenu(menu,MF_STRING|MF_GRAYED,0,"No options available");
}

void DisplaySCM(server_t* server)
{
	static int x, y, temp;
	TV_HITTESTINFO info;
	RECT rect;
	HMENU menu;
	list_t* list;
	ITEM_TYPE* it, tempit;

	temp = GetMessagePos();
	x = LOWORD(temp);
	y = HIWORD(temp);

	GetWindowRect(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&rect);
	info.pt.x = x - rect.left;
	info.pt.y = y - rect.top;

	current_server_info.hItem = TreeView_HitTest(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&info);
	if (!server && !current_server_info.hItem) {
		menu = CreatePopupMenu();
		AppendMenu(menu,MF_STRING,SCM_CONNECT,"Create new list...");
	}
	else {
		if (current_server_info.hItem)
		{
			/* When TreeView_SelectItem is called; it triggers a 'list change' - we should ignore it */
			HashInsertInt(HASH_GLOBAL,"Status","IgnoreListChange",1);
	
			current_server_info.mask = TVIF_PARAM /*| TVIF_TEXT */ | TVIF_CHILDREN;
			TreeView_SelectItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),current_server_info.hItem);
			TreeView_GetItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&current_server_info);

			HashInsertInt(HASH_GLOBAL,"Status","IgnoreListChange",0);
		}
	
		it = (ITEM_TYPE*)current_server_info.lParam;

		if (server) {
			it = &tempit;
			tempit.ptr = server;
			tempit.type = IT_SERVER;
			tempit.owner = CURRENT_LIST;
		}

		if (it->type == IT_SERVER) {
			HashInsert(HASH_GLOBAL,"Status","CurrentServer",it->ptr,HASH_PTR);
			HashInsert(HASH_GLOBAL,"Status","CurrentList",it->owner,HASH_PTR);
		}
		else {
			HashInsert(HASH_GLOBAL,"Status","CurrentServer",NULL,HASH_PTR);
			if (it->type == IT_LIST) {
				list = it->ptr;
				HashInsert(HASH_GLOBAL,"Status","CurrentList",list,HASH_PTR);
			}
			else {
				list = NULL;
				HashInsert(HASH_GLOBAL,"Status","CurrentList",NULL,HASH_PTR);
			}
		}

		menu = CreatePopupMenu();

		if (it->type == IT_SERVER) 
			MENU_CreateServerMenu(menu,it->ptr,it->owner);
	
		else if (it->type == IT_ROOT || it->type == IT_LIST) 
			MENU_CreateListMenu(menu,it->ptr);

		else {
			DestroyMenu(menu);
			return;
		}
	}

	TrackPopupMenu(menu,TPM_LEFTALIGN|TPM_TOPALIGN,x,y,0,PROG_HWND,NULL);
	DestroyMenu(menu);
}


void CreateDesc(HWND hwnd, server_t* svr)
{
	HGLOBAL temp;
	char* str;

	if (OpenClipboard(hwnd)) {
		EmptyClipboard();
		temp = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE,256);
		str = GlobalLock(temp);
		sprintf (str, "%s (%s): %s [%d/%d players]",svr->name,svr->gametype->name,svr->ipaddr,
			HSI(svr,"stats","numplayers"),HSI(svr,"stats","maxplayers"));
		GlobalUnlock(temp);
		SetClipboardData(CF_TEXT,temp);
		CloseClipboard();
	}
	return;
}

void CopyIPToClipboard(HWND hwnd, server_t* svr)
{
	HGLOBAL temp;
	char* str;

	if (OpenClipboard(hwnd)) {
		EmptyClipboard();
		temp = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE,256);
		str = GlobalLock(temp);
		sprintf (str, "%s",svr->ipaddr);
		GlobalUnlock(temp);
		SetClipboardData(CF_TEXT,temp);
		CloseClipboard();
	}
	return;
}
					


void HWND_OnTop(HWND hwnd) {
	int top = (int) HashSelect(HASH_GLOBAL,"Settings","AlwaysOnTop",HASH_INT);
	HWND flag;

	if (top) {
		top = 0;
		flag = HWND_NOTOPMOST;
	} else {
		top = 1;
		flag = HWND_TOPMOST;
	}
	SetWindowPos(hwnd,flag,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
	HashInsert(HASH_GLOBAL,"Settings","AlwaysOnTop",(HASH_PARAM)top,HASH_INT);
}

HTREEITEM GetHitItem (void) {
	static int x, y, temp;
	TV_HITTESTINFO info;
	RECT rect;
	//HMENU menu;
	//list_t* list;
	//ITEM_TYPE* it;

	temp = GetMessagePos();
	x = LOWORD(temp);
	y = HIWORD(temp);

	GetWindowRect(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&rect);
	info.pt.x = x - rect.left;
	info.pt.y = y - rect.top;

	return TreeView_HitTest(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&info);
}

void PingAllServers(list_t* list)
{
	server_t* svr;
	int i, index;
	
	index = HashSelectInt(list,"pinglist","lastsvr");

	// Query this server and the previous 2, if necessary
	for (i=0;i<3;i++)
	{
		if (index - i < 0) 
			break;
		
		svr = HashSelectPtr(list,"servers",(HASH_PARAM)(index - i)); // breaks randomly
	
		// only ping if valid server and last send was over 100ms ago
		// ping current server or previous server if no reply was received
		// (indicated by a non-zero last-req)
		if (svr && (!i || svr->last_req) && 
			GetTickCount() - svr->last_req > SEND_TIMEOUT && !svr->lan) 
		{
			svr->last_req = GetAccTickCount();
			svr->gametype->send(HSIG("Global","Socket"),svr,svr->gametype->port);
			AddRequestCache(svr);
		}
	}

	// Query next item next time
	HashInsertInt(list,"pinglist","lastsvr",++index);


	if (index >= HashCount(list,"servers"))
	{
		EVENTS_Cancel(HashSelectPtr(list,"pinglist","event"));

		HashDelete(list,"pinglist","event");

		if (HSIG("Settings","RefreshSelList")>=0)
		{
			// Register a new one to start after the user-selected time
			// elapses
			HashInsertPtr(list,"pinglist","event",
				EVENTS_Register(PingAllServers,list,(float)HSIG("Settings","RefreshSelList"),(float)0.25));
		
			HII(list,"pinglist","lastsvr",0);
		}
		
	}

	ServerUpdate(svr);
}

	
void ListPingAllServers(list_t* list)
{
	// Don't start again if we're already going
	if (HashSelectPtr(list,"pinglist","event"))
	{
		if (HSI(list,"pinglist","lastsvr") == 0)
			EVENTS_Cancel(HSP(list,"pinglist","event"));
		else
			return;
	}

	if (strcmp(GroupOfList(list)->name,"LANServers") == 0)
		return;

	HII(list,"pinglist","lastsvr",0);
	
	HashInsertPtr(list,"pinglist","event",
		EVENTS_Register(PingAllServers,list,(float)0.25,(float)0.25));
}

void ShowNewServerList(void)
{
	LV_ITEM lvi;
	int i;
	server_t* svr;
	list_t* list;
	
	HWND hwnd = HashSelectPtr(HASH_GLOBAL,"Global","ServersLV");
	ListView_DeleteAllItems(hwnd);

	if (HSIG("Settings","DualPane") == 0)
		return;
	
	list = HashSelectPtr(HASH_GLOBAL,"Status","CurrentList");
	if (!list)
		return;

	i = 0;

	for (svr=HashReset(list,"Servers",HASH_PTR);svr;
		svr = HashNext(list,"Servers",HASH_PTR))
	{
		lvi.mask = LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE;
		if (svr->gametype->iImage) 
			lvi.iImage = svr->gametype->iImage;
		else 
			lvi.iImage = 0;
			
		//HashSelectInt(HASH_GLOBAL,"Indexes","Open");
		HashInsertInt(svr,"LV","item",i);
		lvi.iItem = (int)i;
		lvi.iSubItem = 0;
		lvi.state = lvi.stateMask = 0;
		lvi.pszText = svr->name;
		lvi.lParam = (long)svr;
		i = ListView_InsertItem(hwnd,&lvi);
		HII(svr,"lv","item",i);
		i++;
		/*if (j != i)
		{
			sprintf(buf,"%d != %d",i,j);
			MSG_DialogBox(NULL,NULL,buf,IDD_ERROR);
		}*/
	}

	HII(list,"lv","item",i);
	UI_SortList(list);
	
	if (strcmp(GroupOfList(list)->name,"LANServers") != 0)
	{
		ListPingAllServers(list);
		UI_SetStatusText("Refreshing server list.");
		UI_ProgressBar(1);
	}
	return;
}

/*#define DENOM 32000
int heightnum[3][3] = { 
	{0, 10000, 32000 }, 
	{0, 22000, 32000 },
	{0, 32000, 0} 
};
int widthnum[4] = { 0, 15000, 27000, 32000 };

int wn = 15000, wrn = 5000, hn = 22000, hn2 = 10000;*/

int paneHeight(int col, int row, int set)
{
	char buf[128];
	sprintf(buf,"PaneHeight%d%d",col,row);
	if (set != -1)
		HIIG("Settings",buf,set);

	return HSIG("Settings",buf);
}

int paneWidth(int col, int set)
{
	char buf[128];
	sprintf(buf,"PaneWidth%d",col);
	if (set != -1)
		HIIG("Settings",buf,set);

	return HSIG("Settings",buf);
}

#define DENOM (HSIG("Settings","PaneDenom"))

#define STATUS_HEIGHT 14
#define HEIGHT_TOP(y,x) (5+ (r.bottom-5)*paneHeight(y,x,-1)/DENOM)
#define WIDTH_LEFT(x) (5 + (r.right-5)*paneWidth(x,-1)/DENOM)
#define HEIGHT_BOTTOM(y,x) (HEIGHT_TOP(y,x+1)-HEIGHT_TOP(y,x) -5)
#define HEIGHT_BOTTOM2(z,x,y) (HEIGHT_TOP(z,y+1)-HEIGHT_TOP(z,x) -5)
#define WIDTH_RIGHT(x) (WIDTH_LEFT(x+1) - WIDTH_LEFT(x) - 5)
#define WIDTH_RIGHT2(x,y) (WIDTH_LEFT(y+1) - WIDTH_LEFT(x) - 5)


static int allowResizing, resizing;
static HCURSOR resizingCursor;

int WINAPI CHAT_UsersProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);

int WINAPI WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	NMHDR* hdr;
	NMLVKEYDOWN* lvkey;
	NMTVKEYDOWN* tvkey;
	NM_TREEVIEW* tv;
	NM_LISTVIEW* lv;
	LV_ITEM* lvi;
	ITEM_TYPE* it;
	list_t* list;
	server_t* svr;
	HMENU menu;
	RECT r;
	int x, y, w, i;

	if (HSIG("Status","NoWndProc"))
		return 0;

#ifdef _SQCHAT
	if (wParam == IDC_USERS)
		return CHAT_UsersProc(hwnd,msg,wParam,lParam);
#endif _SQCHAT

	if (SystrayProc(hwnd,msg,wParam,lParam) )
		return 0;

	list = HashSelect(HASH_GLOBAL,"Status","CurrentList",HASH_PTR);

	switch (msg) {
		case WM_LBUTTONDOWN:
			if (allowResizing)
			{
				SetCapture(hwnd);
				resizing = allowResizing;
				SetCursor(resizingCursor);
			}
			break;
		case WM_LBUTTONUP:
			if (allowResizing)
			{
				resizing = 0;
				ReleaseCapture();
			}
			
			if (HSIG("Status","Dragging"))
			{
				HIIG("Status","Dragging",0);
				//SetCursor( LoadCursor(NULL,MAKEINTRESOURCE(IDC_ARROW)) );
				ReleaseCapture();
			}

			break;

		case WM_MOUSEMOVE:
			if (HSIG("Status","Dragging"))
				break;

			x = LOWORD(lParam);
			y = HIWORD(lParam);
			
			GetClientRect(hwnd,&r);
			if (x>=5 && x<=r.right-5 && y>=5 && y<= r.bottom-5)
			{
				if (x>=WIDTH_RIGHT(0) && x<= WIDTH_LEFT(1))
				{
					SetCursor(resizingCursor = LoadCursor(NULL,MAKEINTRESOURCE(IDC_SIZEWE)) );
					allowResizing = 1;
				}
				else if (x>=WIDTH_LEFT(1) && y>=HEIGHT_BOTTOM(1,0) && y<=HEIGHT_TOP(1,1))
				{
					SetCursor(resizingCursor = LoadCursor(NULL,MAKEINTRESOURCE(IDC_SIZENS)) );
					allowResizing = 2;
				}
				else if (x<=WIDTH_RIGHT(0) && y>=HEIGHT_BOTTOM(0,0) && y<=HEIGHT_TOP(0,1))
				{
					SetCursor(resizingCursor = LoadCursor(NULL,MAKEINTRESOURCE(IDC_SIZENS)) );
					allowResizing = 3;
				}
				else if (x>=WIDTH_RIGHT(1) && x<=WIDTH_LEFT(2))
				{
					SetCursor(resizingCursor = LoadCursor(NULL,MAKEINTRESOURCE(IDC_SIZEWE)) );
					allowResizing = 4;
				}
			}
			else {
				if (allowResizing)
					SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_ARROW)) );
				allowResizing = 0;
			}

			if (resizing == 1)
				 paneWidth(1, DENOM * x / r.right);
			else if (resizing == 2)
				paneHeight(1,1,DENOM * y / r.bottom);
			else if (resizing == 3)
				paneHeight(0,1,DENOM * y / r.bottom);
			else if (resizing == 4)
				paneWidth(2, DENOM * x / r.right);
			else break;

		case WM_SIZE:
			GetClientRect(hwnd,&r);

			if (HSIG("Settings","DualPane") == 1)
			{
				MoveWindow(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),
					WIDTH_LEFT(0), HEIGHT_TOP(0,1),
					WIDTH_RIGHT(0),HEIGHT_BOTTOM(0,1) - STATUS_HEIGHT,
					1);
				MoveWindow(GetDlgItem(hwnd,IDC_TREE1),
					WIDTH_LEFT(0),HEIGHT_TOP(0,0),
					WIDTH_RIGHT(0),HEIGHT_BOTTOM(0,0),
					1);
				x = WIDTH_LEFT(0)-3;
				w = WIDTH_RIGHT2(0,2) + 6;

				if (HSIG("Status","ShowProgressBar"))
				{
					MoveWindow(GetDlgItem(hwnd,IDT_STATUS),
						x,HEIGHT_TOP(1,2) - 16,
						w*5/6-5,16,
						1);
					MoveWindow(GetDlgItem(hwnd,IDC_PROGRESS),
						x+w*5/6,HEIGHT_TOP(1,2) - 16,
						w/6,16,
						1);
				}
				else
				{
					MoveWindow(GetDlgItem(hwnd,IDT_STATUS),
						x,HEIGHT_TOP(1,2) - 16,
						w,16,
						1);
					MoveWindow(GetDlgItem(hwnd,IDC_PROGRESS),
						0,0,0,0,1);
				}
			}
			else {
				MoveWindow(GetDlgItem(hwnd,IDC_TREE1),
					WIDTH_LEFT(0), HEIGHT_TOP(0,0),
					WIDTH_RIGHT(0), HEIGHT_BOTTOM2(0,0,1),
					1);

				MoveWindow(GetDlgItem(hwnd,IDT_STATUS),
					0,0,
					0,0,
					1);
				MoveWindow(GetDlgItem(hwnd,IDC_PROGRESS),
					0,0,0,0,1);
			}
				
			if (HSIG("Settings","SideChat") && HSPG("Status","ChatActive"))
			{
				MoveWindow(GetDlgItem(hwnd,IDL_PLAYERINFO),
					WIDTH_LEFT(1),HEIGHT_TOP(1,0),
					WIDTH_RIGHT(1),HEIGHT_BOTTOM(1,0),
					1);
				MoveWindow(GetDlgItem(hwnd,IDL_SERVERINFO),
					WIDTH_LEFT(1), HEIGHT_TOP(1,1),
					WIDTH_RIGHT(1), HEIGHT_BOTTOM(1,1) - (HSIG("Settings","DualPane") == 1 ? STATUS_HEIGHT:0),
					1);
				MoveWindow(GetDlgItem(hwnd,IDC_USERS),
					WIDTH_LEFT(2), HEIGHT_TOP(2,0),
					WIDTH_RIGHT(2), HEIGHT_BOTTOM(2,0) - (HSIG("Settings","DualPane") == 1 ? STATUS_HEIGHT:0),
					1);
			}
			else {
				MoveWindow(GetDlgItem(hwnd,IDL_PLAYERINFO),
					WIDTH_LEFT(1), HEIGHT_TOP(1,0),
					WIDTH_RIGHT2(1,2), HEIGHT_BOTTOM(1,0),
					1);
				MoveWindow(GetDlgItem(hwnd,IDL_SERVERINFO),
					WIDTH_LEFT(1), HEIGHT_TOP(1,1),
					WIDTH_RIGHT2(1,2), HEIGHT_BOTTOM(1,1) - (HSIG("Settings","DualPane") == 1 ? STATUS_HEIGHT :0),
					1);
				MoveWindow(GetDlgItem(hwnd,IDC_USERS),0,0,0,0,1);
				i = WIDTH_RIGHT(1) + WIDTH_RIGHT(2);
				x = WIDTH_LEFT(1);

			}
			UpdateWindow(hwnd);
			break;

		case WM_INITMENU:
			// Dont bother during shutdown 
			if (HSIG("Status","Shutdown"))
				break;

			// server menu
			if ((menu = GetSubMenu((HMENU)wParam,1)) != NULL)
			{
				while (DeleteMenu(menu,0,MF_BYPOSITION))
					;
				MENU_CreateServerMenu(menu,CURRENT_SERVER,GetCurrentList());
			}
			
			if ((menu = GetSubMenu((HMENU)wParam,2)) != NULL)
			{
				while (DeleteMenu(menu,0,MF_BYPOSITION))
					;
				MENU_CreateListMenu(menu,CURRENT_LIST);
			}

			if ((menu = GetSubMenu((HMENU)wParam,3)) != NULL)
			{
				i = MF_UNCHECKED;
				if (HashSelect(HASH_GLOBAL,"Global","ServerInfo",HASH_PTR))
					i = MF_CHECKED;
				CheckMenuItem(menu,IDM_FULLSERVERINFO,i);

				i = MF_UNCHECKED;
				if (HashSelect(HASH_GLOBAL,"Settings","AlwaysOnTop",HASH_INT))
					i = MF_CHECKED;
				CheckMenuItem(menu,IDM_ONTOP,i);
			}
			/*if ((menu = GetSubMenu((HMENU)wParam,4)) != NULL)
			{
				while (DeleteMenu(menu,0,MF_BYPOSITION))
					;
				
				AppendMenu(menu,MF_STRING,IDM_REGWEB,"Registration Website");
				if (HSPG("Status","ChatActive"))
					AppendMenu(menu,MF_STRING,IDM_CHAT_LOGIN,"Log out...");
				else AppendMenu(menu,MF_STRING,IDM_CHAT_LOGIN,"Log in...");
			}*/
			break;

		case WM_NOTIFY:
			hdr = (NMHDR*)lParam;
			if (hdr->code == LVN_BEGINDRAG)
			{
				HIIG("Status","Dragging",1);
				SetCapture(PROG_HWND);
				SetCursor( LoadCursor(NULL,MAKEINTRESOURCE(IDC_NO)) );
			}
			if (hdr->code == LVN_GETDISPINFO)
			{
				if (hdr->idFrom == IDL_PLAYERINFO)
				{
					lvi = (LV_ITEM*)(hdr + 1);
					if (lvi->iSubItem == 1)
						strcpy(lvi->pszText,HashSelectText(HASH_GLOBAL,"PlayerinfoPing",(HASH_PARAM)lvi->iItem));
					else strcpy(lvi->pszText,HashSelectText(HASH_GLOBAL,"PlayerinfoName",(HASH_PARAM)lvi->iItem));
					return 0;
				}
				if (hdr->idFrom == IDL_SERVERS)
				{
					lvi = (LV_ITEM*)(hdr + 1);
					svr = (server_t*)lvi->lParam;
					if (svr) {
						if (lvi->iSubItem == 0)
							sprintf(lvi->pszText,"%s",svr->name);
						if (lvi->iSubItem == 1)
							sprintf(lvi->pszText,"%d / %d",HashSelectInt(svr,"stats","numplayers"),HashSelectInt(svr,"stats","maxplayers"));
						if (lvi->iSubItem == 2)
							sprintf(lvi->pszText,"%d",svr->ping*5/6);
						if (lvi->iSubItem == 3)
							sprintf(lvi->pszText,"%d",svr->avg_ping*5/6);
						if (lvi->iSubItem == 4)
							sprintf(lvi->pszText,"%s",HashSelectText(svr,"stats","map"));
						if (lvi->iSubItem == 5)
							sprintf(lvi->pszText,"%s",svr->gametype->name);

					}					
					return 0;
				}
				if (hdr->idFrom == IDL_SERVERINFO)
				{
					lvi = (LV_ITEM*)(hdr + 1);
					strcpy(lvi->pszText,HashSelectText(HASH_GLOBAL,"ServerInfo",(HASH_PARAM)lvi->iItem));
					if (lvi->iItem == 5)
						strcpy(lvi->pszText,HashSelectText(HASH_GLOBAL,"Status","Ping") );

					return 0;
				}
			}
			if (hdr->idFrom == IDL_SERVERS && hdr->code == LVN_COLUMNCLICK)
			{
				lv = (NM_LISTVIEW*)lParam;
				HIIG("Settings","ServersLV-SortType",lv->iSubItem+1);
				//ShowNewServerList();
				UI_SortList(CURRENT_LIST);

			}
			if (hdr->idFrom == IDL_SERVERS && hdr->code == LVN_ITEMCHANGED)
			{
				if (HSIG("Status","IgnoreItemChange"))
					return 0;

				lv = (NM_LISTVIEW*)lParam;
				SET_CURRENT_SERVER((server_t*)lv->lParam);

				Refresh(1);

				if (!HashSelectInt(HASH_GLOBAL,"Settings","PingAllNetServers")) 
				{ // do an update now if we aren't constantly updating
					CURRENT_SERVER->timeout = 0;
					WRITE_DEBUG("Item changed, requesting new data");
					// Schedule an immediate update
					HIIG("Status","NextRefresh",GetTickCount()+200);
					WRITE_DEBUG("END: Item changed, requesting new data");
				}
				

				return 0;
			}
			if (hdr->idFrom == IDC_TREE1 && hdr->code == TVN_ITEMEXPANDED)
			{
				tv = (NM_TREEVIEW*)lParam;
				tv->itemNew.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE;
				if (tv->action == 2) {
					tv->itemNew.iImage = HashSelectInt(NULL,"Indexes","Open");
					tv->itemNew.iSelectedImage = HashSelectInt(NULL,"Indexes","Open");
				}
				else {
					tv->itemNew.iImage = HashSelectInt(NULL,"Indexes","Closed");
					tv->itemNew.iSelectedImage = HashSelectInt(NULL,"Indexes","Closed");
				}
				TreeView_SetItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&tv->itemNew);
			}
			if (hdr->code == TVN_KEYDOWN || hdr->code == LVN_KEYDOWN)
			{
				tvkey = (NMTVKEYDOWN*)lParam;
				if (tvkey->wVKey == VK_F5)
				{
					if (list && GroupOfList(list)->refresh_list)
						GroupOfList(list)->refresh_list(list); 
				}
				if (tvkey->wVKey == VK_DELETE && hdr->idFrom == IDC_TREE1)
				{
					if (!HSIG("Settings","DualPane") && current_server_info.hItem == GetHitItem())
						SendMessage(hwnd,WM_COMMAND,SCM_DELETE,0);
					else if (CURRENT_LIST)
						SendMessage(hwnd,WM_COMMAND,SCM_DELLIST,0);
				}

				lvkey = (NMLVKEYDOWN*)lParam;
				if (lvkey->wVKey == VK_DELETE && hdr->idFrom == IDL_SERVERS)
				{
						SendMessage(hwnd,WM_COMMAND,SCM_DELETE,0);
				}

				if ((lvkey->wVKey == VK_INSERT || tvkey->wVKey == VK_INSERT) && list && GroupOfList(list)->add_server)
					GroupOfList(list)->add_server(list);
			}


			if (hdr->code == NM_RCLICK)
			{
				if (hdr->idFrom == IDC_TREE1)
				{
					hdr->code = TVN_SELCHANGED;
					//SendMessage(hwnd,msg,wParam,lParam);
					DisplaySCM(NULL);
				}
				else if (hdr->idFrom == IDL_SERVERS)
					DisplaySCM(CURRENT_SERVER);
			}
			if ((hdr->idFrom == IDC_TREE1 || hdr->idFrom == IDL_SERVERS)
				&& hdr->code == NM_DBLCLK)
			{
				if (hdr->idFrom == IDL_SERVERS ||
						current_server_info.hItem == GetHitItem())
					ConnectServer(CURRENT_SERVER);
			}
			if (hdr->idFrom == IDC_TREE1 && hdr->code == TVN_SELCHANGED && !HSIG("Status","shutdown"))
			{
				tv = (NM_TREEVIEW*)lParam;

				current_server_info.mask = TVIF_PARAM | TVIF_CHILDREN;
				current_server_info.hItem = tv->itemNew.hItem;
				TreeView_GetItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&current_server_info);
				it = (ITEM_TYPE*)current_server_info.lParam;

				if (it->type == IT_SERVER) {
					SET_CURRENT_SERVER(it->ptr);
					// User selected a different server 
					if (!HashSelectInt(HASH_GLOBAL,"Settings","PingAllNetServers")) 
					{ // do an update now if we aren't constantly updating
						CURRENT_SERVER->timeout = 0;
						WRITE_DEBUG("Item changed on tree, requesting new data");
						RequestNewData(0);
						WRITE_DEBUG("END: Item changed on tree, requesting new data");
					}
					HashInsertPtr(HASH_GLOBAL,"Status","CurrentList",it->owner);
					Refresh(1);
				}
				else if (it->type == IT_ROOT || IT_LIST) {
					SET_CURRENT_SERVER(NULL);
					HashInsert(HASH_GLOBAL,"Status","CurrentList",it->ptr,HASH_PTR);
	
					if (HashSelectInt(HASH_GLOBAL,"Status","IgnoreListChange") == 1)
						HashInsertInt(HASH_GLOBAL,"Status","IgnoreListChange",0);
					else 
						ShowNewServerList();
				}
			}
			break;
/*		case WM_CHAR:
			if (wParam == VK_DELETE && GetActiveWindow() == GetDlgItem(PROG_HWND,IDL_SERVERS))
			{
				GroupOfList(CURRENT_LIST)->
					delete_server(CURRENT_SERVER,CURRENT_LIST); 
			}
			break;
*/


				
		case WM_COMMAND:
			switch(wParam) {
				case IDM_ONLINELISTS:
					ONLINELISTS_Download((void*)1);
					return 0;
#ifdef _SQCHAT
				case IDM_CHAT_LOGIN:
					CHAT_Login();
					return 0;
#endif
				case IDM_FULLSERVERINFO:
					if (HashSelect(HASH_GLOBAL,"Global","ServerInfo",HASH_PTR))
						SendMessage(HashSelect(HASH_GLOBAL,"Global","ServerInfo",HASH_PTR),WM_CLOSE,0,0);
					else 
						SERVERINFO_DialogBox();
					return 0;
				case IDM_BROWSERWEB:
					ShellExecute(NULL,NULL,IDS_BROWSERURL,NULL,NULL,0);
					return 0;
				case IDM_QGLWEB:
					ShellExecute(NULL,NULL,"http://www.qgl.org/",NULL,NULL,0);
					return 0;
				case IDM_REGWEB:
					ShellExecute(NULL,NULL,"http://register.ausgamers.com/",NULL,NULL,0);
					return 0;
				case IDM_GCWEB:
					ShellExecute(NULL,NULL,"http://gamecreate.qgl.org/",NULL,NULL,0);
					return 0;

				case IDM_COMM_WEBSITE: ShellExecute(NULL,NULL,"http://games.telstra.com/gamearena/",NULL,NULL,0); return 0;
				case IDM_COMM_FORUM: ShellExecute(NULL,NULL,"http://games.telstra.com/gamearena/community/messageboards/",NULL,NULL,0); return 0;
				case IDM_COMM_GAMECREATE: ShellExecute(NULL,NULL,"http://games.telstra.com/gamearena/services/gamecreate/",NULL,NULL,0); return 0;
				case IDM_COMM_LADDERS: ShellExecute(NULL,NULL,"http://games.telstra.com/gamearena/services/ladders/",NULL,NULL,0); return 0;
				case IDM_COMM_FILES: ShellExecute(NULL,NULL,"http://games.telstra.com/gamearena/resources/files/",NULL,NULL,0); return 0;
				case IDM_COMM_STATS: ShellExecute(NULL,NULL,"http://games.telstra.com/gamearena/services/stats/",NULL,NULL,0); return 0;
				
				
				case IDM_OPENFILE:
					DISK_OpenFile();
					return 0;
				case IDM_SAVEFILE:
					DISK_SaveFile();
					return 0;
				case SCM_PARAM:
					if (PARAM_DialogBox())
						ConnectServer(CURRENT_SERVER);
					return 0;
				case SCM_PASSWORD:
					if (! PASSWORD_DialogBox(PROG_HINSTANCE,hwnd,CURRENT_SERVER) )
						return 0;
				case IDM_CONNECT:
				case IDB_CONNECT:
				case SCM_CONNECT:
					ConnectServer(CURRENT_SERVER); return 0;
				case SCM_WAITFREE:
					HashInsertPtr(HASH_GLOBAL,"Status","WaitingServer",CURRENT_SERVER); return 0;
				case IDM_ONTOP:
					HWND_OnTop(hwnd); return 0;
				case IDM_CVARS:
				case SCM_CVARS:
					CVARS_DialogBox(CURRENT_SERVER); return 0;
				case IDM_GENERAL:
					GENCFG_DialogBox(); return 0;
				case IDM_PROGCFG:
					PROGCFG_DialogBox(); return 0;
				case IDM_BUDDY:
					BUDDY_DialogBox(); return 0;
				case SCM_CREATELIST:
					CREATELIST_DialogBox(NULL,hwnd); return 0;
				case SCM_DELLIST:
					if (list && GroupOfList(list)->delete_list)
						GroupOfList(list)->delete_list(list); return 0;
				case SCM_RENAMELIST:
					GroupOfList(list)->properties_list(list); return 0;
				case SCM_REFRESHLIST:
					GroupOfList(list)->refresh_list(list); return 0;
				case SCM_LISTMVUP:
					ListMoveUp(list); return 0;
				case SCM_LISTMVDOWN:
					ListMoveDown(list); return 0;

				case SCM_EXPORTQSTAT:
					DISK_ExportQstatList(list); break;
				case SCM_ADDSERVER:
					GroupOfList(list)->add_server(list); break;
				case SCM_PASTE:
					GroupOfList(list)->paste_server(list); break;
				case SCM_PROPERTIES:
					PROPERTIES_DialogBox(NULL,hwnd,CURRENT_SERVER,0); break;
				case SCM_DELETE:
					if (GroupOfList(CURRENT_LIST)->delete_server)
						GroupOfList(CURRENT_LIST)->
							delete_server(CURRENT_SERVER,CURRENT_LIST); break;
				case SCM_COPYIP:
					CopyIPToClipboard(hwnd,CURRENT_SERVER); break;
				case SCM_COPYDESC:
					CreateDesc(hwnd,CURRENT_SERVER); break;
				case SCM_ADDFAVOURITES:
					GroupOfList(CURRENT_LIST)->add_favourites(CURRENT_SERVER); break;
				case SCM_CUT:
					GroupOfList(CURRENT_LIST)->
						cut_server(CURRENT_SERVER); break;
				case SCM_COPY:
					GroupOfList(CURRENT_LIST)->
						copy_server(CURRENT_SERVER); break;
				case IDM_ABOUT:
					ABOUT_DialogBox(NULL,hwnd); return 0;
				case IDM_EXIT:
					PostQuitMessage(0); 
					return 0;
			}
			break;
		case WM_CLOSE:
			PostQuitMessage(0);
			return 0;
	}		
	return 0;
}

server_t* DDE_ParseServer(char* type, char* a)
{
	server_t* svr;
	char* str, *addr;
	char* ptr;
	char* mod = NULL;
	types_t* t;

	str = addr = strdup(a);

	ptr = strchr(addr,'/');
	if (ptr)
	{
		addr = ptr + 1;
		while (*addr == '/')
			addr++;
		ptr = addr;
		while (*ptr && *ptr != '/')
			ptr ++;

		if (*ptr)
		{
			*ptr = 0;
			mod = ptr+1;
		}
	}

	// addr string needs massaging
	svr = NewServer("New Server",addr,GetServerType_qstat(type)->qstat_type);
	t = GetServerType_qstat(type);
	if (mod)
		HIT(svr,"items",GetServerType_qstat(type)->mod_cvar,mod);

	free(str);

	return svr;
}

int DDE_AddFavourites(char* type, char* addr)
{
	server_t* svr;
	svr = DDE_ParseServer(type,addr);

	ServerAddFavourites(svr);
	FreeServer(svr);

	return TRUE;
}

int DDE_LaunchServer(char* type, char* addr)
{
	server_t* svr;
	svr = DDE_ParseServer(type,addr);

	SET_CURRENT_SERVER(svr);
	return ConnectServer(svr);
}

int DDE_PromptUser(char* type, char* addr)
{
	server_t* svr;
	int x;

	svr = DDE_ParseServer(type,addr);

	x = SCHEME_DialogBox(type,svr->ipaddr);
	FreeServer(svr);

	return x;
}

int DDE_CheckForCommand(void)
{
	char* command;
	char* ptr;
	char* end;
	char* next;
	char* func;
	char* args[10];
	char delim;
	int num_args = 0;
	int do_quit = 1, action;
	int quoted = 0;

	command = ptr = HSTG("status","ddecommand");
	
	if (!*ptr) return 1;

	
	// skip through the string until the end
	for (;;)
	{
		if (!*command)
			break;

		// skip through to the next ] not in a string
		next = command;
		while ((next = strpbrk(next,"\"]")) != NULL)
		{
			if (*next == '"')
			{
				if (*(next-1) != '\\') 
					quoted = !quoted;
			}
			if (*next == ']' && !quoted)
				break;
			next++;
		}

		if (!next)
			break;

		if (*command == '[') command++;
		ptr = command; command = next;
		*command = 0; command++; // erase and skip the ]

	
		// command show now be of the form
		// foobar("1","2",etc);
		// we dont support more than two parameters
		func = ptr;

		ptr = strchr(ptr,'('); // move to parameters
		if (!ptr) break; // invalid
		*ptr = 0; ptr ++; // end the command
		delim = *ptr;
		num_args = 0;
		// run through until end
		while (delim && delim != ')')
		{
			if (*ptr == '"')
			{
				ptr++;
				end = ptr;
				while ((end = strpbrk(end,"\"")) != NULL)
				{
					// ignore backslashed quotes
					if (*(end-1) == '\\') { end++; continue; }
					// if broken, just chop off the last character
					if (*end == '"') { *end = 0; end++; }
					break;
				}
			}
			else 
			{
				end = strpbrk(ptr,",)");
				
			}
			args[num_args++] = ptr;

			if (!end) end = ptr + strlen(ptr);
			delim = *end;
			*end = 0;
				
			ptr = end + 1;
		}

		// do something based on this command
		if (SQstricmp(func,"server") == 0 && num_args == 2 && strstr(args[1],"//"))
		{
			action = HSIG("Settings","SchemeAction");
			//if (action == 2)
				action = DDE_PromptUser(args[0],args[1]);
			
			if (action == 1) DDE_AddFavourites(args[0],args[1]); 
			if (action == 0) do_quit = DDE_LaunchServer(args[0],args[1]);
		}
		if (SQstricmp(func,"quit") == 0 && num_args == 0)
		{
			if (!do_quit)
			{
				ShowWindow(PROG_HWND,SW_SHOW);
				return 1;
			}
			else
				return 0;
		}
	}

	
	
	HashDelete(HASH_GLOBAL,"status","ddecommand");
	if (HSIG("status","startmin"))
		ShowWindow(PROG_HWND,SW_SHOW);
	return 1;
}


HDDEDATA CALLBACK DdeCallbackFunc(
  UINT uType,     // transaction type
  UINT uFmt,      // clipboard data format
  HCONV hconv,    // handle to the conversation
  HSZ hsz1,       // handle to a string
  HSZ hsz2,       // handle to a string
  HDDEDATA hdata, // handle to a global memory object
  DWORD dwData1,  // transaction-specific data
  DWORD dwData2   // transaction-specific data
)
{
	char command[512];

	// allow connection
	if (uType == XTYP_CONNECT)
		return (HDDEDATA)1;

	// ignore confirmation
	if (uType == XTYP_CONNECT_CONFIRM)
		return 0;
	
	if (uType == XTYP_EXECUTE)
	{
		DdeGetData(hdata,command,500,0);
		HITG("Status","DDECommand",command);
		return (HDDEDATA)DDE_FACK;
	}

	return 0;
}
 

// ***
// Create a socket, and allow it to broadcast.
// ***
SOCKET InitSocket(void)
{
	SOCKET s;
	int i = 1;
	
	s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char*) &i, sizeof(int));
	i = 1;
	ioctlsocket(s,FIONBIO,&i); // don't block on send
	
	return s;
}

void checkSocket(void)
{
	FD_SET wfds;
	struct timeval tv;
	int i;

	SOCKET s = HSIG("Global","Socket");

	FD_ZERO(&wfds);
	FD_SET(s,&wfds);

	tv.tv_sec = 0;
	tv.tv_usec = 20;

	//do {
		i = select(0,NULL,&wfds,NULL,&tv);
		if (!i) {
			//MSG_DialogBox(NULL,NULL,"Writing unavailable",IDD_ERROR);
			WRITE_DEBUG("Writing unavailable, creating new socket");
			CloseSocket(s);
			s = InitSocket();
			HIIG("Global","Socket",s);
		}
	//} while (!i);
}





/* Sub-classed Treeview handler.
   If requested, this handler will translate mouse wheel rotations from
   the usual list-scrolling into selecting the next or previous item
   on the list, expanding and collapsing the list as necessary. 
*/
WNDPROC oldTVProc;
int WINAPI HookedTVProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	HTREEITEM item;
//	char buf[128];

	/*sprintf(buf,"Message %d sent to treeview",msg);
	WRITE_DEBUG(buf);
	*/

	if (HashSelectInt(NULL,"Settings","MouseWheel") == 0)
		return CallWindowProc(oldTVProc,hwnd,msg,wparam,lparam);
	
	if (msg == 522) {
		item = TreeView_GetSelection(hwnd);

		
		if ((short) HIWORD(wparam) < 0) {
			TreeView_Expand(hwnd,item,TVE_EXPAND);
			item = TreeView_GetNextVisible(hwnd,item);
		}
		else  {
			TreeView_Expand(hwnd,item,TVE_COLLAPSE);
			item = TreeView_GetPrevVisible(hwnd,item);
		}
		if (item != NULL)
			TreeView_SelectItem(hwnd, item);
		
		return 0;
	}

	return CallWindowProc(oldTVProc,hwnd,msg,wparam,lparam);
}


HWND CreateMainWindow(HWND parent) {
	HWND hwnd;
	LV_COLUMN lvc;
	//char title[128];

	if ((hwnd = HashSelectPtr(HASH_GLOBAL,"Global","MainWnd")) != NULL) {
		SaveWindowPosition(hwnd);
		DestroyWindow(hwnd);
	}
	HashInsertPtr(HASH_GLOBAL,"Global","MainWnd",CreateDialog(PROG_HINSTANCE, MAKEINTRESOURCE(IDD_MAIN), parent, WndProc));
	SetWindowText(PROG_HWND,IDS_TITLE);



	DeleteMenu(GetMenu(PROG_HWND),MF_BYPOSITION,0);
	DrawMenuBar(PROG_HWND);


	SetClassLong(PROG_HWND, GCL_HICON, (LONG) LoadIcon(PROG_HINSTANCE, MAKEINTRESOURCE(IDI_ICON2))   ); 

	// Save the handle to the treeview that contains
	// the various servers
	HashInsert(HASH_GLOBAL,"Global","ServersTree", GetDlgItem(PROG_HWND, IDC_TREE1) , HASH_PTR);
	
	(long) oldTVProc = SetWindowLong(HashSelectPtr(HASH_GLOBAL,"Global","ServersTree"), 
		GWL_WNDPROC, (long)HookedTVProc);

	HashInsertPtr(HASH_GLOBAL,"Global","ServerInfoLV",GetDlgItem(PROG_HWND,IDL_SERVERINFO));
	memset(&lvc,0,sizeof(LV_COLUMN));
	lvc.mask = LVCF_WIDTH | LVCF_TEXT;
	lvc.cx = 70;
	lvc.pszText = "Item";
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServerInfoLV"),1,&lvc);

	lvc.pszText = "Setting";
	lvc.cx = 130;
	lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 1;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServerInfoLV"),2,&lvc);

	HashInsertPtr(HASH_GLOBAL,"Global","PlayerInfoLV",GetDlgItem(PROG_HWND,IDL_PLAYERINFO));
	memset(&lvc,0,sizeof(LV_COLUMN));
	lvc.mask = LVCF_WIDTH | LVCF_TEXT;
	lvc.cx = 50;
	lvc.pszText = "Kills";
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","PlayerInfoLV"),1,&lvc);

	lvc.pszText = "Ping";
	lvc.cx = 50;
	lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 1;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","PlayerInfoLV"),2,&lvc);

	lvc.pszText = "Player Name";
	lvc.cx = 100;
	//lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 2;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","PlayerInfoLV"),3,&lvc);

  if (HSIG("Settings","DualPane") == 1)
  {
	HashInsertPtr(HASH_GLOBAL,"Global","ServersLV",GetDlgItem(PROG_HWND,IDL_SERVERS));
	memset(&lvc,0,sizeof(LV_COLUMN));
	lvc.mask = LVCF_WIDTH | LVCF_TEXT;
	lvc.cx = 120;
	lvc.pszText = "Server";
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),1,&lvc);

	lvc.pszText = "Players";
	lvc.cx = 50;
	lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 1;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),2,&lvc);

	lvc.pszText = "Ping";
	lvc.cx = 40;
	//lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 2;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),3,&lvc);

	lvc.pszText = "Avg Ping";
	lvc.cx = 40;
	//lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 3;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),4,&lvc);

	lvc.pszText = "Map";
	lvc.cx = 50;
	//lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 4;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),5,&lvc);
	
	lvc.pszText = "Game";
	lvc.cx = 50;
	//lvc.mask |= LVCF_SUBITEM;
	lvc.iSubItem = 5;
	ListView_InsertColumn(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),6,&lvc);
	
	ListView_SetImageList(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"), 
		HashSelectPtr(NULL,"Global","ImageList"),LVSIL_SMALL);	
	ListView_SetImageList(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"), 
		HashSelectPtr(NULL,"Global","StateList"),LVSIL_STATE);	
  }
	else {
		SetWindowPos(GetDlgItem(PROG_HWND,IDL_SERVERS),HWND_BOTTOM,
			0,0,0,0,0);
		ShowWindow(GetDlgItem(PROG_HWND,IDL_SERVERS),SW_HIDE);
		DestroyWindow(GetDlgItem(PROG_HWND,IDL_SERVERS));
	}
	
	return PROG_HWND;
}


void setWorkingDir(void) {
	char str[512];
	char* ptr;
	int i;
	strcpy(str,GetCommandLine());

	if (str[0] == '"') {
		ptr = str+1;
		for (i=1;str[i]!='"' && str[i];i++)
			;
		str[i] = 0;
	}
	else {
		ptr = str;
		for (i=0;str[i] != ' ' && str[i];i++)
			;
		str[i] = 0;
	}
	
	for (i=strlen(ptr);i>=0;i--)
		if (ptr[i] == '/' || ptr[i] == '\\')
			break;
	if (i>=0)
		ptr[i] = 0;

	_chdir(ptr);
}

// There should be a neater way to do this
void RespawnWindow(void)
{

	HIIG("Status","Systray",-1);
	SystrayUpdate();
	ServersListUpdate(0);
	ListRebuildTreeview();
}

int accessInternet(unsigned s)
{
	SOCKADDR_IN addr;
	int r;
	
	addr.sin_family = AF_INET;
	addr.sin_port = htons(27501);
	addr.sin_addr.s_addr = inet_addr("202.139.235.147");

	r = sendto(s,"test",4,0,(SOCKADDR*)&addr,sizeof(SOCKADDR_IN));
	return (r>0);
}
extern int allowNSLookups; // comm.c
extern int DEBUG_ShowMessages;



// ***
// Initial entry point for Server Query
// Performs all initialization code and
// runs the main execution loop.
// ***
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR cmdline, int show)
{
	HWND temp;
	MSG msg;
	FILE* tmp;
	//int tabstops[10];
	SOCKET s;
	list_t* list;
	DWORD ddeinit;
	HSZ appname;
	int running;



	
	//Make sure working directory is the location of query.exe
	setWorkingDir();

	//Start log
	WRITE_DEBUG("date");

	//remove legacy files
	remove("game/kingpin.dll");
	remove("game/sin.dll");
	remove("*.dll");

	// if not starting multiple windows, look for an old one
	if (!strstr(cmdline,"/MULTIPLE") && !strstr(cmdline,"/multiple"))
	{
		temp = FindWindow(0,IDS_TITLE);
		if (temp)
		{
			SetForegroundWindow(temp);
			return 0;
		}
	}

	// disable domain name look ups on LAN (for machines that lock)
	if (strstr(cmdline,"/lan") || strstr(cmdline,"/LAN"))
		allowNSLookups = 0;

	if (strstr(cmdline,"/net") || strstr(cmdline,"/NET"))
		allowNSLookups = 2;			

	// Initialise hash stuff
	WRITE_DEBUG("Init: hashes");
	HashInit(G_Hashes,&G_NumHashes);

	// Start up Winsock
	WRITE_DEBUG("Init: winsock");
	if (initWinSock() != 0) {
		MSG_DialogBox(NULL,NULL,"Server Query was unable to initialize a suitable version of WinSock",IDD_ERROR);
		return 0;
	}

	// Exit if socket failed to initialise
	WRITE_DEBUG("Init: socket");
	if ((s = InitSocket()) <= 0)
	{
		MSG_DialogBox(NULL,NULL,"Server Query was unable to initialize a UDP socket",IDD_ERROR);
		return 0;
	}

	// Start DDE
	WRITE_DEBUG("Init: DDE");
	ddeinit = 0;
	if (DdeInitialize(&ddeinit,DdeCallbackFunc,
			APPCMD_FILTERINITS /* | 
			(CBF_FAIL_ALLSVRXACTIONS & (~CBF_FAIL_EXECUTES) )*/,0) != 0)
	{
		MSG_DialogBox(NULL,NULL,"Could not initialize DDE",IDD_ERROR);
		return 0;
	}
	appname = DdeCreateStringHandle(ddeinit,"query",CP_WINANSI);
	DdeNameService(ddeinit, appname, 0, DNS_REGISTER);
	
	

	// test for access to net
	if (!accessInternet(s) && allowNSLookups == 1)
	{
		//MSG_DialogBox(NULL,NULL,"Server Query has detected that internet access is unavailable.  For stability reasons, domainname lookups have been disabled for the duration of this session.",IDD_ERROR);
		allowNSLookups = 0;
	}

	// test for crash on last run
	if ((tmp = fopen("sq.tmp","rt")) != NULL) {
		fclose(tmp);
		/*
		if (allowNSLookups)
			MSG_DialogBox(NULL,NULL,"Server Query has detected that it locked up when last run.  To try and prevent this from re-occuring, Server Query has temporarily disabled name resolution.\n\nYou can also achieve this from the command line by running\n\n x:\\path\\query.exe /LAN",IDD_ERROR);
		allowNSLookups = 0;
		*/
	}

	if (GetAsyncKeyState(VK_SHIFT) & 0x8000)
		allowNSLookups = 0;		

	if (GetAsyncKeyState(VK_CONTROL) & 0x8000)
		DEBUG_ShowMessages = 1;

	if (allowNSLookups == 2)
		allowNSLookups = 1;			

	// write sq.tmp; remove at end of execution
	if ((tmp = fopen("sq.tmp","wt")) != NULL) {
		fprintf(tmp,"SQ");
		fclose(tmp);
	}

	// For random client-ports, used primarily for the Q2
	// connect code
	WRITE_DEBUG("Init: random seed");
	srand( (unsigned)time( NULL ) );

	
	WRITE_DEBUG("Init: common contols");
	// Common controls
	InitCommonControls();

	WRITE_DEBUG("Init: default settings");
	
	if (strstr(cmdline,"/DEBUG") || strstr(cmdline,"/debug"))
		DEBUG_ShowMessages = 1;

	// The global hash table holds the values
	// that won't change during execution
	HashInsertPtr(HASH_GLOBAL,"Funcs","malloc",malloc);
	HashInsertPtr(HASH_GLOBAL,"Funcs","calloc",calloc);
	HashInsertPtr(HASH_GLOBAL,"Funcs","realloc",realloc);

	HashInsertInt(HASH_GLOBAL,"Global","Socket",s);
	HashInsertInt(HASH_GLOBAL,"Global","DDE",ddeinit);
	HashInsertPtr(HASH_GLOBAL,"Global","Hinstance",hInstance);
	HashInsertText(HASH_GLOBAL,"Global","Cmdline",cmdline);
	//HashAlloc(HASH_GLOBAL,"ServerInfo",2);

	// The settings hash table is stored on disk
	// These provide reasonable defaults
	HashInsertText(HASH_GLOBAL,"Settings","ClanId","");
	HashInsertInt(HASH_GLOBAL,"Settings","Refresh",2);
	HashInsertInt(HASH_GLOBAL,"Settings","Timeout",5);
	HashInsertInt(HASH_GLOBAL,"Settings","LANRefresh",15);
	HashInsertInt(HASH_GLOBAL,"Settings","LANTimeout",30);
	
	//HashInsertInt(HASH_GLOBAL,"Settings","NoLAN",0);
	HashInsertInt(HASH_GLOBAL,"Settings","ShowGameTypeOnLan",0);
	HashInsertInt(HASH_GLOBAL,"Settings","PingAllNetServers",0);
	HashInsertInt(HASH_GLOBAL,"Settings","ShowPing",1);
	HashInsertInt(HASH_GLOBAL,"Settings","ShowPupServers",1);
	HashInsertInt(HASH_GLOBAL,"Settings","RefreshPupServers",1);
	HashInsertInt(HASH_GLOBAL,"Settings","PingWhileBackground",0);
	HashInsertInt(HASH_GLOBAL,"Settings","ChatStart",1);
	HashInsertInt(HASH_GLOBAL,"Settings","DualPane",1);
	HashInsertInt(HASH_GLOBAL,"Settings","UpdateOnlineLists",1);
	HashInsertInt(HASH_GLOBAL,"Settings","UpdateNames",0);
	HashInsertInt(HASH_GLOBAL,"Settings","ExtraServInfo",0);
	HashInsertInt(HASH_GLOBAL,"Settings","EnableSchemes",1);
	HashInsertInt(HASH_GLOBAL,"Settings","SchemeAction",2); // prompt
#ifdef LISTALLSERVERS
	HashInsertInt(HASH_GLOBAL,"Settings","ListAllServers",1);
#else
	HashInsertInt(HASH_GLOBAL,"Settings","ListAllServers",0);
#endif
	HashInsertInt(HASH_GLOBAL,"Settings","RefreshSelList",60);
#ifndef NOLANLIST	
	HashInsertInt(HASH_GLOBAL,"Settings","ShowLANList",1);
#endif

	// Default pane widths and heights
	HIIG("Settings","PaneHeight00",0);
	HIIG("Settings","PaneHeight01",17000);
	HIIG("Settings","PaneHeight02",32000);
	HIIG("Settings","PaneHeight10",0);
	HIIG("Settings","PaneHeight11",22000);
	HIIG("Settings","PaneHeight12",32000);
	HIIG("Settings","PaneHeight20",0);
	HIIG("Settings","PaneHeight21",32000);
	HIIG("Settings","PaneHeight22",32000);

	HIIG("Settings","PaneWidth0",0);
	HIIG("Settings","PaneWidth1",20500);
	HIIG("Settings","PaneWidth2",27000);
	HIIG("Settings","PaneWidth3",32000);
	HIIG("Settings","PaneDenom",32000);

	// Put 'exported' functions into a hash
	HashInsertPtr(HASH_GLOBAL,"Funcs","GetServerPort",GetServerPort);
	HashInsertPtr(HASH_GLOBAL,"Funcs","GetServerAddr",GetServerAddr);
	HashInsertPtr(HASH_GLOBAL,"Funcs","ServersListUpdate",ServersListUpdate);
	HashInsertPtr(HASH_GLOBAL,"Funcs","AddServer",AddServer);
	HashInsertPtr(HASH_GLOBAL,"Funcs","NewServer",NewServer);
	HashInsertPtr(HASH_GLOBAL,"Funcs","DeleteServer",DeleteServer);
	HashInsertPtr(HASH_GLOBAL,"Funcs","EmptyList",EmptyList);
	HashInsertPtr(HASH_GLOBAL,"Funcs","HashAlloc",HashAlloc);
	
	HashInsertPtr(HASH_GLOBAL,"Funcs","GetServerType",GetServerType);
	HashInsertPtr(HASH_GLOBAL,"Funcs","GetServerType_qstat",GetServerType_qstat);
	HashInsertPtr(HASH_GLOBAL,"Funcs","ListNamedWithParent",ListNamedWithParent);
	HashInsertPtr(HASH_GLOBAL,"Funcs","CreateList",CreateList);
	HashInsertPtr(HASH_GLOBAL,"Funcs","GroupOfList",GroupOfList);
	HashInsertPtr(HASH_GLOBAL,"Funcs","HTTP_Download",HTTP_Download);
	HashInsertPtr(HASH_GLOBAL,"Funcs","EVENTS_Register",EVENTS_Register);
	HashInsertPtr(HASH_GLOBAL,"Funcs","ListPingAllServers",ListPingAllServers);
	HashInsertPtr(HASH_GLOBAL,"Funcs","GetAccTickCount",GetAccTickCount);
	HashInsertPtr(HASH_GLOBAL,"Funcs","RefreshLANServers",RefreshLANServers);
	FUNC_Export();
	

	// Load the various DLL's related to
	// processing the different types of servers
	WRITE_DEBUG("Init: Game DLLs");
	RegisterGameTypes();

	WRITE_DEBUG("Init: Treeview Image List");
	TreeviewImageList("ImageList");
	StateImageList("StateList");
	
	// Register the available server list types
	WRITE_DEBUG("Init: group types");
	RegisterGroupTypes();

	WRITE_DEBUG("Check: default.dll available");
	if (!GroupNamed("NetServers")) {
		MSG_DialogBox(PROG_HINSTANCE,NULL,"Unable to load 'list/default.dll'.  SQ halted.\n\nIn all likelihood, you installed from a .zip file without enabling 'use folder names'.",IDD_ERROR);
		exit(0);
	}

	// Create the default lists
	WRITE_DEBUG("Init: default lists");
	GetAllServersList(); // create this list
	CreateList("LAN","LANServers",NULL);
	CreateList("Favourites","Favourites",NULL);
	//CreateList("All","ListAllServers",NULL);

	// Load settings hash, paths to various
	// games and finally the server list itself.
	WRITE_DEBUG("Load: stored settings");
	LoadFromDisk();

	// Set up schemes
	SCHEME_Update();	
	if (strstr(cmdline,"/schemes") || strstr(cmdline,"/SCHEMES"))
	{
		SCHEME_AddEntries();
		ShutDown();
		return 0;
	}



	WRITE_DEBUG("Init: recv thread");
	RecvThreadInit();

	// Create the window and store the HWND in
	// the global hash.  Also set the icon and current
	// HWND to the main HWND.

	WRITE_DEBUG("Init: Main Window");
	CreateMainWindow(NULL);

	WRITE_DEBUG("Modify: Always on top");
	if (HashSelect(HASH_GLOBAL,"Settings","AlwaysOnTop",HASH_INT))
		SetWindowPos(PROG_HWND,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);

	// Create the tab stops on the player pane.
	/*WRITE_DEBUG("Init: tab stops");
	temp = GetDlgItem(PROG_HWND, IDL_PLAYERS);
	tabstops[0] = 12;  tabstops[1] = 50;  tabstops[2] = 80;
	SendMessage(temp, LB_SETTABSTOPS, 3, (LPARAM)tabstops);*/


	// Program continues until running is 0 [occurs
	// when PostQuitMessage() is called ]
	running = 1;
	
	WRITE_DEBUG("Init: timed events");
	EVENTS_Register(ONLINELISTS_Download,NULL,1,0);
	EVENTS_Register(MOTD_Download,NULL,2,0);
	
	WRITE_DEBUG("Update: treeview");
	ListRebuildTreeview();

	LoadWindowPosition(PROG_HWND, SW_HIDE);

	WRITE_DEBUG("Init: System tray");
	SystrayUpdate();

	if (strstr(cmdline,"/min") || strstr(cmdline,"/MIN"))
	{
		HII(HASH_GLOBAL,"status","startmin",1);
		ShowWindow(PROG_HWND,SW_HIDE);
	}
	else
		ShowWindow(PROG_HWND,SW_SHOW);


#ifdef _SQCHAT
	if (HashSelectInt(HASH_GLOBAL,"Settings","ChatStart"))
	{
		HashInsertInt(HASH_GLOBAL,"Status","ChatStartup",1);
		CHAT_Login();
	}
#endif

	// Force DNS lookup
	GetServerAddr("www.serverquery.com");

	if (!allowNSLookups)
		list = HashFind(HASH_GLOBAL,"Lists","LANServers",(int)GroupOfList(NULL)->name);
		// Select LAN list
	else
		list = HashFind(HASH_GLOBAL,"Lists","Favourites",(int)GroupOfList(NULL)->name);
		// Select Favourites list

	
	if (list)
		TreeView_SelectItem(GetDlgItem(PROG_HWND,IDC_TREE1),list->hTreeItem);

	WRITE_DEBUG("start up");
	timeTaken();

	HITG("status","ddecommand","[abc(\"testing],\\\"asdf\",\"asdf\")][quit_two()]");

	
	while (running) 
	{
		WRITE_DEBUG("Run: start of loop");

		// check socket for writing
		checkSocket();

		// check for DDE command messages
		if (!DDE_CheckForCommand())
			break;

		// run any scheduled events
		WRITE_DEBUG("Run: events");
		EVENTS_Check();
		//timeTaken();

		// see if any data is available via requested
		// TCP connections
		WRITE_DEBUG("Run: HTTP");
		HTTP_Update();
		//timeTaken();


		// Check for incoming data packets
		WRITE_DEBUG("Run: data check");
		if (CheckIncomingData()) 
			CheckWaitingServer();
		WRITE_DEBUG("End: data check");
		//timeTaken();
		
		// if we're the foreground app, or the user
		// doesn't care, do normal server status stuff
		//WRITE_DEBUG("Run: foreground");
		if (HashSelect(HASH_GLOBAL,"Settings","PingWhileBackground",HASH_INT) ||
			GetForegroundWindow() == PROG_HWND ||
			GetActiveWindow() == GetForegroundWindow() ||
			HashSelectPtr(HASH_GLOBAL,"Status","WaitingServer"))
		{
			// Send out packets if the refresh
			// period has elapsed
			
			if ((unsigned)HashSelect(HASH_GLOBAL,"Status","NextRefresh",HASH_INT) < GetTickCount()) 
			{

				//MOTD_Download(0);

				WRITE_DEBUG("Running: data request");
				RequestNewData(1);
				WRITE_DEBUG("END: data request");
				//timeTaken();
			}

			// The same function is called to both begin and continue the LAN server requests
			// This stuff would probably be better handled by the event code... ?
			if ((unsigned)HashSelect(HASH_GLOBAL,"Status","NextLANRefresh",HASH_INT) < GetTickCount()) 
			{
				WRITE_DEBUG("Running: lan data request");
				RequestLANServersStatus();
				WRITE_DEBUG("END: lan data request");
				HashInsertInt(HASH_GLOBAL,"Status","NextLANRefresh",GetTickCount() + HSIG("Settings","LANRefresh") * 1000 );
				//timeTaken();
			}

			
			if ((int) HashSelect(HASH_GLOBAL,"Status","LANRefresh",HASH_INT) &&
				(unsigned) HashSelect(HASH_GLOBAL,"Status","LANRefresh",HASH_INT) < GetTickCount())
			{
				WRITE_DEBUG("Running: lan data request");
				RequestLANServersStatus();
				WRITE_DEBUG("Finished lan data request");
				//timeTaken();
			}
		}

		//WRITE_DEBUG("Message loop");
		// Standardish windows message loop
		if (PeekMessage(&msg, (HWND)NULL, 0, 0, PM_NOREMOVE))
		{
			if ((running = GetMessage(&msg, (HWND) NULL, 0, 0)) != 0 
				&& (!GetActiveWindow() || !IsDialogMessage(GetActiveWindow(),&msg)))
			{
				TranslateMessage(&msg);         
				DispatchMessage(&msg);     
			}
		} 
		//timeTaken();

		if (HashSelectPtr(HASH_GLOBAL,"Status","NewWindow")) {
			SetActiveWindow(HashSelectPtr(HASH_GLOBAL,"Status","NewWindow"));
			//BringWindowToTop(HashSelectPtr(HASH_GLOBAL,"Status","NewWindow"));
			//if (GetActiveWindow() != GetForegroundWindow())
				
			HashDelete(HASH_GLOBAL,"Status","NewWindow");
		}
		// Give some CPU time back to windows
		//WRITE_DEBUG("Sleep");
		Sleep(1L);
		timeTaken();
	}
	// Do necessary shut-down stuff, like
	// unloading DLL's etc
	WRITE_DEBUG("Shutdown");
	remove("sq.tmp");
	ShutDown();
	return msg.wParam;
}

