/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include "types.h"
#include "list.h"
#include "resource.h"
#include <stdio.h>
#include "disk.h"
#include "main.h"
#include <commctrl.h>
#include "ui.h"
#include "limits.h"

#define SV_NAME_LEN 64
#define SV_IPADDR_LEN 64

#define PROPERTIES_GetServer(x) ((server_t*)GetProp(x,"server"))
#define PROPERTIES_GetList(x) ((list_t*)GetProp(x,"list"))


int PROPERTIES_Save(HWND hwnd, int newserver) 
{
	char buf[64];
	server_t* svr = PROPERTIES_GetServer(hwnd);
	list_t* list = PROPERTIES_GetList(hwnd);
	int i;

	// Read server settings from dialog
	free(svr->name);
	svr->name = (char*)malloc(SV_NAME_LEN);
	SendMessage(GetDlgItem(hwnd, IDT_NAME), WM_GETTEXT, SV_NAME_LEN, (LPARAM)svr->name);

	free(svr->ipaddr);
	svr->ipaddr = (char*)malloc(SV_IPADDR_LEN);
	SendMessage(GetDlgItem(hwnd, IDT_IPADDR), WM_GETTEXT, SV_IPADDR_LEN, (LPARAM)svr->ipaddr);

	i = SendMessage(GetDlgItem(hwnd, IDL_TYPE), CB_GETCURSEL, 0, 0);
	SendMessage(GetDlgItem(hwnd, IDL_TYPE), CB_GETLBTEXT, i, (LPARAM)buf);
	for (i=0;i<128;i++)
		if (!strcmp(buf, GetServerType(i)->name)) {
			svr->gametype = GetServerType(i);
			HIIG("Settings","DefaultServerType",i);
			break;
		}

	if (IsDlgButtonChecked(hwnd, IDC_HOSTNAME) == BST_CHECKED)
	{
		free(svr->name);
		svr->name = malloc(SV_NAME_LEN);
		strcpy(svr->name,"New Server");
	}

	// Do something
	SendMessage(hwnd,WM_GETTEXT,32,(LPARAM)buf);

	if (!strcmp(buf,"Add Server"))
	{
		if (newserver)
		{
			AddServer(list,svr);
			svr = NewServer("New Server","","default");
			HashInsertPtr(svr,"lists",HASH_ANY,list);
			SetProp (hwnd,"server",(HANDLE)svr);
		}
		else FreeServer(svr);
	}
	else UI_UpdateServer(svr);
	
	return 0;
}			


int PROPERTIES_Close(HWND hwnd) 
{
	RemoveProp(hwnd,"server");
	RemoveProp(hwnd,"list");
	DestroyWindow(hwnd);
	return 0;
}			



int WINAPI PROPERTIES_DlgProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_COMMAND:
			if (LOWORD(wParam) == IDB_OK)
			{
				PROPERTIES_Save(hwnd,0);
				return PROPERTIES_Close(hwnd);
			}
				
			/*
			if (LOWORD(wParam) == IDB_MOVEDOWN)
			{
				server = PROPERTIES_GetServer(hwnd);
				list = PROPERTIES_GetList(hwnd);
				i = (int) HashExistsValue(list,"Servers",server,HASH_PTR);
				if (HashCount(list,"Servers") - 1 != i) {
					other = HashSelectPtr(list,"Servers",(HASH_PARAM)(i+1));
					HashMove(list,"Servers",i,i+1);
					temp.hTreeItem = server->hTreeItem;
					server->hTreeItem = other->hTreeItem;
					other->hTreeItem = temp.hTreeItem;
					UI_UpdateServer(server);
					UI_UpdateServer(other);
					TreeView_SelectItem(PROG_TREEVIEW,server->hTreeItem);
				}
				return 0;
			}
			if (LOWORD(wParam) == IDB_MOVEUP)
			{
				server = PROPERTIES_GetServer(hwnd);
				i = (int) HashExistsValue(server->list,"Servers",server,HASH_PTR);
				if (i != 0) {
					other = HashSelectPtr(server->list,"Servers",(HASH_PARAM)(i-1));
					HashMove(server->list,"Servers",i,i-1);
					temp.hTreeItem = server->hTreeItem;
					server->hTreeItem = other->hTreeItem;
					other->hTreeItem = temp.hTreeItem;
					UI_UpdateServer(server);
					UI_UpdateServer(other);
					TreeView_SelectItem(PROG_TREEVIEW,server->hTreeItem);
				}
			}
			*/
			else if (LOWORD(wParam) == IDB_ADD)
				PROPERTIES_Save(hwnd,1);

/*			else if (LOWORD(wParam) == IDT_NAME && HIWORD(wParam) == EN_KILLFOCUS)
			{
				server = PROPERTIES_GetServer(hwnd);
				free(server->name);
				server->name = (char*)malloc(SV_NAME_LEN);
				SendMessage(GetDlgItem(hwnd, IDT_NAME), WM_GETTEXT, SV_NAME_LEN, (LPARAM)server->name);
			}
			else if (LOWORD(wParam) == IDT_IPADDR && HIWORD(wParam) == EN_KILLFOCUS)
			{
				server = PROPERTIES_GetServer(hwnd);
				free(server->ipaddr);
				server->ipaddr = (char*)malloc(SV_IPADDR_LEN);
				SendMessage(GetDlgItem(hwnd, IDT_IPADDR), WM_GETTEXT, SV_IPADDR_LEN, (LPARAM)server->ipaddr);
			}
			else if (LOWORD(wParam) == IDL_TYPE && HIWORD(wParam) == CBN_SELCHANGE)
			{
				server = PROPERTIES_GetServer(hwnd);
				i = SendMessage(GetDlgItem(hwnd, IDL_TYPE), CB_GETCURSEL, 0, 0);
				SendMessage(GetDlgItem(hwnd, IDL_TYPE), CB_GETLBTEXT, i, (LPARAM)buf);
				for (i=0;i<20;i++)
					if (!strcmp(buf, GetServerType(i)->name)) {
						server->type = i;
						break;
					}
			}*/

			if (LOWORD(wParam) == IDC_HOSTNAME)
			{
				if (IsDlgButtonChecked(hwnd, IDC_HOSTNAME) == BST_CHECKED)
					EnableWindow(GetDlgItem(hwnd,IDT_NAME),0);
				else EnableWindow(GetDlgItem(hwnd,IDT_NAME),1);
			}
			if (wParam == IDCANCEL) {
				DestroyWindow(hwnd);
				return 0;
			}
			return 0;
		case WM_CLOSE:
			PROPERTIES_Close(hwnd);
			return 0;
	}
	return 0;
}


void PROPERTIES_DialogBox(HINSTANCE hinstance, HWND parent_hwnd, void* ptr, int add)
{
	HWND hwnd;
	int i, j = 0;
	server_t* svr;
	list_t* list;

	hwnd = CreateDialog(PROG_HINSTANCE, MAKEINTRESOURCE(IDD_PROPERTIES), parent_hwnd, PROPERTIES_DlgProc);
	
	if (add) 
	{
		//i = HSIG("Settings","DefaultServerType");
		//if (!i) i = 4;
		svr = NewServer("New Server","","HLS");
		//HashInsertPtr(svr,"lists",HASH_ANY,ptr);
		//svr->list = ptr;
		
		SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)"Add Server");
		SendMessage(GetDlgItem(hwnd,IDCANCEL),WM_SETTEXT,0,(LPARAM)"Close");
		//DestroyWindow(GetDlgItem(hwnd,IDCANCEL));
		ShowWindow(GetDlgItem(hwnd,IDB_OK),SW_HIDE);
	}
	else {
		svr = (server_t*) ptr;
		SendMessage(hwnd,WM_SETTEXT,0,(LPARAM)"Properties");
		//DestroyWindow(GetDlgItem(hwnd,IDB_ADD));
		ShowWindow(GetDlgItem(hwnd,IDB_ADD),SW_HIDE);

		list = HashSelectPtr(ptr,"lists",0);
		if (strcmp(GroupOfList(list)->name,"NetServers") &&
			strcmp(GroupOfList(list)->name,"Favourites") )
		{
			EnableWindow(GetDlgItem(hwnd,IDT_NAME),0);
			EnableWindow(GetDlgItem(hwnd,IDT_IPADDR),0);
			EnableWindow(GetDlgItem(hwnd,IDL_TYPE),0);
		}
		if (strcmp(GroupOfList(ptr)->name,"LANServers"))
		{
			EnableWindow(GetDlgItem(hwnd,IDB_MOVEUP),1);
			EnableWindow(GetDlgItem(hwnd,IDB_MOVEDOWN),1);
		}

	}
	for (i=0;i<MAX_SERVER_TYPES;i++) { 
		if (!GetServerType(i)->invalid) {
			SendMessage(GetDlgItem(hwnd, IDL_TYPE), CB_ADDSTRING, strlen(GetServerType(i)->name), (LPARAM)GetServerType(i)->name);
		}
	}
	
	SendMessage(GetDlgItem(hwnd, IDT_NAME), WM_SETTEXT, 0, (LPARAM)svr->name);
	SendMessage(GetDlgItem(hwnd, IDT_IPADDR), WM_SETTEXT, 0, (LPARAM)svr->ipaddr);
	SendMessage(GetDlgItem(hwnd, IDL_TYPE), CB_SELECTSTRING, -1, (LPARAM)svr->gametype->name);

	SetProp (hwnd,"server",(HANDLE)svr);
	SetProp (hwnd,"list",(HANDLE)ptr);
	
	ShowWindow(hwnd,SW_SHOW);
	return;
}
