/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/

#include <stdio.h>
#include "comm.h"
#include "resolve.h"

#include "types.h"
#include "list.h"
#include "http.h"

#include "groupshared.h"
#include "funcimport.h"


// Function definitions
/*
short int (*MAIN_GetServerPort)(char* addr, short otherwise);
unsigned long (*MAIN_GetServerAddr)(char* addr);
void (*MAIN_ServersListUpdateEnable)(int setting);
server_t* (*MAIN_NewServer)(char* name, char* ipaddr, char* qstat_type);
server_t* (*MAIN_AddServer)(list_t* list, server_t* svr);
void (*MAIN_ServersListUpdate)(int forced);
void (*MAIN_EmptyList)(list_t* list);
HASH* (*MAIN_HashAlloc)(void* owner, char*name, int size);
*/


char* master = "master3.idsoftware.com:27950";

int Q3MASTER_IsPacketFor(list_t* list, SOCKADDR* addr)
{
	SOCKADDR_IN* dest_addr = (SOCKADDR_IN*)addr;
	char* m = HashSelectText(list,"Extra","URL");
	int defport = 27950;

	if (strcmp(GroupOfList(list)->name,"QWMaster") == 0)
		defport = 27000;

	if (dest_addr->sin_addr.s_addr == MAIN_GetServerAddr(m)
		&& dest_addr->sin_port == htons(MAIN_GetServerPort(m, 27950)) )
		return 1;
	return 0;
}

group_t* GroupOfList(list_t* list) {
	return &(list->group);
}

void Q3MASTER_Packet(list_t* list, char* data, int len)
{
	char* end = data + len;
	char addr[128];
	int type = 6; // quake3

	char* intheader = "\xff\xff\xff\xff";
	char* q3header = "getserversResponse";
	char* qwheader = "d\n";
	
	// skip (int) -1 header
	if (memcmp(data,intheader,strlen(intheader)) == 0)
		data += strlen(intheader);
	if (memcmp(data,q3header,strlen(q3header)) == 0)
		data += strlen(q3header);
	else if (memcmp(data,qwheader,strlen(qwheader)) == 0)
	{
		type = 3; //qw 
		data += strlen(qwheader);
	}
	

	MAIN_ServersListUpdate(UPDATE_DISABLE);
	
	while (data < end) {
		
		if (type == 6)
		{
			data++; // skip the backslash
			if (end - data >=3 && memcmp(data,"EOT",3) == 0)
				break;
		}
		sprintf(addr,"%d.%d.%d.%d:%d",((unsigned char*)data)[0],((unsigned char*)data)[1],((unsigned char*)data)[2],((unsigned char*)data)[3],((unsigned char*)data)[4]*256+((unsigned char*)data)[5]);
		
		MAIN_AddServer(list, MAIN_NewServer(addr,addr,type));
		data += 6;
	}

	MAIN_ServersListUpdate(UPDATE_ENABLE);
	MAIN_ServersListUpdate(0);

	return;
}

void Q3MASTER_Refresh(list_t* list) {
	SOCKADDR_IN addr;
	char *str;
	int str_len = 0;
	int defport = 27950;

	char* m = HashSelectText(list,"Extra","URL");

	if (strcmp(GroupOfList(list)->name,"WolfMaster") == 0)
		str = "\xff\xff\xff\xffgetservers\r55\r\n";
	else if (strcmp(GroupOfList(list)->name,"Q3Master2") == 0)
		str = "\xff\xff\xff\xffgetservers\r48\r\n";
	else if (strcmp(GroupOfList(list)->name,"QWMaster") == 0)
	{
		str = "c\n\x0"; str_len = 3;
		defport = 27000;
	}
	else str = "\xff\xff\xff\xffgetservers\r45\r\n";

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = MAIN_GetServerAddr(m);
	addr.sin_port = htons(MAIN_GetServerPort(m, (short)defport));

	
	sendto(HashSelectInt(HASH_GLOBAL,"Global","Socket"), str, 
		str_len ? str_len : strlen(str), 0, (SOCKADDR*)&addr, sizeof(SOCKADDR_IN));

	MAIN_EmptyList(list);

	return;
}

// Create the 'Q3Master' group type
int __stdcall Init(group_shared_t* info) {

	group_t* grp;
	
	group_t* def = (group_t*)info->def;

	HashInit(info->hashes,info->numHashes);

	FUNC_Import();

	/*
	MAIN_GetServerPort = HashSelectPtr(HASH_GLOBAL,"Funcs","GetServerPort");
	MAIN_GetServerAddr = HashSelectPtr(HASH_GLOBAL,"Funcs","GetServerAddr");
	MAIN_ServersListUpdate = HashSelectPtr(HASH_GLOBAL,"Funcs","ServersListUpdate");
	MAIN_NewServer = HashSelectPtr(HASH_GLOBAL,"Funcs","NewServer");
	MAIN_AddServer = HashSelectPtr(HASH_GLOBAL,"Funcs","AddServer");
	MAIN_EmptyList = HashSelectPtr(HASH_GLOBAL,"Funcs","EmptyList");
	MAIN_HashAlloc = HashSelectPtr(HASH_GLOBAL,"Funcs","HashAlloc");
	*/

	/*grp = (group_t*)info->groups[0];
	strcpy(grp->name, "Q3Master");
	strcpy(grp->desc, "Quake3 Master (v1.17)");
	grp->delete_list = def->delete_list;
	grp->copy_server = def->copy_server;
	grp->refresh_list = Q3MASTER_Refresh;
	grp->is_packet_for = Q3MASTER_IsPacketFor;
	grp->packet = Q3MASTER_Packet;
	grp->properties_list = def->properties_list;
	MAIN_HashAlloc(grp,"extra",4);
	HashInsert(grp,"extra", "URL", "master3.idsoftware.com:27950",HASH_STRING);
	grp->location = grp;
	*/

	grp = (group_t*)info->groups[0];
	strcpy(grp->name, "Q3Master2");
	strcpy(grp->desc, "Quake3 Master");
	grp->delete_list = def->delete_list;
	grp->copy_server = def->copy_server;
	grp->refresh_list = Q3MASTER_Refresh;
	grp->is_packet_for = Q3MASTER_IsPacketFor;
	grp->packet = Q3MASTER_Packet;
	grp->properties_list = grp->properties_list;
	MAIN_HashAlloc(grp,"extra",4);
	HashInsert(grp,"extra", "URL", "master3.idsoftware.com:27950",HASH_STRING);
	grp->location = grp;
	

	grp = (group_t*)info->groups[1];
	strcpy(grp->name, "QWMaster");
	strcpy(grp->desc, "QuakeWorld Master");
	grp->delete_list = def->delete_list;
	grp->copy_server = def->copy_server;
	grp->refresh_list = Q3MASTER_Refresh;
	grp->is_packet_for = Q3MASTER_IsPacketFor;
	grp->packet = Q3MASTER_Packet;
	grp->properties_list = def->properties_list;
	MAIN_HashAlloc(grp,"extra",4);
	HashInsert(grp,"extra", "URL", "203.55.240.100:27000",HASH_STRING);
	grp->location = grp;

	
	grp = (group_t*)info->groups[2];
	strcpy(grp->name, "WolfMaster");
	strcpy(grp->desc, "RtCW Master");
	grp->delete_list = def->delete_list;
	grp->copy_server = def->copy_server;
	grp->refresh_list = Q3MASTER_Refresh;
	grp->is_packet_for = Q3MASTER_IsPacketFor;
	grp->packet = Q3MASTER_Packet;
	grp->properties_list = grp->properties_list;
	MAIN_HashAlloc(grp,"extra",4);
	HashInsert(grp,"extra", "URL", "wolfmaster.idsoftware.com:27950",HASH_STRING);
	grp->location = grp;
	

	return 3;
}