/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "resource.h"
#include "quake\resource.h"
#include "types.h"
#include "shared.h"
#include "quakex.h"
#include "servinfo.h"

#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))

SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);

void QW_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	sendto(s, "\xff\xff\xff\xffstatus", 11, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	return;
}


void QW_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
	HWND temp, hwnd;
	int i, j;
	char buf[128];
	//int num_players;

	hwnd = mainwnd;
	
	ServinfoInit(svr);

	strcpy(buf, GetValue(svr, "hostname") );
	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"items"); // wipe all current settings so we get blanks
		HII(svr,"info","num_players",0);
	} else if (HashCount(svr,"items") == 0) {
		HII(svr,"info","num_players",0);
		strcpy(buf, "CONNECTING...");
	}
	SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname",buf);
	ServinfoAdd("Address", svr->ipaddr);

	sprintf(buf, "QuakeWorld %s", GetValue(svr, "*version"));
	SetDlgItemText(hwnd, IDT_VERSION, buf); 
	ServinfoAdd("Game",buf);

	strcpy(buf, GetValue(svr,"mapname") );
	if (!strcmp( GetValue(svr, "mapname"), "") || GetValue(svr, "mapname") == NULL)
		strcpy(buf, GetValue(svr, "map") );
	SetDlgItemText(hwnd, IDT_MAPNAME, buf);
	ServinfoAdd("Map", buf);

	if (strcmp(GetValue(svr, "maxclients"), ""))
	{
		sprintf(buf, "%d / %s", SVR_NUMPLAYERS, GetValue(svr, "maxclients") );
	}
	else strcpy(buf,"");
	SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	ServinfoAdd("Players",buf);

	ServinfoAdd("Ping","");

	if (HSIG("Settings","ExtraServInfo")) {
	strcpy(buf, GetValue(svr, "fraglimit") );
	SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf);
	ServinfoAdd("Fraglimit", buf);
	
	strcpy(buf, GetValue(svr, "timelimit") );
	SetDlgItemText(hwnd, IDT_TIMELIMIT, buf );
	ServinfoAdd("Timelimit", buf);

	SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Deathmatch:" );
	strcpy(buf, GetValue(svr, "deathmatch") );
	SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
	ServinfoAdd("Deathmatch",buf);

	SetDlgItemText(hwnd, IDC_DMFLAGS, "Teamplay:" );
	strcpy(buf, GetValue(svr, "teamplay") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf);
	ServinfoAdd("Teamplay", buf);

	SetDlgItemText(hwnd,IDT_MISC,"");
}
	PlayerinfoInit(svr);

	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"\tFrags\tPing\tName");
	
	for (i=0;i<SVR_NUMPLAYERS;i++)
	{
		strcpy(buf,"\t");

		if (strcmp(clan_id,""))
		{
			for (j=0;j<(int)(strlen(SVR_NAME(i)) - strlen(clan_id));j++)
				if (!memcmp(SVR_NAME(i)+j, clan_id, strlen(clan_id)))
					strcpy(buf,"*\t");
		}
		
		if (!ignore_changes && SVR_CHANGE(i))
			sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
				SVR_FRAGS(i),
				SVR_CHANGE(i) > 0 ? '+' : ' ',
				SVR_CHANGE(i),
				SVR_PING(i),
				SVR_NAME(i)
			);
		else 
			sprintf(buf + strlen(buf), "%d\t%d\t%s", 
				SVR_FRAGS(i),
				SVR_PING(i),
				SVR_NAME(i)
			);
		
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);

		PlayerinfoAdd(SVR_NAME(i),SVR_FRAGS(i),SVR_CHANGE(i),SVR_PING(i));
	}
	return;
}

char* QW_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}

int QW_Receive(server_t* svr, char* buf, int len)
{
	buf[len] = 0;
	WRITE_DEBUG(buf);
	DecodeString(svr, buf, len);

	// Update available information
	HashInsertText(svr,"stats","map",
		HashSelectText(svr,"items","map"));
	HashInsertInt(svr,"stats","numplayers",
		SVR_NUMPLAYERS );
	HashInsertText(svr,"stats","maxplayers",
		HashSelectText(svr,"items","maxclients") );

	return 1; // update displayed info
}

int QUAKE2_IsFull(server_t* svr)
{
	if (SVR_NUMPLAYERS && atoi( GetValue(svr, "maxclients") ) &&
			SVR_NUMPLAYERS == atoi( GetValue(svr, "maxclients") ) )
		return 1;
	return 0;
}
int __stdcall Init(shared_t* info)
{
	types_t* t;

	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");

	t = info->types[0];
	//t->type = 3;
	t->protocol = 1;
	t->port = 27500;
	strcpy(t->name, "QuakeWorld");
	strcpy(t->path, "C:\\Quake\\glqwcl.exe");
	strcpy(t->qstat_type, "QWS");
	strcpy(t->gamespy_type, "quakeprot");
	strcpy(t->mod_var,"gamename");
	
	t->receive = QW_Receive;
	t->print_info = QW_PrintInfo;
	t->send = QW_Send;
	t->hostname = QW_Hostname;
	t->isfull = QUAKE2_IsFull;

	//HashInsertPtr(t,"funcs","isfull",QUAKE2_IsFull);

/* (void*) t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hinstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)
	); */

	return 1;
}

